/*
  $Id: sp_postgresql_sportsdb_delete_event_data_by_id.sql 2876 2011-04-12 21:36:57Z ian $
  
  WARNING:
  This code removes data from a SportsDB database. It has been reviewed and tested in an attempt
  to ensure that it does not harm the integrity of a SportsDB database or remove data that it
  should not, but it may not function as anticipated. This code comes with absolutely no
  guarantees. It is your responsibility to check this code to ensure that you are satisfied that
  it will work as expected. XML Team Inc. takes no responsibility for any unexpected data loss
  or data corruption that may occur as a result of the execution of this code on your database.  
  
  This is a stored procedure to clear out the events table and all dependent table records by event_id.
*/

CREATE OR REPLACE PROCEDURE sportsdb_delete_event_data_by_id
(in_event_id INTEGER) RETURNS INTEGER AS $$
BEGIN
  DECLARE
    event_id_test INTEGER;

    v_no_more_rows BOOLEAN;
    v_count INT;

    v_stat_id INTEGER;
    v_repository_type VARCHAR(120);
    v_repository_id INTEGER;
    v_period_id INTEGER;

    /* A cursor to look up stats related to periods tied to an event. */
    cur_period_stats CURSOR FOR
    SELECT stats.id, stats.stat_repository_type, stats.stat_repository_id FROM stats
    JOIN participants_events ON (participants_events.event_id = in_event_id)
    JOIN periods ON (periods.participant_event_id = participants_events.id)
    WHERE stats.stat_coverage_type = 'periods'
    AND stats.stat_coverage_id = periods.id;
        
    /* A cursor to look up stats related to an event. */
    cur_event_stats CURSOR FOR
    SELECT stats.id, stats.stat_repository_type, stats.stat_repository_id FROM stats
    WHERE stats.stat_coverage_type = 'events'
    AND stats.stat_coverage_id = in_event_id;

    /* A cursor to look up periods tied to an event. */
    cur_periods CURSOR FOR
    SELECT id FROM periods
    WHERE periods.participant_event_id IN (
      SELECT participants_events.id FROM participants_events
      WHERE participants_events.event_id = in_event_id  
    );

  BEGIN


  DECLARE CONTINUE HANDLER FOR NOT FOUND SET v_no_more_rows = TRUE;

  SELECT id INTO event_id_test FROM events
  WHERE events.id = in_event_id;
  
  /* Leave if no event_id found to match the one passed in */
  IF event_id_test IS NULL THEN
    LEAVE main;
  END IF;
  
  /*
    Stats-table-related clearing SQL
  */

  /* 
    Delete stats tied to periods and events using cursors and
    dynamic queries aimed at the stats table and the
    table each stats record points to. 

    Doing so via individual queries per sports stats table
    is brittle (new tables will be missed) and requires
    repeated expensive joins to the stats table and
    per-sport stats tables that can be very large also. 
  */

  OPEN cur_period_stats;
  period_stats_loop: LOOP
    FETCH cur_period_stats INTO v_stat_id, v_repository_type, v_repository_id;
    
    IF v_no_more_rows THEN
      SET v_no_more_rows = FALSE;
      CLOSE cur_period_stats;
      LEAVE period_stats_loop;
    END IF;

    SET @s = CONCAT('DELETE FROM ', v_repository_type, ' WHERE id = ', v_repository_id);
    PREPARE stmt FROM @s;
    EXECUTE stmt;
    DEALLOCATE PREPARE stmt;
    
    SET @s = CONCAT('DELETE FROM stats where id = ', v_stat_id);
    PREPARE stmt FROM @s;
    EXECUTE stmt;
    DEALLOCATE PREPARE stmt;
  END LOOP period_stats_loop;

  /* Stats tied to events */

  OPEN cur_event_stats;
  event_stats_loop: LOOP
    FETCH cur_event_stats INTO v_stat_id, v_repository_type, v_repository_id;
    
    IF v_no_more_rows THEN
      SET v_no_more_rows = FALSE;
      CLOSE cur_event_stats;
      LEAVE event_stats_loop;
    END IF;

    SET @s = CONCAT('DELETE FROM ', v_repository_type, ' WHERE id = ', v_repository_id);
    PREPARE stmt FROM @s;
    EXECUTE stmt;
    DEALLOCATE PREPARE stmt;
    
    SET @s = CONCAT('DELETE FROM stats where id = ', v_stat_id);
    PREPARE stmt FROM @s;
    EXECUTE stmt;
    DEALLOCATE PREPARE stmt;
  END LOOP event_stats_loop;
  

  /*
    American Football event-related clearing SQL
  */

  SELECT count(*) INTO v_count 
  FROM american_football_event_states 
  WHERE event_id =  in_event_id; 

  IF v_count > 0 THEN
    DELETE FROM american_football_action_participants
    WHERE american_football_action_participants.american_football_action_play_id IN (
      SELECT american_football_action_plays.id FROM american_football_action_plays
      JOIN american_football_event_states ON (
           american_football_event_states.id =  american_football_action_plays.american_football_event_state_id)
       JOIN events ON american_football_event_states.event_id = in_event_id  
     );

    DELETE FROM american_football_action_plays
    WHERE american_football_action_plays.american_football_event_state_id IN (
      SELECT american_football_event_states.id  FROM american_football_event_states
      WHERE american_football_event_states.event_id = in_event_id  
    );

    DELETE FROM american_football_event_states
    WHERE american_football_event_states.event_id = in_event_id;
  END IF;

  /*
    Baseball event-related clearing SQL
  */

  SELECT count(*) INTO v_count 
  FROM baseball_event_states 
  WHERE event_id = in_event_id; 

  IF v_count > 0 THEN
    DELETE FROM baseball_action_contact_details
    WHERE baseball_action_contact_details.baseball_action_pitch_id IN (
      SELECT baseball_action_pitches.id FROM baseball_action_pitches
      JOIN baseball_action_plays ON (baseball_action_plays.id=baseball_action_pitches.baseball_action_play_id)
      JOIN baseball_event_states ON baseball_event_states.id =  baseball_action_plays.baseball_event_state_id
      WHERE baseball_event_states.event_id = in_event_id
    );

    DELETE FROM baseball_action_pitches
    WHERE baseball_action_pitches.baseball_action_play_id IN (
      SELECT baseball_action_plays.id FROM baseball_action_plays
      JOIN baseball_event_states ON baseball_event_states.id =  baseball_action_plays.baseball_event_state_id
      WHERE baseball_event_states.event_id = in_event_id
    );

    DELETE FROM baseball_action_plays
    WHERE baseball_action_plays.baseball_event_state_id IN (
      SELECT baseball_event_states.id  FROM baseball_event_states
      WHERE baseball_event_states.event_id = in_event_id
    );

    DELETE FROM baseball_action_substitutions
    WHERE baseball_action_substitutions.baseball_event_state_id IN (
      SELECT baseball_event_states.id  FROM baseball_event_states
      WHERE baseball_event_states.event_id = in_event_id
    );

    DELETE FROM baseball_event_states
    WHERE baseball_event_states.event_id = in_event_id;
  END IF;

  /*
    Basketball event-related clearing SQL
  */

  SELECT count(*) INTO v_count 
  FROM basketball_event_states 
  WHERE event_id = in_event_id; 

  IF v_count > 0 THEN
    DELETE FROM basketball_event_states
    WHERE basketball_event_states.event_id = in_event_id;
  END IF;

  /*
    Ice Hockey event-related clearing SQL
  */

  SELECT count(*) INTO v_count 
  FROM ice_hockey_event_states 
  WHERE event_id = in_event_id; 

  IF v_count > 0 THEN
    DELETE FROM ice_hockey_action_participants
    WHERE ice_hockey_action_participants.ice_hockey_action_play_id IN (
      SELECT ice_hockey_action_plays.id FROM ice_hockey_action_plays
      JOIN ice_hockey_event_states ON ice_hockey_event_states.id =  ice_hockey_action_plays.ice_hockey_event_state_id
      WHERE ice_hockey_event_states.event_id = in_event_id  
    );

    DELETE FROM ice_hockey_action_plays
    WHERE ice_hockey_action_plays.ice_hockey_event_state_id IN (
      SELECT ice_hockey_event_states.id  FROM ice_hockey_event_states
      WHERE ice_hockey_event_states.event_id = in_event_id  
    );

    DELETE FROM ice_hockey_event_states
    WHERE ice_hockey_event_states.event_id = in_event_id;
  END IF;

  SELECT count(*) INTO v_count 
  FROM tennis_event_states 
  WHERE event_id = in_event_id; 

  IF v_count > 0 THEN
    DELETE FROM tennis_event_states
    WHERE tennis_event_states.event_id = in_event_id;
  END IF;
  
  /*
    General event state tables
  */

  SELECT count(*) INTO v_count 
  FROM event_states 
  WHERE event_id = in_event_id; 

  IF v_count > 0 THEN
    DELETE FROM event_action_fouls
    WHERE event_action_fouls.event_state_id IN (
      SELECT event_states.id  FROM event_states
      WHERE event_states.event_id = in_event_id  
    );

    DELETE FROM event_action_penalties
    WHERE event_action_penalties.event_state_id IN (
      SELECT event_states.id FROM event_states
      WHERE event_states.event_id = in_event_id  
    );

    DELETE FROM event_action_substitutions
    WHERE event_action_substitutions.event_state_id IN (
      SELECT event_states.id FROM event_states
      WHERE event_states.event_id = in_event_id  
    );

    DELETE FROM event_action_participants
    WHERE event_action_participants.event_action_play_id IN (
      SELECT event_action_plays.id FROM event_action_plays
      JOIN event_states ON event_states.id =  event_action_plays.event_state_id
      WHERE event_states.event_id = in_event_id  
    );

    DELETE FROM event_action_plays
    WHERE event_action_plays.event_state_id IN (
      SELECT event_states.id FROM event_states
      WHERE event_states.event_id = in_event_id  
    );

    DELETE FROM event_states
    WHERE event_states.event_id = in_event_id;
  END IF;
  
  
  /*
    Common event-related clearing SQL
  */

  DELETE FROM wagering_moneylines
  WHERE event_id = in_event_id;

  DELETE FROM wagering_odds_lines
  WHERE event_id = in_event_id;

  DELETE FROM wagering_runlines
  WHERE event_id = in_event_id;

  DELETE FROM wagering_straight_spread_lines
  WHERE event_id = in_event_id;

  DELETE FROM wagering_total_score_lines
  WHERE event_id = in_event_id;

  DELETE FROM weather_conditions
  WHERE event_id = in_event_id;

  /* Delete period data using a cursor to speed things up (by 50% in test) */
  OPEN cur_periods;
  periods_loop: LOOP
    FETCH cur_periods INTO v_period_id;

    IF v_no_more_rows THEN
      SET v_no_more_rows = FALSE;
      CLOSE cur_periods;
      LEAVE periods_loop;
    END IF;
    
    DELETE FROM sub_periods
    WHERE sub_periods.period_id = v_period_id;

    DELETE FROM periods
    WHERE id = v_period_id;
  END LOOP periods_loop;  

  DELETE FROM participants_events
  WHERE participants_events.event_id = in_event_id;

  DELETE FROM affiliations_events
  WHERE affiliations_events.event_id = in_event_id;

  DELETE FROM events_sub_seasons
  WHERE events_sub_seasons.event_id = in_event_id;

  DELETE FROM document_fixtures_events
  WHERE document_fixtures_events.event_id = in_event_id;

  DELETE FROM events_documents
  WHERE events_documents.event_id = in_event_id;

  DELETE FROM events_sub_seasons
  WHERE events_sub_seasons.event_id = in_event_id;

  DELETE FROM person_event_metadata
  WHERE person_event_metadata.event_id = in_event_id;

  DELETE FROM events
  WHERE events.id = in_event_id;

END;$$

delimiter ;