# $Id: ActiveFixtures.pm 2167 2008-08-22 18:17:19Z alan $
package Conf::ActiveFixtures;

# An object-oriented active fixtures config reading module.
# Based on the Core::ActiveFixtures module which is not Object Oriented

use strict;
use FileHandle;
use Carp;

use Core::Util;
use Core::LineParser;
use Core::XmlUtil; 
use Core::Configurator;

# Module-scope hash to hold active fixture sets (read once used many times)
my $active_fixture_sets = undef;

sub new {
    my ($class, $doc) = @_;

    my $self = {};
    bless $self, $class;
    
    $self->{doc} = $doc;
    $self->{shared_data} = Util::SharedData->new();
	$self->{cfg} = $self->{shared_data}->get('cfg');
    if (! defined $active_fixture_sets) {
		$active_fixture_sets = {};
	   	$self->load_all($self->{cfg});
    	$self->{shared_data}->set('active_fixture_sets', $active_fixture_sets);
    }
    $self->{active_fixture_sets} = $self->{shared_data}->get('active_fixture_sets');
    
    return $self;   
}

sub load_all {
	my ($self, $cfg) = @_;
	
	foreach my $key (
	qw(
		active-fixtures-events 
		active-fixtures-participants-events 
		active-map-events-fixtures 
		active-fixtures-players
		active-fixtures-injuries
		active-fixtures-league-directory
		active-fixtures-award
		active-fixtures-ranking
		active-fixtures-record
		active-fixtures-standing
		active-fixtures-result
		active-fixtures-subscore
		active-fixtures-wagering
		active-fixtures-weather
		event-level-lookup
	)
	) {
		my $loaded = $self->load($key);
	}
}

# Returns a hash_ref: $active_fixtures->{hi_level.hi_key}->{fixture_key} => value
# There are two versions:
#   1. <hi_level>/<hi_level_key>:<fixture_list>
#   2. <hi_level>/<hi_level_key>:<fixture_list>:<some setting>
# The first version just defaults the setting to 1

sub load {
	my ($self, $key) = @_;

	my $cfg = $self->{cfg};
	# Test for required params
	$cfg->required_param('hierarchy_levels');

	my $active_path;
	if ( !($active_path = $cfg->param($key)) ) {
		my $param_key = "active_fixtures_dir";
		my $active_dir = $cfg->required_param($param_key);
		$active_path = "$active_dir/$key.dat";
	}

	die "No such file $active_path" if !-f $active_path;
	my @lines = Core::LineParser::load($active_path);
	my $active_fixtures = {};
	foreach my $line (@lines) {
		next if $line =~ /^#.*?$/;
		my ($hi, $fixture_list, $value) = split /:/, $line;
		
		$value = trim($value);
		$value = 1 if !defined $value;
		$hi = Core::Util::trim($hi);
		die "Error in $active_path: Bad format, line: $line" if 
			!$hi;
		die "Error in $active_path: Hierarchy Level must be " . 
			"<hierarchy_level>/<hierarchy_key>, line $line" if
			$hi ne '*' && $hi !~ m(/);
		die "Error in $active_path: No fixture list specified, line: $line" if 
			!$fixture_list;
		my @fixtures = split(/, ?/, $fixture_list);
		die "Error in $active_path: No fixtures in list, line: $line" if !@fixtures;

		foreach my $fixture (@fixtures) {
			$fixture = Core::Util::trim($fixture);
			$active_fixtures->{$hi}->{$fixture} = $value;
		}
	}

	$active_fixture_sets->{$key} = $active_fixtures;
	return $active_fixtures;
}

sub excludelookup {
	my ($self, $ref, $hi_level, $hi_key, $fixture_key) = @_;
	my $hi = "$hi_level/$hi_key";
	my $hi_ref = $ref->{$hi};

	return $hi_ref && ($hi_ref->{'*'} || $hi_ref->{$fixture_key});
}

sub lookup {
	my ($self, $ref, $hi_level, $hi_key, $fixture_key) = @_;

#	if ($ref->{'*'}) {
#		return $ref->{'*'}->{'*'} ||
#			$ref->{'*'}->{$fixture_key};
#	} 
#	else {
		my $hi = "$hi_level/$hi_key";

		my $hi_ref = $ref->{$hi};

		return $hi_ref && ($hi_ref->{'*'} || $hi_ref->{$fixture_key});
#	}
}

sub wildcardlookup {
	my ($self, $ref, $hi_level, $hi_key, $fixture_key) = @_;

	if ($ref->{'*'}) {
		return $ref->{'*'}->{'*'} ||
			$ref->{'*'}->{$fixture_key};
	} 
}

# Do lookup for current doc
sub doc_lookup {
	my ($self, $key) = @_;
 
 	my $doc = $self->{doc};
 	my $cfg = $self->{cfg};
	my $ref = $self->{active_fixture_sets}->{$key};

	# TODO - Add a load on demand if an active fixtures set has not been loaded yet

	my $fixture_key = required_attribute_xpath($doc, 
		'/sports-content/sports-metadata/@fixture-key');

	my $supported = $self->load_supported_code_types($cfg);	
	my $content = $self->get_content_codes($supported, $doc);
	
	# first look for excludes
	while ( my ($code_type, $code_key) = each %$content ) {
		my $active = $self->excludelookup($ref, $code_type, $code_key, $fixture_key);
		if ($active eq 'exclude') {
			# print "\n\nIMMEDIATE EXCLUDE!\n\n";
			return undef;
		}
	}

	# look for direct includes
	while ( my ($code_type, $code_key) = each %$content ) {
		my $active = $self->lookup($ref, $code_type, $code_key, $fixture_key);
		if (defined $active) {
			# print "\n\nMatched Directly!\n\n";
			return $active;
		}
	}

	# look for wildcard includes
	while ( my ($code_type, $code_key) = each %$content ) {
		my $active = $self->wildcardlookup($ref, $code_type, $code_key, $fixture_key);
		if (defined $active) {
			# print "\n\nMatched WILDCARD!\n\n";
			return $active;
		}
	}

	return undef;
}

sub load_supported_code_types {
	my ($self) = @_;

	my $cfg = $self->{cfg};
	croak "cfg not defined" if !$cfg;
	my $codes = $cfg->required_param('hierarchy_levels');
	my @codes = split(/,/, $codes);
	my %supported_code_types = map { Core::Util::trim($_) => 1 } @codes;

	return \%supported_code_types;
}

sub get_content_codes {
	my ($self, $supported) = @_;

 	my $doc = $self->{doc};

	my @sport_content_nodes = 
		$doc->findnodes('/sports-content/sports-metadata/' .
		'sports-content-codes/sports-content-code');

	my %codes = ();
	foreach my $node (@sport_content_nodes) {
		my $code_type = $node->getAttribute('code-type');
		my $code_key = $node->getAttribute('code-key');

		if ($code_type && $code_key && $supported->{$code_type}) {
			$codes{$code_type} = $code_key;
		}
	}

	return \%codes;
}


1;