# $Id: ActiveFixtures.pm 2167 2008-08-22 18:17:19Z alan $
package Core::ActiveFixtures;

use strict;
use FileHandle;
use Carp;

use Core::Util;
use Core::LineParser;
use Core::XmlUtil; 
use Core::Configurator;

# Returns a hash_ref: $active_fixtures->{hi_level.hi_key}->{fixture_key} => value
# There are two versions:
#   1. <hi_level>/<hi_level_key>:<fixture_list>
#   2. <hi_level>/<hi_level_key>:<fixture_list>:<some setting>
# The first version just defaults the setting to 1

sub load {
	my ($cfg, $key) = @_;

	# Test for required params
	$cfg->required_param('hierarchy_levels');

	my $active_path;
	if ( !($active_path = $cfg->param($key)) ) {
		my $param_key = "active_fixtures_dir";
		my $active_dir = $cfg->required_param($param_key);
		$active_path = "$active_dir/$key.dat";
	}

	die "No such file $active_path" if !-f $active_path;
	my @lines = Core::LineParser::load($active_path);
	my $active_fixtures = {};
	foreach my $line (@lines) {
		next if $line =~ /^#.*?$/;
		my ($hi, $fixture_list, $value) = split /:/, $line;
		
		$value = trim($value);
		$value = 1 if !defined $value;
		$hi = Core::Util::trim($hi);
		die "Error in $active_path: Bad format, line: $line" if 
			!$hi;
		die "Error in $active_path: Hierarchy Level must be " . 
			"<hierarchy_level>/<hierarchy_key>, line $line" if
			$hi ne '*' && $hi !~ m(/);
		die "Error in $active_path: No fixture list specified, line: $line" if 
			!$fixture_list;
		my @fixtures = split(/, ?/, $fixture_list);
		die "Error in $active_path: No fixtures in list, line: $line" if !@fixtures;

		foreach my $fixture (@fixtures) {
			$fixture = Core::Util::trim($fixture);
			$active_fixtures->{$hi}->{$fixture} = $value;
		}
	}

	return $active_fixtures;
}

sub excludelookup {
	my ($ref, $hi_level, $hi_key, $fixture_key) = @_;
	my $hi = "$hi_level/$hi_key";
	my $hi_ref = $ref->{$hi};

	return $hi_ref && ($hi_ref->{'*'} || $hi_ref->{$fixture_key});
}

sub lookup {
	my ($ref, $hi_level, $hi_key, $fixture_key) = @_;

#	if ($ref->{'*'}) {
#		return $ref->{'*'}->{'*'} ||
#			$ref->{'*'}->{$fixture_key};
#	} 
#	else {
		my $hi = "$hi_level/$hi_key";

		my $hi_ref = $ref->{$hi};

		# print hi_ref: " . $hi_ref . ($hi_ref->{'*'} || $hi_ref->{$fixture_key}) . "\n";

		return $hi_ref && ($hi_ref->{'*'} || $hi_ref->{$fixture_key});
#	}
}

sub wildcardlookup {
	my ($ref, $hi_level, $hi_key, $fixture_key) = @_;

	if ($ref->{'*'}) {
		return $ref->{'*'}->{'*'} ||
			$ref->{'*'}->{$fixture_key};
	} 
}

# Do lookup for current doc
sub doc_lookup {
	my ($ref, $doc, $cfg) = @_;
 
	my $fixture_key = required_attribute_xpath($doc, 
		'/sports-content/sports-metadata/@fixture-key');

	my $supported = load_supported_code_types($cfg);	
	my $content = get_content_codes($supported, $doc);
	
	# first look for excludes
	while ( my ($code_type, $code_key) = each %$content ) {
		my $active = excludelookup($ref, $code_type, $code_key, $fixture_key);
		if ($active eq 'exclude') {
			# print "\n\nIMMEDIATE EXCLUDE!\n\n";
			return undef;
		}
	}

	# look for direct includes
	while ( my ($code_type, $code_key) = each %$content ) {
		my $active = lookup($ref, $code_type, $code_key, $fixture_key);
		if (defined $active) {
			# print "\n\nMatched Directly!\n\n";
			return $active;
		}
	}

	# look for wildcard includes
	while ( my ($code_type, $code_key) = each %$content ) {
		my $active = wildcardlookup($ref, $code_type, $code_key, $fixture_key);
		if (defined $active) {
			# print "\n\nMatched WILDCARD!\n\n";
			return $active;
		}
	}

	return undef;
}

sub load_supported_code_types {
	my ($cfg) = @_;

	croak "cfg not defined" if !$cfg;
	my $codes = $cfg->required_param('hierarchy_levels');
	my @codes = split(/,/, $codes);
	my %supported_code_types = map { Core::Util::trim($_) => 1 } @codes;

	return \%supported_code_types;
}

sub get_content_codes {
	my ($supported, $doc) = @_;

	my @sport_content_nodes = 
		$doc->findnodes('/sports-content/sports-metadata/' .
		'sports-content-codes/sports-content-code');

	my %codes = ();
	foreach my $node (@sport_content_nodes) {
		my $code_type = $node->getAttribute('code-type');
		my $code_key = $node->getAttribute('code-key');

		if ($code_type && $code_key && $supported->{$code_type}) {
			$codes{$code_type} = $code_key;
		}
	}

	return \%codes;
}


1;