# $Id: AsciiTable.pm 2125 2008-07-23 19:27:14Z ian $
# AsciiTable - used in conjunction with SimpleHtmlFormatter

package Core::AsciiTable;

use strict;

use HTML::Entities;

sub new {
	my ($class, $col_sep) = @_;

	my $self = {};
	bless $self, $class;
	$self->_initialize($col_sep);
	return $self;
}

sub _initialize {
	my ($self, $col_sep) = @_;

	$self->{'rows'} = [];
	$self->{'col_sep'} = $col_sep;
}

sub text {
	my ($self, $text) = @_;

	$self->{'cell_text'} .= $text if defined $self->{'cell_text'};
}

sub cell_start {
	my ($self, $align, $colspan) = @_;

	$self->{'cell_text'} = "";
	$self->{'align'} = $align;
	$self->{'colspan'} = $colspan;
}

sub cell_end {
	my ($self) = @_;

	die "Bad HTML structure: unmatched closing tag" if (!defined $self->{'cell_text'});
	$self->add_cell(purge($self->{'cell_text'}), $self->{'align'}, $self->{'colspan'});
	$self->{'cell_text'} = undef;
	$self->{'align'} = undef;
}

sub row_start {
	my ($self) = @_;

	$self->{'row'} = [];
}

sub row_end {
	my ($self) = @_;

	$self->add_row();
}

sub init_table {
	my ($self, $column_headings) = @_;

	$self->{'rows'} = [];
}

sub add_cell {
	my ($self, $text, $align, $colspan) = @_;

	push @{$self->{'row'}}, {  
		'text' => $text, 'align' => $align, 'colspan' => $colspan 
	};
}

sub add_row {
	my ($self) = @_;

	die "Bad HTML structure: unmatched row closing tag" 
		if !defined $self->{'row'};
	die "Bad HTML structure: row must be inside table" 
		if !defined $self->{'rows'};
	push @{$self->{'rows'}}, $self->{'row'};
	$self->{'row'} = undef;
}

sub table_lines {
	my ($self) = @_;

	my $idx = 0;
	my $cell;

	# Gather intelligence
	my @widths = $self->compute_widths();

	# Write data
	my @rows = ();
	foreach my $row (@{$self->{'rows'}}) {
		$idx = 0;
		my @rowtext = ();
		foreach $cell (@$row) {
			my $text = $cell->{'text'};
			my $align = $cell->{'align'} || 'center';
			my $colspan = $cell->{'colspan'} || 1;
			my $end_idx = $idx + $colspan - 1;
			my $separator_width = (length $self->{'col_sep'}) * ($colspan - 1);
			my $colwidth = sum(@widths[$idx..$end_idx]) + $separator_width;
			
			my $width = length $text;
			my $celltext;
			if ($align eq 'left') {
				$celltext = $text . (" " x ($colwidth - $width));
			} elsif ($align eq 'right') {
				$celltext = (" " x ($colwidth - $width)) . $text;
			} else {	# Default is center
				my $pad = $colwidth - $width;
				my $start = int($pad / 2);
				my $end = $pad - $start;
				$celltext = (" " x $start) . $text . (" " x $end);
			}

			push @rowtext, $celltext;
			$idx += $colspan;
		}
		push @rows, join($self->{'col_sep'}, @rowtext);
	}

	return @rows;
}

sub sum {
	my @elements = @_;

	my $sum = 0;
	foreach my $element (@elements) { $sum += $element; }

	return $sum;
}

sub compute_widths {
	my ($self) = @_;

	my @widths = ();

	foreach my $row (@{$self->{'rows'}}) {
		my $idx = 0;
		foreach my $cell (@$row) {
			my $text = $cell->{'text'};
			my $colspan = $cell->{'colspan'} || 1;
			my $separator_width = (length $self->{'col_sep'}) * ($colspan - 1);
			my $width = $separator_width + (length $text);

			# A little arithmetical wizardry
			my $nom_width = int( $width / $colspan );
			my $residual = $width % $colspan;
			my $non_residual = $colspan - $residual;
			my @col_widths = ( ( $nom_width + 1 ) x $residual, 
				( $nom_width ) x $non_residual );

			foreach $width (@col_widths) {
				$widths[$idx] = $width if !$widths[$idx] || $width > $widths[$idx];
				$idx++;
			}
		}
	}
	return @widths;
}

sub purge {
	my ($txt) = @_;
	$txt =~ s/^\s+//;
	$txt =~ s/\s+$//; 
	$txt =~ s/[\r\n\t]/ /g;
	$txt =~ s/  +/ /g;
	return decode_entities($txt);
}

sub purge_pre {
	my ($txt) = @_;
	return decode_entities($txt);
}

1;