# $Id: Configurator.pm 3081 2016-04-07 20:13:07Z ian $

package Core::Configurator;

use strict;

use FileHandle;
use Carp;

use File::Basename;
use Core::Util;
use Core::Configurator_overrides;

sub new {
    my ($class, $home_path, $config_name) = @_;

	my $self = {};
	bless $self, $class;
	$self->_initialize($home_path, $config_name);
	return $self;
}

sub _initialize {
	my ($self, $home_path, $config_name) = @_;

	$self->{'home_path'} = $home_path;
#	my $config_path = $self->{'config_path'} = _get_config_path($home_path, $config_name);
	my $config_path = $self->{'config_path'} = _get_config_path($home_path, $config_name);
	$self->{name} = $config_name;
	$self->{'keys'} = {};

	# Read in config file
	my $file = new FileHandle $config_path or 
		die "Cannot open configuration file $config_path: $!";

	while (my $line = <$file>) {
		$line = _genchomp($line);
		next if $line =~ /^#/ || $line =~ /^\s*$/;
		
		my ($key, $val) = split /\s*=\s*/, $line, 2;
		next if (!defined $key || !defined $val);
		$key = lc(trim($key));
		$self->{'keys'}->{$key} = trim($val);
	}
	
	undef $file;

    # From the module Config_overrides
    $self->do_override_config();
}

sub required_param {
	my ($self, $key) = @_;

	my $display_key = ucfirst($key);
	my $param = $self->param($key);
	defined $param or croak "$display_key is required in config file: " . 
		$self->cfg_file();

	return $param;
}

sub name {
	my ($self) = @_;

	return $self->{name};
}

sub cfg_file {
	my ($self) = @_;

	return $self->{'config_path'};
}

sub param {
	my ($self, $key, $val) = @_;

	if (defined $val) {
		return $self->_set($key, $val);
	} else {
		my $val = $self->_get($key);
		$val = Core::Util::trim($val) if defined $val;
		my $home_dir = $self->get_home_dir();
		$val =~ s/{home}/$home_dir/gi if (defined $val);
		return $val;
	}
}

sub params {
	my ($self, $key) = @_;

	my @params = ();
	push @params, $self->param($key) if $self->param($key);
	for (my $i = 1; 1; $i++) {
		my $param = $self->param("${key}_" . $i);
		last if !$param;
		push @params, $param;
	}

	return @params;
}

sub required_params {
	my ($self, $key) = @_;

	my @params = $self->params($key);
	my $display_key = ucfirst($key);
	@params or croak "$display_key is required in config file: " . 
		$self->cfg_file();

	return @params;
}

sub set_no_save {
	my ($self, $key, $val) = @_;
	$self->{'keys'}->{$key} = trim($val);
}

sub get_home_dir {
	my ($self) = @_;

	return $self->{'home_path'};
}

sub _get_config_path {
	my ($home_path, $config_name) = @_;

	die "No home_path given" if !$home_path;
	die "No configuration name given" if !$config_name;
	return File::Spec->catfile($home_path, 'conf', "${config_name}.cfg");
}

sub _set {
	my ($self, $key, $val) = @_;
	
	$self->{'keys'}->{$key} = $val;
	$self->_save;
	return $val;
}

sub _get {
	my ($self, $key, $val) = @_;

	return $self->{'keys'}->{$key};
}

sub _save {
	my ($self) = @_;
	my $config_path = $self->{'config_path'};

	# Read in current file
	my $file = new FileHandle "$config_path" or 
		die "Couldn't read config file from $config_path: $!";
	my @lines = <$file>;
	undef $file;

	# Write out config file
	$file = new FileHandle ">$config_path" or 
		die "Couldn't save config file to $config_path: $!";
	
	my $written_table = {};
	foreach my $line (@lines) {
		if ($line =~ /^#/ || $line =~ /^\s*$/) {
			print $file $line;
		}else {
			$line = _genchomp($line);
			my ($key, $val) = split /\s*=\s*/, $line, 2;
			next if (!defined $key);
			my $display_key = $key;	# Preserve case for printing
			$key = lc($key);
			print $file "$display_key=".$self->{'keys'}->{$key}."\n";
			$written_table->{$key} = 1;
		}
	}
		
	print $file "\n\n# Additional parameters\n\n" 
		if keys %$written_table < keys %{$self->{'keys'}};
	
	my @keys = sort keys %{$self->{'keys'}};
	foreach my $key (@keys) {
		next if ($written_table->{$key});
		my $display_key = ucfirst($key);
		print $file "$display_key=".$self->{'keys'}->{$key}."\n";
	}

	undef $file;
}

sub _genchomp {
	my ($txt) = @_;

	$txt =~ s/[\r\n]+$//;
	return $txt;
}

1;