# $Id: DbUtil.pm 3017 2013-02-01 04:04:56Z ian $
package Core::DbUtil;

use strict;

use Carp;
use DBI();
use DBI::Const::GetInfoType;

use Date::Pcalc qw( Date_to_Text_Long );

sub connect {
	my ( $cfg, $dbname ) = @_;

	my $db_prefix = $dbname ? "${dbname}_" : "";
	my $db_type   = $cfg->required_param( $db_prefix . 'dbtype' );
	my $db        = $cfg->required_param( $db_prefix . 'dbdatabase' );
	my $user      = $cfg->param( $db_prefix . 'dbuser' );
	my $pass      = $cfg->param( $db_prefix . 'dbpass' );

	# Test to see if we're running under ActiveState's Perlapp (compiled code)
	my $running_under_perlapp = eval "length(\$PerlApp::VERSION) eq '0' ? '0': '1'";

	my $dsn;

	# SQLServer
	if ( $db_type =~ /sqlserver/i ) {
		my $db_name = $cfg->required_param( $db_prefix . 'dbname' );
		$dsn = "DBI:ODBC:$db_name";
	}

	# MySQL
	elsif ( $db_type =~ /mysql/i ) {
		my $host        = $cfg->required_param( $db_prefix . 'dbhost' );
		my $got_mysqlpp = 0;

		# Just use pure perl driver if running under perlapp.
		if ( $running_under_perlapp eq '1' ) {
			eval { require DBD::mysql };
			unless ($@) {
				$got_mysqlpp = 0;
				DBD::mysql->import();

				#			print "Using mysql\n";
			} else {
				eval { require DBD::mysqlPP };
				unless ($@) {
					$got_mysqlpp = 1;
					DBD::mysqlPP->import();

					#				print "Using mysqlPP\n";
				} else {
					die
"Could not find a working mysql database driver. Tried compiled code and native.\nDetails: \n$@\n";
				}
			}
		}

		# Try native code driver first then fall back on pure perl driver (should not reach that part if
		# running as per FF install instructions.
		else {
			eval { require DBD::mysql };
			unless ($@) {
				$got_mysqlpp = 0;
				DBD::mysql->import();

				#			print "Using mysql\n";
			} else {
				eval { require DBD::mysqlPP };
				unless ($@) {
					$got_mysqlpp = 1;
					DBD::mysqlPP->import();

					#				print "Using mysqlPP\n";
				} else {
					die
"Could not find a working mysql database driver. Tried compiled code and native.\nDetails: \n$@\n";
				}
			}
		}
		my $module_str = ( $got_mysqlpp eq '1' ) ? 'DBI:mysqlPP' : 'DBI:mysql';

		$dsn = "DBI:mysql:database=$db;host=$host";
	}

	# PostgreSQL
	elsif ( $db_type =~ /postgresql/i ) {
		my $got_pgpp = 0;

		# Just use pure perl driver if running under perlapp.
		if ( $running_under_perlapp eq '1' ) {
			eval { require DBD::Pg };
			unless ($@) {
				$got_pgpp = 0;
				DBD::Pg->import();
			} else {
				eval { require DBD::PgPP };
				unless ($@) {
					$got_pgpp = 1;
					DBD::PgPP->import();
				} else {
					die
"Could not find a working PostgreSQL database driver. Tried compiled code and native.\nDetails: \n$@\n";
				}
			}
		}

		# Try native code driver first then fall back on pure perl driver (should not reach that part if
		# running as per FF install instructions.
		else {
			eval { require DBD::Pg };
			unless ($@) {
				$got_pgpp = 0;
				DBD::Pg->import();
			} else {
				eval { require DBD::PgPP };
				unless ($@) {
					$got_pgpp = 1;
					DBD::PgPP->import();
				} else {
					die
"Could not find a working PostgreSQL database driver. Tried compiled code and native.\nDetails: \n$@\n";
				}
			}
		}
		my $module_str = ( $got_pgpp eq '1' ) ? 'DBI:PgPP' : 'DBI:Pg';

		my $host = $cfg->required_param( $db_prefix . 'dbhost' );
		$dsn = "${module_str}:dbname=$db;host=$host";

		#		print "DSN: $dsn\n";
	}

	# NOTE: PostgreSQL seems to behave the same way as MySQL here
	#                     {'RaiseError' => 1, 'PrintError' => 0, odbc_cursortype => 2, MARS_Connection => 'YES' });
	my $dbh = undef;
	eval {
		if ( $db_type =~ /sqlserver/i )
		{
			my $h = { 'RaiseError' => 1, 'PrintError' => 0, odbc_cursortype => 2 };

			my $odbc_default_bind_type = $cfg->param('db_odbc_default_bind_type');
			if ( defined $odbc_default_bind_type ) {
				$h->{odbc_default_bind_type} = $odbc_default_bind_type;
			}
			$dbh = DBI->connect( $dsn, $user, $pass, $h );

		   #                     {'RaiseError' => 1, 'PrintError' => 0, odbc_cursortype => 2, odbc_force_rebind => 1 });
			$dbh->do("USE $db");
			$dbh->{LongReadLen} = 16777216;
		} else {
			$dbh = DBI->connect( $dsn, $user, $pass, { 'RaiseError' => 1, 'PrintError' => 0 } );
		}
	};
	if ( defined $@ && ( !$@ eq '' ) ) {
		Croak $@;
	}

	return $dbh;
}

# Ascertain the database type (e.g. sqlserver, mysql, postgresql)
sub db_type {
	my ($dbh) = @_;

	$dbh or croak("No valid database handle was given");
	return $dbh->get_info( $GetInfoType{SQL_DBMS_NAME} )
	  || $dbh->{Driver}->{Name};
}

# Test to see if the current database connection is to a MySQL database
sub mysql {
	my ($dbh) = @_;

	my $db_type = db_type($dbh);
	return $db_type =~ /mysql/i;
}

# Test to see if the current database is PostgreSQL
# Function tested - 2007.01.11 - imarsman
sub postgresql {
	my ($dbh) = @_;

	my $db_type = db_type($dbh);
	return $db_type =~ /PostgreSQL/i;
}

sub sqlserver {
	my ($dbh) = @_;

	my $db_type = db_type($dbh);

	return $db_type =~ /sql ?server/i;
}

sub retrieve_max_id {
    my ($dbh, $table) = @_;

    my $id;
    # NOTES:
    # This needs to be called ONLY directly after an insert. It should be safe for PostgreSQL and MySQL
    
    # PostgreSQL: http://search.cpan.org/dist/DBD-Pg/Pg.pm#last_insert_id
    # PostgreSQL: also: http://archives.postgresql.org/pgsql-general/2008-08/msg01058.php
    # perhaps for SQL Server:
    # http://coding.derkeiler.com/Archive/Perl/perl.dbi.users/2005-04/msg00387.html
    # SELECT @@identity AS id
    # http://www.sqlteam.com/article/alternatives-to-identity-in-sql-server-2000
    # SELECT SCOPE_IDENTITY() is probably what we want
    # http://social.msdn.microsoft.com/Forums/en-US/sqldataaccess/thread/a8ff34e6-5bb9-494e-a790-2a50cc53abfd/
    # Does not look good
    if (mysql($dbh) || postgresql($dbh)) {
        # http://cipri4ph.wordpress.com/2008/04/05/perl-dbi-last-insert-id/
        $id = $dbh->last_insert_id(undef, undef, $table, "id");
    }
    else {
        my $stmt = $dbh->prepare("SELECT max(id) as id FROM ".$table);
        $stmt->execute();
        my $hashref = $stmt->fetchrow_hashref();
        $stmt->finish;
        $id = $hashref->{id};
    }
    return $id
}


sub get_next_id {
	my ( $dbh, $key_id ) = @_;

	my $id;
	my $is_mysql = mysql($dbh);
	$dbh->do(q{ LOCK TABLES unique_key WRITE }) if $is_mysql;
	eval {
		my $id_query        = q{ SELECT next_key FROM unique_key WHERE key_id=? };
		my $id_update_query = q{ UPDATE unique_key SET next_key=next_key+1 
			WHERE key_id=? };
		my $id_sth        = $dbh->prepare($id_query);
		my $id_update_sth = $dbh->prepare($id_update_query);
		$id_sth->execute($key_id);
		my @id_row = $id_sth->fetchrow_array();
		$id_sth->finish();
		if ( !@id_row ) {    # No row exists - create one
			my $new_row_query = q{ INSERT INTO unique_key (key_id, next_key) 
				VALUES (?, ?) };
			my $new_row_sth = $dbh->prepare($new_row_query);
			$new_row_sth->execute( $key_id, 1 );
			$id = 1;
		} else {
			$id = $id_row[0];
		}
		$id_update_sth->execute($key_id);
	};
	my $error = $@;
	$dbh->do(q{ UNLOCK TABLES }) if $is_mysql;
	die $error if $error;

	return $id;
}

sub limit {
	my ( $dbh, $limit, $sql ) = @_;

	return $sql if !$limit;

	# PostgreSQL seems to have the same limit clause syntax as MySQL
	if ( mysql($dbh) || postgresql($dbh) ) {
		return $sql . qq( LIMIT $limit );
	} elsif ( sqlserver($dbh) ) {
		$sql =~ s/^\s*SELECT\s+(DISTINCT)?/SELECT $1 TOP $limit /
		  or die "Invalid SELECT statement: $sql";
		return $sql;
	} else {
		die "Database type not supported";
	}
}

sub date2str {
	my ($date) = @_;

	$date =~ /^(\d\d\d\d)-(\d\d)-(\d\d) (\d\d):(\d\d):(\d\d)$/ or die "No date: $date!";
	my ( $year, $month, $day, $hour, $minute, $second ) = ( $1, $2, $3, $4, $5, $6 );

	return Date_to_Text_Long( $year, $month, $day ) . " $hour:$minute";
}

sub parse_date {
	my ($date) = @_;

	$date =~ /^(\d\d\d\d)-(\d\d)-(\d\d) (\d\d):(\d\d):(\d\d)$/ or die "No date: $date!";
	my ( $year, $month, $day, $hour, $minute, $second ) = ( $1, $2, $3, $4, $5, $6 );

	return ( $year, $month, $day, $hour, $minute, $second );
}

1;
