# $Id: Log.pm 2973 2012-07-18 04:23:46Z ian $

package Core::Log;

use Carp;
use FileHandle;
use File::Path;
use File::Basename;
use Fcntl qw(:DEFAULT :flock);

use Core::Util;

use strict;

# Takes log_path, console
sub new {
    my ($class, $hashref) = @_;

	my $self = {};
	
	bless $self, $class;
	initialize($self, $hashref);
	return $self;
}

sub initialize {
	my ($self, $hashref) = @_;

	my $log_path = $hashref->{log_path};
	my $instance = $hashref->{instance};
	$log_path or croak "No log path supplied";
	my ($name, $path, $suffix) = fileparse($log_path,
		qr{\.[^\.]+$});
	$name = "$name-$instance";

	$self->{name} = $name;
	$self->{path} = $path;
	$self->{suffix} = $suffix;
	$self->{cur_logpath} = join("", $path, "current-$name", $suffix);
	
	my $thresh = $hashref->{level_threshold};
	$self->init_levels($thresh);
	
	$self->{'console'} = $hashref->{'console'};
}

sub init_levels {
	my ($self, $thresh) = @_;
	
	my @level_list = qw( DBG ADV HI MED LOW ERR );
	my %levels = map { $_ => 1 } @level_list;
	$thresh = trim($thresh);
	$thresh = "MED" if !$thresh;

	my @active_list;
	if ($thresh eq 'DBG') {
		@active_list = qw( DBG ADV HI MED LOW ERR );
	} elsif ($thresh eq 'ADV') {
		@active_list = qw( ADV HI MED LOW ERR );
	} elsif ($thresh eq 'HI') {
		@active_list = qw( HI MED LOW ERR );
	} elsif ($thresh eq 'MED') {
		@active_list = qw( MED LOW ERR );
	} elsif ($thresh eq 'LOW') {
		@active_list = qw( LOW ERR );
	} else {
		my $joined_threshes = join(", ", qw( DBG ADV HI MED LOW ));
		die "Unsupported logging threshold $thresh, must be " . 
			"one of: $joined_threshes";
	}

	$self->{levels} = \%levels;
	$self->{active_levels} = { map { $_ => 1 } @active_list }; 
}

sub _logpath {
	my ($self) = @_;

	my $name = $self->{name};
	my $path = $self->{path};
	my $suffix = $self->{suffix};

	my @time_arr = localtime;
	my ($year, $mon, $day) = ($time_arr[5] + 1900, $time_arr[4]	+ 1, $time_arr[3]);
	my $dated_name = sprintf("%s-%04d-%02d-%02d", $name, $year, $mon, $day);
	mkpath($path);
	return join("", $path, $dated_name, $suffix);
}

sub _cur_logpath {
	my ($self) = @_;

	my $path = $self->{path};

	mkpath($path);
	return $self->{cur_logpath};
}

sub log {
	my ($self, $msg) = @_;

	chomp($msg);

	$msg =~ /^\[(\w+)\s*\]/;
	my $level = $1;	
	$self->log("[ADV] unrecognized level [$level] in msg: $msg") if 
		!$self->{levels}->{$level};

	# Ignore log row if this level isn't active
	return if !$self->{active_levels}->{$level};
	
	my $date = localtime;
	my $print_msg = "$date - $msg\n";

	# Write to console
	print $print_msg if $self->{'console'};

	# Write to dated log file
	my $log_path = $self->_logpath();
	my $fh = new FileHandle ">>$log_path" or 
		die "Can't open $log_path for writing";
	flock($fh, LOCK_EX);
	print $fh $print_msg if $fh;
	flock($fh, LOCK_UN);
	$fh->close();

	# Write to current log file
	my $cur_log_path = $self->_cur_logpath();
	my $write_mode = -s $cur_log_path > 100_000 ? '>' : '>>';
	# Changed logfile write behaviour to NOT fail if a write could not occur
	# but rather to print out a warnign to stdout.
	eval {
		my $cfh = new FileHandle "${write_mode}$cur_log_path";
		# or 
			#die "Can't open $cur_log_path for writing";
		flock($cfh, LOCK_EX);
		print $cfh $print_msg if $cfh;
		flock($cfh, LOCK_UN);
		$cfh->close();
	};
	if ($@) {
		$print_msg =~ s/\n//g;
		print "Can't write message \"$print_msg\" to $cur_log_path\n";
		#return;
	}
}

sub console { return $_[0]->{console}; }

# Logs exceptions but doesn't die
sub trap {
	my ($self, $sub, $doc_id, @params) = @_;

	my $result = undef;
	eval { $result = &$sub(@params); };
	$self->log("[ADV] For $doc_id: $@") if $@;

	return $result;
}

1;

