# $Id: SimpleHtmlFormatter.pm 2125 2008-07-23 19:27:14Z ian $
package Core::SimpleHtmlFormatter;

use strict;

use HTML::TokeParser;
use Text::Wrap;

use Core::AsciiTable;

sub new {
	my ($class, $params) = @_;

	my $self = {};
	bless $self, $class;
	$self->_initialize($params);
	return $self;
}

sub _initialize {
	my ($self, $params) = @_;

	$self->{'screen_width'} = 80;
	$self->{'output'} = '';
	$self->{'nl'} = $params->{'nl'} or die "requires 'nl' (newline) param";
	$self->{'indent'} = " " x $params->{'indent'} or die "requires indent param";
	$self->{'endp'} = $params->{'endp'} or 
		die "requires 'endp' (end paragraph) param";
	$self->{'col_sep'} = $params->{'col_sep'} || ' ';
	$Text::Wrap::columns = $self->{'screen_width'};
}

sub transform {
	my ($self, $html) = @_;

	$html or die "No HTML given";
	my $parser = HTML::TokeParser->new(\$html);
	while (my $token = $parser->get_token) {
		my @arr = @$token;
		my $type = shift @arr;
		if ($type eq 'S') {	# Start tag
			my ($tag, $attr) = @arr;
			$self->start_tag($tag, $attr);
		} elsif ($type eq 'T') {
			my ($cur_text) = @arr;
			$self->text($cur_text);
		} elsif ($type eq 'E') {
			my ($tag) = @arr;
			$self->end_tag($tag);
		}
	} 

	return $self->{'output'};
}

sub start_tag {
	my ($self, $tag, $attr) = @_;

	if ($tag =~ /^table$/i) {
		$self->{'table'} = new Core::AsciiTable($self->{'col_sep'});
	} elsif ($tag =~ /^tr$/i) {
		$self->{'table'}->row_start();
	} elsif ($tag =~ /^td$/i || $tag =~ /^th$/i) {
		$self->{'table'}->cell_start($attr->{'align'}, $attr->{'colspan'});
	} elsif ($tag =~ /^p|pre|h1|h2|h3|div|span$/i) {
		$self->{'text'} = "";
	} elsif ($tag =~ /^br$/i) {
		$self->{'text'} .= $self->{'nl'};
	}
}

sub end_tag {
	my ($self, $tag) = @_;

	my $nl = $self->{'nl'};
	if ($tag =~ /^table$/i) {
		my @lines = $self->{'table'}->table_lines();
		$self->{'table'} = undef;
		foreach my $line (@lines) { 
			$self->cat(" $line$nl"); 
		}
	} elsif ($tag =~ /^tr$/i) {
		$self->{'table'}->row_end();
	} elsif ($tag =~ /^td$/i || $tag =~ /^th$/i) {
		$self->{'table'}->cell_end();
	} elsif ($tag =~ /^p$/i) {
		$self->cat(wrap($self->{'indent'}, '', $self->pull_text()) . 
			$self->{'endp'});
	} elsif ($tag =~ /^pre$/i) {
		$self->cat(wrap('', '', $self->pull_text_pre()) . 
			$self->{'endp'});
	} elsif ($tag =~ /^h1|h2|h3$/i) {
		$self->cat(wrap('', '', $self->pull_text()) . $nl);
	} elsif ($tag =~ /^div$/i) {
		my $text = $self->pull_text();
		$self->cat("$text$nl");
	} elsif ($tag =~ /^span$/i) {
		my $text = $self->pull_text();
		$self->cat($text);
	}
}

sub text {
	my ($self, $cur_text) = @_;

	if ($self->{'table'}) { $self->{'table'}->text($cur_text); }
	else { $self->{'text'} .= $cur_text; }
}

sub pull_text {
	my ($self) = @_;
	my $text = Core::AsciiTable::purge($self->{'text'});
	$self->{'text'} = "";
	return $text;
}

sub pull_text_pre {
	my ($self) = @_;
	my $text = Core::AsciiTable::purge_pre($self->{'text'});
	$self->{'text'} = "";
	return $text;
}

sub cat {
	my ($self, $text) = @_;
	$self->{'output'} .= $text;
}

1;