# $RCSfile$; $Date: 2008-01-10 22:22:59 -0800 (Thu, 10 Jan 2008) $; $Revision: 1672 $; $Name$
package Feedfetcher::Plugins::AP;

use strict;
use Feedfetcher::Plugins::Retriever;
our @ISA = ("Feedfetcher::Plugins::Retriever");

use Core::LineParser;
use Core::XmlUtil;
use Core::ActiveFixtures;
use Feedfetcher::Plugins::Local;
use Feedfetcher::Plugins::LinescoreGenerator;

use Carp;
use News::NNTPClient;
use MIME::Entity;
use MIME::Parser;
use XML::LibXML;
use URI;
use DateTime;
use DateTime::Format::Mail;
use File::Path;
use FileHandle;

sub name { return 'AP feed'; }
sub code { return 'ap'; }

sub initialize {
	my ($self) = @_;

	my $ls_file = $self->cfg->required_param('ap_linescore_fixture_table');
	my $ls_update_file = 
		$self->cfg->required_param('ap_linescore_update_fixture_table');
	$self->{'ls_gen'} = 
		new Feedfetcher::Plugins::LinescoreGenerator('ap.org',
		'ap_linescore_fixture_table', 'ap_linescore_update_fixture_table',
		$self->cfg, $self->get_log);
	$self->load_last_articles;
	$self->load_tables;
}

sub ls_gen { my ($self)= @_; return $self->{'ls_gen'}; }

sub load_tables {
	my ($self) = @_;

	$self->{'map_league_sport'} = 
		Core::LineParser::load_table($self->cfg, 'ap_map_league_sport');
	$self->{'map_season'} = 
		Core::LineParser::load_table_list($self->cfg, 'ap_map_season');
	$self->{'map_leagues'} =
		Core::LineParser::load_table($self->cfg, 'ap_map_leagues');
	$self->{'insert_teams'} = 
		Core::ActiveFixtures::load($self->cfg, 'ap-insert-teams');
}

sub nntp_connect {
	my ($self) = @_;

	my $cfg = $self->cfg;
	my $server = $cfg->required_param('ap_server');
	my $user = $cfg->required_param('ap_user');
	my $pass = $cfg->required_param('ap_pass');
	my $c = new News::NNTPClient($server, "", 0);
	my $msg = $c->message;
	chomp $msg;
	my $code = $c->code;
	my $status = $msg ? "$msg ($code)" : $code;
	$c->ok or die "Can't connect to AP server $server: $status";
	$c->authinfo($user, $pass) or 
		die "Can't authenticate: $!";

	return $c;
}

sub get_list {
	my ($self) = @_;

	my @files = ();
	my $date_map = {};
	my @groups = $self->get_groups;
	$self->{'server'} = $self->retry(\&nntp_connect);
	foreach my $group (@groups) {
		my $first_last = $self->retry(
			sub { 
				my ($self, $group) = @_; 
				my $first_last = $self->{'server'}->group($group) or die "$group: $!"; 
				return $first_last;
			}, $group);
		my ($first, $last) = split(/-/, $first_last);
		$first = $self->get_first($first, $last, $group);
		for (my $i = $first; $i <= $last; $i++) {
			my $art_id = $self->retry(
				sub { 
					my ($self, $art_number) = @_; 
					return $self->{'server'}->stat($art_number) or die $!; 
				}, $i) or next;
			my $header_ref = $self->retry(
			sub {
				my ($self, $art_id) = @_;
				my $lines_ref = $self->{'server'}->head($art_id); 
				return $lines_ref;
			}, $art_id) or next;
			my $date = get_date($header_ref);
			push @files, $art_id;
			$date_map->{$art_id} = $date;
			$self->map_article($art_id, $group, $i);
			return undef if $self->check_abort;
		}
	}

	my @sorted_files = sort { $date_map->{$a} <=> $date_map->{$b} } @files;

	return @sorted_files;
}

sub get_date {
	my ($header_ref) = @_;

	foreach my $line (@$header_ref) {
		chomp $line;

		if ($line =~ /^Date:\s+(.+)$/) {
			my $news_date = $1;
			my $p = DateTime::Format::Mail->new->loose;
			my $dt = $p->parse_datetime(
				$news_date
			);
			return $dt->strftime("%Y%m%d%H%M%S");
		}
	}

	die "No date in news header";
}

sub retrieve_doc {
	my ($self, $art_id) = @_;

	$self->{'last_article'} = $self->map_article($art_id);
	my $lines_ref = $self->retry(
		sub {
			my ($self, $art_id) = @_;
			my $lines_ref = $self->{'server'}->article($art_id); 
			return $lines_ref;
		}, $art_id) or next;
	my $article = join("", @$lines_ref);

	my $parser = new MIME::Parser;
	$parser->output_to_core(1);	# Don't print out to disk
	my $entity = $parser->parse_data($article);
	my $xml_entity  = $entity->parts(1);
	my $body = $xml_entity->bodyhandle;
	my $xml = $body->as_string;

	return $xml
}

sub post_retrieve {
	my ($self) = @_;
	my ($group, $last) = @{$self->{'last_article'}};
	$self->set_last($group, $last);
	$self->store_last_articles;
}

sub cleanup {
	my ($self) = @_;

	$self->store_last_articles;
	undef $self->{'server'};
}

sub map_article {
	my ($self, $art_id, $group, $art_num) = @_;

	$self->{'map_article'} = {} if !defined $self->{'map_article'};
	if (defined $group) {
		$self->{'map_article'}->{$art_id} = [ $group, $art_num ];
	} else {
		return $self->{'map_article'}->{$art_id};
	}
}

sub get_groups {
	my ($self) = @_;

	my $groups_file = $self->cfg->required_param('ap_groups_file');
	my @groups = Core::LineParser::load($groups_file);
	return @groups;
}

sub get_first {
	my ($self, $first, $last, $group) = @_;

	my $new_first = undef;

	my $force_last = $self->arg('last');
	$new_first = $last - ($force_last - 1) if $force_last;

	if (!defined $new_first) {
		my $last_loaded = $self->{'last_article_table'}->{$group};
		$new_first = $last_loaded + 1 if defined $last_loaded;
	} 

	if (!defined $new_first) {
		my $cfg_last = $self->cfg->required_param('ap_last');
		return $last - ($cfg_last - 1);
	}

	return $new_first > $first ? $new_first : $first;
}

sub load_last_articles {
	my ($self) = @_;

	$self->{'last_article_table'} = {};
	my $home_path = Core::Util::get_home_path();
	my $last_article_file = "$home_path/ap_last_article";
	return if !-f $last_article_file;

	my @lines = Core::LineParser::load($last_article_file);
	foreach my $line (@lines) {
		my ($group, $last_art) = split(/:/, $line);
		die "Corrupt ap_last_article file: $line" if 
			!defined $group || !defined $last_art;
		$self->{'last_article_table'}->{$group} = $last_art;
	}
}

sub store_last_articles {
	my ($self) = @_;

	return if !$self->{'last_article_table'};
	my $home_path = Core::Util::get_home_path();
	my $last_article_file = "$home_path/ap_last_article";
	my $fh = new FileHandle ">$last_article_file" or 
		die "Can't write $last_article_file: $!";
	my %last_article_table = %{$self->{'last_article_table'}};
	foreach my $group (keys %last_article_table) {
		my $last_article = $last_article_table{$group};
		print $fh "$group:$last_article\n";
	}
}

sub set_last {
	my ($self, $group, $last) = @_;
	
	$self->{'last_article_table'}->{$group} = $last;
}

sub filter {
	my ($self, $xml) = @_;

	my $doc = $self->parser->parse_string($xml);	

	# Modify doc-id
	my $metanode = solonode($doc, "/sports-content/sports-metadata");
	my $doc_id = required_attribute($metanode, "doc-id");
	$doc_id = "ap.$doc_id";
	$metanode->setAttribute('doc-id', $doc_id);

	# Fix date (add EDT timezone)
	my $date = $metanode->getAttribute("date-time");
	$date = $self->fix_date($date);
	$metanode->setAttribute('date-time', $date);

	# Fix event dates
	my @event_nodes = $doc->findnodes("//sports-event");
	my $event_date = undef;
	foreach my $event_node (@event_nodes) {
		my $event_metanode = solonode($event_node, "event-metadata");
		my $date = $event_metanode->getAttribute("start-date-time");
		if (defined $date) {
			$date = $self->fix_date($date);
			$event_metanode->setAttribute('start-date-time', $date);
			$event_date = $date if !$event_date;
		}
	}

	# Add publisher-key
	my $content_codes_node = solonode($metanode, "sports-content-codes");
	my $content_code = 
		$self->create_content_code($doc, 'publisher', 'ap.org', '');
	$content_codes_node->appendChild($content_code);

	my $fixture_key = required_attribute($metanode, 'fixture-key');
	my $league_key = $self->map_league($content_codes_node);
	my $season_key = 
		$self->map_season($league_key, $event_date, $content_codes_node, $doc) if
		$event_date;
	$self->map_event_key($league_key, $season_key, $doc) if 
		$event_date;
	die "Cannot establish season" if !$event_date && 
		$doc->findnodes('//event-metadata[@event-key]');
	$self->insert_teams($league_key, $fixture_key, $content_codes_node, $doc);

	# Fix sports
	$self->fix_sports($league_key, $content_codes_node);

	# Fix Misc
	$self->fix_misc($league_key, $fixture_key, $doc);

	my $ls_doc = $self->ls_gen->process($league_key, $doc);
	
	my @docs = ($doc->toString);
	push @docs, $ls_doc->toString if $ls_doc;
	return @docs;
}

sub fix_date {
	my ($self, $datetime) = @_;

	confess "datetime not specified" if !$datetime;
	$datetime =~ /(\d\d\d\d)-?(\d\d)-?(\d\d)T?(\d\d):?(\d\d):?(\d\d)/ or
		die "Invalid datetime: $datetime";
	my ($year, $mon, $day, $hour, $min, $sec) = ($1, $2, $3, $4, $5, $6);
	my $dt = DateTime->new( year   => $year,
                       month  => $mon,
                       day    => $day,
                       hour   => $hour,
                       minute => $min,
                       second => $sec,
                       time_zone => 'America/New_York',
                     );

	return $dt->strftime("%Y%m%dT%H%M%S%z");
}

sub map_league {
	my ($self, $content_codes_node) = @_;

	my $league = required_attribute_xpath($content_codes_node, 
		'sports-content-code[@code-type="league"]/@code-key'); 
	my $sport = attribute_xpath($content_codes_node, 
		'sports-content-code[@code-type="sport"]/@code-key');
	my $key = join '/', grep defined, $league, $sport;
	my $good_league = $self->{map_leagues}->{$key} ||
		$self->{map_leagues}->{$league} or 
		die "$key not configured in mapping file";

	my $league_node = solonode($content_codes_node, 
		'sports-content-code[@code-type="league"]');
	$league_node->setAttribute('code-key', $good_league);
	return $good_league;
}

sub fix_sports {
	my ($self, $league, $content_codes_node) = @_;

	my $sport_node = solonode_optional($content_codes_node, 
		'sports-content-code[@code-type="sport"]');

	return if !$sport_node;

	my ($sport_key, $sport_name) = $self->map_league_sport($league);
	$sport_node->setAttribute('code-key', $sport_key);
	$sport_node->setAttribute('code-name', $sport_name);
}

sub map_league_sport {
	my ($self, $league) = @_;

	my $rec = $self->{'map_league_sport'}->{$league} or 
		die "$league league not supported [map_league_sport]";
	my @fields = split(/,/, $rec);
	die "Couldn't parse map_league_sport file" if @fields != 2;

	return ($fields[0], $fields[1]);
}

sub map_season {
	my ($self, $league, $datetime, $content_code_nodes, $doc) = @_;

	$datetime =~ /(\d+)T/ or die "Invalid datetime: $datetime";
	my $date = $1;
	my $listrec = $self->{'map_season'}->{$league} or 
		die "$league league not supported [map_season]";
	foreach my $rec (@$listrec) {
		my ($start, $end, $season_key) = split(/,/, $rec);
		if ($date >= $start && $date <= $end) {
		my $content_code = 
			$self->create_content_code($doc, 'season', $season_key);
			$content_code_nodes->appendChild($content_code);
			return $season_key;
		}
	}

	die "Event not within valid season";
}

sub fix_misc {
	my ($self, $league_key, $fixture_key, $doc) = @_;

	# Fix baseball linescore
	my ($sport_key, $sport_name) = $self->map_league_sport($league_key);
	if ($sport_key == 15007000 && 
		$fixture_key eq 'linescore') {

		my $meta_baseball = solonode($doc, 
			'sports-content/sports-event/event-metadata/event-metadata-baseball');
		$meta_baseball->setAttribute('outs', '3');
		$meta_baseball->setAttribute('runnner-on-first', 'false');
		$meta_baseball->setAttribute('runnner-on-second', 'false');
		$meta_baseball->setAttribute('runnner-on-third', 'false');
	}
}

sub insert_teams {
	my ($self, $league_key, $fixture_key, $content_code_nodes, $doc) = @_;

	my $table = $self->{'insert_teams'};
	return if !Core::ActiveFixtures::doc_lookup($table, $doc, $self->cfg);
	
	my @team_nodes = $doc->findnodes('//team-metadata[@team-key]');
	foreach my $team_node (@team_nodes) {
		my $team_key = $team_node->getAttribute('team-key');
		my $name_node = solonode_optional($team_node, 'name');
		my $name = undef;
		if ($name_node) {
			my $first = $name_node->getAttribute('first');
			my $last = $name_node->getAttribute('last');
			if ($first && $last) {
				$name = "$first $last";
			} elsif ($first) {
				$name = $first;
			} elsif ($last) {
				$name = $last;
			}
		}
		my $content_code = 
			$self->create_content_code($doc, 'team', $team_key, $name);
		$content_code_nodes->appendChild($content_code);
	}
}

sub map_event_key {
	my ($self, $league_key, $season_key, $doc) = @_;

	my @nodes = $doc->findnodes('//event-metadata[@event-key]');
	foreach my $node (@nodes) {
		my $event_key = $node->getAttribute('event-key');
		$event_key = "$league_key-$season_key-$event_key";
		$node->setAttribute('event-key', $event_key);
	}
}

sub create_content_code {
	my ($self, $doc, $code_type, $code_key, $code_name) = @_;

	my $content_code = $doc->createElement('sports-content-code');
	$content_code->setAttribute('code-type', $code_type);
	$content_code->setAttribute('code-key', $code_key);
	$content_code->setAttribute('code-name', $code_name) 
		if $code_name;  

	return $content_code;
}

1;