# $RCSfile$; $Date: 2008-01-10 22:23:17 -0800 (Thu, 10 Jan 2008) $; $Revision: 1673 $; $Name$
package Feedfetcher::Plugins::LinescoreGenerator;

use strict;

use FileHandle;
use DateTime;
use Carp;
use XML::LibXML;
use File::Path;

use Core::XmlUtil;

sub new {
    my ($class, $publisher_key, 
		$fixture_table, $update_table, $cfg, $log) = @_;

	my $self = {};
	bless $self, $class;
	$self->{'cfg'} = $cfg or croak "No configuration given";
	$self->{'log'} = $log or croak "No logger given";
	$self->initialize($publisher_key, $fixture_table, $update_table);
	return $self;
}

sub cfg { my ($self) = @_; return $self->{'cfg'}; }
sub log { my ($self, $msg) = @_; $self->{'log'}->log($msg); }
sub pub_key { my ($self) = @_; return $self->{'pub_key'}; }

sub initialize {
	my ($self, $publisher_key, $fixture_table, $update_table) = @_;

	my $cfg = $self->cfg;
	$self->{'period_table'} = Core::LineParser::load_table($cfg, 
		'linescore_period_table');
	$self->{'fixture_table'} = Core::ActiveFixtures::load($cfg, $fixture_table);
	$self->{'update_table'} = Core::ActiveFixtures::load($cfg, $update_table);
	$self->{'pub_key'} = $publisher_key;
	$self->{'parser'} = new XML::LibXML;
}

sub process {
	my ($self, $league_key, $doc) = @_;

	my $fixture_key = required_attribute_xpath($doc,
		'/sports-content/sports-metadata/@fixture-key');
	
	# Check if it's linescore
	my $is_linescore = Core::ActiveFixtures::lookup($self->{'fixture_table'},
		'league', $league_key, $fixture_key);
	if ($is_linescore) {
		$self->proc_linescore($doc);
	} 

	# Check if it's an update doc
	my $result = undef;
	if ($self->is_update($league_key, $fixture_key)) {
		$result = $self->generate_linescore($league_key, $doc);
	}

	if ($is_linescore) {
		#Check for end of game
		my $event_status = required_attribute_xpath($doc,
			'/sports-content/sports-event/event-metadata/@event-status');
		$self->cleanup_linescore($doc) if $event_status eq 'post-event';
	}

	return $result;
}

sub is_update {
	my ($self, $league_key, $fixture_key) = @_;

	return Core::ActiveFixtures::lookup($self->{'update_table'}, 
		'league', $league_key, $fixture_key);
}

# Should be called when linescore doc comes in
sub proc_linescore {
	my ($self, $doc) = @_; 

	my $event_key = $self->get_event_key($doc);
	my $path = $self->linescore_file($event_key);
	$doc->toFile($path);
}

# Should be called when game is over, i.e. boxscore
sub cleanup_linescore {
	my ($self, $doc) = @_; 

	my $event_key = $self->get_event_key($doc);
	my $path = $self->linescore_file($event_key);
	unlink $path if -f $path;
}

# Should be called when an update doc comes in
sub generate_linescore {
	my ($self, $league, $update_doc) = @_;

	my $event_key = $self->get_event_key($update_doc);
	my $doc = $self->get_linescore($event_key);
	if (!$doc) {
		my $u_event = solonode($update_doc, 
			qq{/sports-content/sports-event[event-metadata/\@event-key="$event_key"]});
		my $u_period = $self->period($league, $u_event); 
		if ($u_period == 1) {
			$self->new_linescore($update_doc) if $u_period; 
			$doc = $self->get_linescore($event_key);
		}
	}
	return if !$doc;
	my $orig_doc_id = $self->update_content_meta($doc, $update_doc);
	my $update_doc_id = required_attribute_xpath($update_doc, 
		'/sports-content/sports-metadata/@doc-id');
	$self->insert_content_code($doc, $orig_doc_id, $update_doc_id);

	# Write with updated doc-id and content code for next run
	$self->proc_linescore($doc);
	
	$self->update_metadata($doc, $update_doc, $event_key);
	$self->update_subscore($league, $doc, $update_doc, $event_key);

	return $doc;
}

# This only works for first period
sub new_linescore {
	my ($self, $update_doc) = @_;

	$self->proc_linescore($update_doc);
}

sub update_subscore {
	my ($self, $league, $ls_doc, $update_doc, $event_key) = @_;

	my $ls_event = solonode($ls_doc, 
		qq{/sports-content/sports-event[event-metadata/\@event-key="$event_key"]});

	my $u_event = solonode($update_doc, 
		qq{/sports-content/sports-event[event-metadata/\@event-key="$event_key"]});
	my $u_period = $self->period($league, $u_event); 
	my $sport = required_attribute_xpath($update_doc,
		q{/sports-content/sports-metadata/sports-content-codes/} .
		q{sports-content-code[@code-type="sport"]/@code-key});

	# Get teams
	my @team_nodes = $u_event->findnodes('team');
	foreach my $team_node (@team_nodes) {
		my $meta = solonode($team_node, 'team-metadata');
		my $team_key = required_attribute($meta, 'team-key');
		my $stats_node = solonode($team_node, 'team-stats');
		my $score = required_attribute($stats_node, 'score');
		
		my $ls_team_node = solonode($ls_event, 
			qq{team[team-metadata/\@team-key="$team_key"]});
		my $ls_subscore_node = lastnode($ls_team_node, 'team-stats/sub-score');
		my $ls_stats_node = solonode($ls_team_node, 'team-stats');
		my $ls_period = $ls_subscore_node ? 
			required_attribute($ls_subscore_node, 'period-value'): undef; 

		# Get rid of spurious children
		my @children = $ls_stats_node->childNodes;
		foreach my $child (@children) {
			$ls_stats_node->removeChild($child) if 
				$child->nodeName ne 'sub-score';
		}

		# We must be in period (or half-inning) after last linescore update
		my $new_subscore = undef;
		if (!$ls_period) {
			$new_subscore = $score;
		} else {
			my $period_diff = $u_period - $ls_period;
			if ($sport == 15007000) {	# Baseball - special case
				my $u_half = required_attribute_xpath($u_event,
					q{event-metadata/event-metadata-baseball/@inning-half});
				my $ls_half = required_attribute_xpath($ls_event,
					q{event-metadata/event-metadata-baseball/@inning-half});
				my $alignment = required_attribute($meta, 'alignment');
				die "Bad inning-half in linescore doc: $ls_half" if
					$ls_half ne 'top' && $ls_half ne 'bottom';

				# We're in the top of the inning - so don't update home team
				next if ($alignment eq 'home' && $ls_half eq 'top');

				if ($ls_half eq 'bottom') {		# Bottom of inning
					next if $period_diff != 1;
				} else {						# Top of inning
					die "Bad inning-half in update doc: $u_half" if
						$u_half ne 'top' && $u_half ne 'bottom';
					next if ($u_half eq 'bottom' && $period_diff != 0) ||
						($u_half eq 'top' && $period_diff != 1);
				}
			} else {					# Regular case
				next if $u_period - $ls_period != 1; 
			}

			my $ls_score = required_attribute($ls_stats_node, 'score');
			$new_subscore = $score - $ls_score;
			next if $new_subscore < 0;
		}

		# Add new subscore node
		my $new_subscore_node = $ls_doc->createElement('sub-score');
		$new_subscore_node->setAttribute('period-value', $u_period);
		$new_subscore_node->setAttribute('score', $new_subscore);
		$ls_stats_node->appendChild($new_subscore_node);

		# Update score
		$ls_stats_node->setAttribute('score', $score);
	}
}

sub insert_content_code {
	my ($self, $doc, $orig_doc_id, $update_doc_id) = @_;

	my $content_code = solonode_optional($doc, 
		'/sports-content/sports-metadata' .
		'/sports-content-codes/sports-content-code[@code-type="derivative"]');
	if ($content_code) {
		my $orig_value = required_attribute($content_code, 'code-key');
		($orig_doc_id, undef) = split(/, /, $orig_value);
	} else {
		$content_code = $doc->createElement('sports-content-code');
	}

	my $value = "$orig_doc_id, $update_doc_id";
	$content_code->setAttribute('code-type', 'derivative');
	$content_code->setAttribute('code-key', $value);
	$content_code->setAttribute('code-source', 'xmlteam.com');

	my $content_codes = 
		solonode($doc, '/sports-content/sports-metadata/sports-content-codes');
	$content_codes->appendChild($content_code);
}

sub get_linescore {
	my ($self, $event_key) = @_;

	my $path = $self->linescore_file($event_key);
	return undef if !-f $path;
	my $doc = $self->{'parser'}->parse_file($path);
	return $doc;
}

sub update_metadata {
	my ($self, $ls_doc, $update_doc, $event_key) = @_;

	my $ls_event = solonode($ls_doc, 
		qq{sports-content/sports-event[event-metadata/\@event-key="$event_key"]});
	my $u_event = solonode($update_doc, 
		qq{sports-content/sports-event[event-metadata/\@event-key="$event_key"]});
	my $u_meta = solonode($u_event, 'event-metadata');
	my $ls_meta = solonode($ls_event, 'event-metadata');
	my $newnode = $u_meta->cloneNode( 1 );
	$ls_meta->replaceNode($newnode);
}

sub update_content_meta {
	my ($self, $doc, $update_doc) = @_;

	my $meta_node = solonode($doc, '/sports-content/sports-metadata');

	my $datetime = required_attribute_xpath($update_doc,
		'/sports-content/sports-metadata/@date-time');
	$datetime = $self->generate_datetime($datetime);	
	$meta_node->setAttribute('date-time', $datetime);
	$meta_node->setAttribute('document-class', 'event-summary');
	$meta_node->setAttribute('fixture-key', 'derived-linescore');
	$meta_node->setAttribute('fixture-name', 'derived linescore');
	$meta_node->setAttribute('fixture-source', 'xmlteam.com');

	my $doc_id = required_attribute($meta_node, 'doc-id');
	$meta_node->setAttribute('doc-id', $self->update_doc_id($doc_id));

	# sports-title
	my $title_text = solonode($meta_node, 'sports-title/text()');
	my $text = $title_text->data;
	$text .= ' [derived]' if $text !~ /\[derived\]/;
	$title_text->setData($text);

	# Return original doc-id
	return $doc_id;	
}

sub generate_datetime {
	my ($self, $date_time) = @_;

	$date_time =~ 
		/(\d\d\d\d)-?(\d\d)-?(\d\d)T?(\d\d):?(\d\d):?(\d\d)([+-]\d+)$/ or
		die "Invalid datetime: $date_time";
	my ($year, $mon, $day, $hour, $min, $sec, $tz) = 
		($1, $2, $3, $4, $5, $6, $7);
	my $dt = DateTime->new( year   => $year,
				   month  => $mon,
				   day    => $day,
				   hour   => $hour,
				   minute => $min,
				   second => $sec,
				   time_zone => $tz,
				 );
	$dt = $dt + DateTime::Duration->new( seconds => 1 ); 
	return $dt->strftime("%Y%m%dT%H%M%S%z");
}

sub update_doc_id {
	my ($self, $doc_id) = @_;

	if ($doc_id =~ /^dv\./) {
		$doc_id =~ /\.(\d+)$/;
		my $sn = $1 + 1;
		$doc_id =~ s/(.*)\.\d+$/$1.$sn/;
	} else {
		$doc_id =~ s/(.*)$/dv.$1.1/ or 
			die "Bad format for doc_id: $doc_id";
	} 

	return $doc_id;
}

sub linescore_file {
	my ($self, $event_key) = @_; 

	my $dir = $self->cfg->required_param("linescore_dir");
	mkpath($dir);
	my $event_id = $self->pub_key . "-" . $event_key; 
	my $path = "$dir/$event_id-linescore.xml";
	return $path;
}

sub period {
	my ($self, $league, $event_node) = @_;

	my $xpath = $self->{'period_table'}->{$league} or 
		die "$league league not supported [Period table]";
	my $period_attr = solonode($event_node, "event-metadata/$xpath");

	return $period_attr->getValue();
}

sub get_event_key {
	my ($self, $doc) = @_;

	my $query = '/sports-content/sports-event/event-metadata/@event-key';
	my $event_key = required_attribute_xpath($doc, $query);

	return $event_key;
}

1;