# $RCSfile$; $Date: 2016-04-19 13:33:45 -0400 (Tue, 19 Apr 2016) $; $Revision: 3109 $; $Name$

package Feedfetcher::Boss;

use strict;

use Feedfetcher::Pull;
use Feedfetcher::Push;
use Feedfetcher::Process;
use Feedfetcher::DocLogger;
use Feedfetcher::WorkerData;
use Core::Configurator;
use Core::Log;
use Util::SharedData;

use threads;
use threads::shared;
use Thread::Semaphore;
use Thread::Queue;

use Getopt::Std;
use File::Spec;
use File::Slurp;

sub new {
    my ($class, $args, $home_dir, $cb_abort) = @_;

	my $self = {};
	bless $self, $class;
	$self->initialize($args, $home_dir, $cb_abort);
	return $self;
}

sub initialize {
	my ($self, $args, $home_dir, $cb_abort) = @_;

	# Instantiate a SharedData instance. Used to cache config
	# module instance later.
	$self->{shared_data} = Util::SharedData->new();
	$self->{cb_abort} = $cb_abort;
	$args ||= {};
	my $instance = $args->{config} || 'feedfetcher';
	if (defined $home_dir) {
		$ENV{'HOMEDIR'} = $home_dir;
	}
	$home_dir ||= Core::Util::get_home_path();
	my $cfg = Core::Configurator->new($home_dir, $instance);
	$self->check_cfg($cfg);
	
	if ($args->{version}) { 
		$self->print_version($cfg);
		exit 0;
	}

	# Set up logging
	my $log = new Core::Log { 'log_path' => $cfg->required_param('log_path'), 
		'console' => $args->{print}, 'instance' => $instance,
		'level_threshold' => $cfg->param('log_level') };
	my $debug_log = $log;
	
	if (defined $cfg->param('ignore_last_update')) {
		if ($cfg->param('ignore_last_update') eq '1' || $cfg->param('ignore_last_update') =~ /true/i) {
#            print "ignore_last_update found\n";
    		$args->{ignore_last_update} = 1;
		}
	}
	
	# Set up process control
	my $abort = 0;
	my $abort_ref = \$abort;
	share($abort_ref);
	my $parser = new XML::LibXML;
	share($parser);
	my $main_lock = 1;
	my $main_lock_ref = \$main_lock;
	share($main_lock_ref);

	my $params = { args => $args, instance => $instance, cfg => $cfg, 
		log => $log };

	# Set up shared objects
	$params->{abort} = $abort_ref;
	my $doc_logger_table = &share({});
	my $doc_logger_fn = "doc_logger_$instance.dat";
#	print "seeting doc_logger: $doc_logger_fn\n";
	$params->{doc_logger} = Feedfetcher::DocLogger->new($doc_logger_table, $cfg, $log, 
		$doc_logger_fn, $args->{all});
	$params->{raw_table} = &share({});
	$params->{queue_semaphore} = Thread::Semaphore->new(0);
	$params->{push_queue} = Thread::Queue->new;
	$params->{pull_queue} = Thread::Queue->new;
	$params->{parser} = $parser;
	$params->{main_lock} = $main_lock_ref;
	$self->{use_database} = ((defined $cfg->param('use_database')) && (lc($cfg->param('use_database')) eq 'true'));

#	if ($self->{use_database}) {
#		my $last_time = time;
#		my $last_time_ref = \$last_time;
#		$params->{last_dbstore_time} = share($last_time_ref);
#		my $db_store_finished = 0;
#		my $db_store_finished_ref = \$db_store_finished;
#		$params->{db_store_finished} = share($db_store_finished_ref);
#		$params->{last_fetched_time} = share($last_time_ref);
#	}
	

	# Cache the config so that it can be retrieved elsewhere
	# without having to be passed in.
	$self->{shared_data}->set('cfg', $cfg);
	
	$self->{data} = Feedfetcher::WorkerData->new($params);
}

sub data {
	my ($self, $key) = @_;

	return $self->{data}->data->{$key};
}

sub launch {
	my ($self) = @_;

	my $cfg = $self->data('cfg');
	if ($cfg->param('push_mode') ne 'active' && 
		$cfg->param('pull_mode') ne 'active') {
		die "At least one of Push_mode or Pull_mode must be 'active'.\n";
	}
	$self->data('log')->log("[MED] Push Mode: " . $cfg->param('push_mode') );
	$self->data('log')->log("[MED] Pull Mode: " . $cfg->param('pull_mode') );
	my $args = $self->data('args');
	$self->set_pid;

	eval {
		$self->launch_push if ((! $args->{test}) && ($cfg->param('push_mode') eq 'active'));

		$self->launch_pull if $args->{test} ||
			($cfg->param('pull_mode') eq 'active');
		$self->launch_process;

		until ($self->check_abort) { sleep 1; }
		$self->data('log')->log("[MED] Feedfetcher was shutdown by user") if 
			$self->{interrupt} == 1;
		$self->cleanup;
	};
	if ($@) {
		$self->data('log')->log("[ERR] $@");
	}
	$self->remove_pid;
}

sub run_pull {
	my ($self) = @_;

	my $ff = Feedfetcher::Pull->new($self->{data});
	$ff->launch(); 

	# Workaround for perl thread bug
	sleep 1;
}

sub run_push {
	my ($self) = @_;

	Feedfetcher::Push::launch($self->{data}); 
}

sub run_proc {
	my ($self) = @_;

	my $ff_proc = Feedfetcher::Process->new($self->{data});
	$ff_proc->launch();
}

sub launch_pull {
	my ($self) = @_;

	$self->{pull_thread} = threads->create("run_pull", $self);
}

sub launch_push {
	my ($self) = @_;

	$self->{push_thread} = threads->create("run_push", $self);
}

sub launch_process {
	my ($self) = @_;

	$self->{proc_thread} = threads->create("run_proc", $self);
}

sub cleanup {
	my ($self) = @_;

	$self->{pull_thread}->join if $self->{pull_thread};
	$self->{push_thread}->join if $self->{push_thread};
	$self->{proc_thread}->join if $self->{proc_thread};
}

sub set_interrupt {
	my ($self) = @_;

	$self->{interrupt} = 1;
	$self->{data}->set_abort;
}

sub check_abort {
	my ($self) = @_;

	$self->{data}->set_abort if $self->{cb_abort} && &{$self->{cb_abort}};
	return ${$self->data('abort')};
}

sub set_pid {
	my ($self) = @_;

	my $cfg = $self->data('cfg');
	my $home_dir = $cfg->get_home_dir();
	my $file = $cfg->name || 'feedfetcher';
	my $pid_path = File::Spec->catfile($home_dir, "${file}.pid");
	write_file($pid_path, $$);
}

sub remove_pid {
	my ($self) = @_;

	my $cfg = $self->data('cfg');
	my $home_dir = $cfg->get_home_dir();
	my $file = $cfg->name || 'feedfetcher';
	my $pid_path = File::Spec->catfile($home_dir, "${file}.pid");
	unlink $pid_path if -f $pid_path;	
}

sub print_version {
	my ($self) = @_;

    # This is supposed to be searched/replaced
    # with Name: [version date]
    # e.g. s/\$Name\$/Name: 20080619/
    # When this is done the proper date for the version number will print out
    # NOTE:
    # If you are looking at a non-repository (i.e. built) copy of this file
    # you will probably see something like
    # my $version = 'Name: 2008.06.19';
    # rather than
    # my $version = '$Name$';
    my $version = q%$Name$%;
    $version =~ /Name:\s*(\S+)\s*$/ or $version = undef;
    $version = $1 || 'unknown';

    # Get year - no external dependencies
    my ($sec,$min,$hour,$mday,$mon,$year,
      $wday,$yday,$isdst) = localtime;
    $year += 1900;

	$version = $self->get_version_number($version);
	my $pretty_version = $version ? "version $version" : "(unknown version)";
	print "FeedFetcher, $pretty_version.\n(C) $year XML Team Solutions, Inc.\n";

    #    Sample pritout:
    #    FeedFetcher, version 2008.06.19.
    #    (C) 2008 XML Team Solutions, Inc.
}

sub check_cfg {
	my ($self, $cfg) = @_;

	$cfg->param('feed_error_dir') or 
		die "Feed_error_dir is not defined in config";
}


sub get_version_number {
	my ($self, $version_text) = @_;

	$version_text =~ /([\d-.]+)$/;
	my $num = $1;
	$num =~ s/-/./g;

	return $num;
}

sub print_help {
	my $msg = <<'END';
-h print this help message
-x extra feed arguments. Equivalent to the arguments from a browser. Put double quotes around the value.
   You must use an ampersand between arguments. There is no need to lead or trail with '&' or '?'.
   If you use "start", that value will be used instead of the last_accessed value or any value you supply
   via the -S parameter (minutes before now). {other args} means any other arguments you need to use.
   
   Examples
   {other args} -x "fixture-keys=publisher-keys=nhl.com&fixture-keys=-heartbeat,schedule-day&start=PT1H"
   {other args} -x "fixture-keys=event-stats&event-status=mid-event&league-keys=l.mlb.com&start=PT30M"
   {other args} -x "start=PT3H"
   
-p used for PRINTING the log to the command window
-i means IGNORE the last_update behavior when loading game subscores, etc., for FeedFetcher Deluxe
-t used for running ff just one TIME (as opposed to every [Interval] minutes)
-a used for retrieving ALL content, even content pulled previously
-S VALUE used to specify a particular [Since] value (minutes before now) on the command line, instead of below
         (Note: Uses a Captial S, not lower-case like below.)
-s used to ignore last-accessed time and instead get all content SINCE some time interval ago;
         (Gets content from the last x minutes defined by the [Since] param)
-w used for running ff WITHOUT throttling (will ignore Stop_threshold)
         (Throttling prevents too many documents from being downloaded
         by FeedFetcher at once.)
-c CONFIG_FILE reads from a named feedfetcher.cfg file, rather than /conf/feedfetcher.cfg
         (leave the .cfg extension part off of the command-line, as in: -c alternate-config
         in order to use conf/alternate-config.cfg)
-l used to load LOCAL content, as defined by [Source_dir]
-k used along with -l for loading local content. Means: KEEP files there after loading
        (as opposed to the usual behavior of deleting them)
-R used to write out incoming files into a RAW working directory, specified in Raw_dir
-L VALUE used to specify the number of most recent documents to pull, when using an alternate Plugin as a source
-F PLUGIN_FILTER_NAME used to specify that content should get filtered through the named Plugin

General Examples

start in perl directory with 'perl Feedfetcher/ff.pl' or '.\Feedfetcher\ff.exe' on Windows

Get last 60 minutes of files, running once then exiting.
    -pwaitS 60 OR -pwait -x "start=PT60M"
Process files from local-source, running once then exiting.
    -pitwalk
Get files for sportsdirectinc.com for league l.mlb.com for last 10 minutes.
    -pwait -x "publisher-keys=sportsdirectinc.com&league-keys=l.mlb.com&start=PT10M"
Get all authorized files for last 60 minutes. Skip heartbeat files.	
    -pwait -x "fixture-keys=-heartbeat&start=PT1H"
END
    print $msg;
}

sub init_args {
	my (%opts) = @_;

    if (defined $opts{'h'}) {
    	print_help();
    	exit 0;
    }

	return {
		'wink' => $opts{'w'}, 'test' => $opts{'t'},
		'print' => $opts{'p'}, 'local' => $opts{'l'},
		'all' => $opts{'a'}, 'use_since' => $opts{'s'},
		'since' => $opts{'S'}, 'config' => $opts{'c'},
		'last' => $opts{'L'}, 'raw' => $opts{'R'},
		'leave_local' => $opts{'k'}, 'filter_module' => $opts{'F'},
        'extra_args' => $opts{'x'}, 'xt_feed_path' => $opts{'P'},
		'version' => $opts{'v'}, 'iteration' => $opts{'I'},
		'suppress_db' => $opts{'d'}, 'ignore_last_update' => $opts{'i'}
	};
#        'extra_args' => $opts{'x'}, 'xt_feed_path' => $opts{'P'},
}

1;
