# $RCSfile$; $Date: 2013-04-01 13:11:11 -0700 (Mon, 01 Apr 2013) $; $Revision: 3022 $; $Name$
package Feedfetcher::DocLogger;

use strict;

use FileHandle;
use URI;
use File::Spec;

use Core::CircularQueue;

sub new {
  my ($class, $shared_table, $cfg, $log, $filename, $disabled) = @_;

	my $self = {};
	bless $self, $class;
	$self->initialize($shared_table, $cfg, $log, $filename, $disabled);
	return $self;
}

sub initialize {
	my ($self, $shared_table, $cfg, $log, $filename, $disabled) = @_;

	$self->{disabled} = $disabled;
#	print "disabled: $disabled\n";
	return if $disabled;

	my $max_files = $cfg->param('max_logging_files') || 100;
	$self->{'log'} = $log;

	$self->{max_files} = $max_files;
	$self->{dirty_writes} = 0;

	# Circular queue to make sure we don't save too many docs
	$self->{'queue'} = new Core::CircularQueue($max_files);

	# Holds doc_ids of files we've recently processed
	# Used as a lookup table to check duplicate processing
	$self->{'table'} = $shared_table;

	my $path = File::Spec->catfile($cfg->get_home_dir, $filename);
	$self->{'path'} = $path;
#	print "path: $path\n";
	$self->_load($path);
	$self->{'fh'} = new FileHandle ">>$path" or 
		die "Can't append to $path: $!";
	$self->{'fh'}->autoflush;
}

sub _load {
	my ($self, $path) = @_;

	my $fh = new FileHandle $path or return;
	while (my $line = <$fh>) {
		chomp $line;
		$self->_add($line);
	}
	undef $fh;
}

sub _add {
	my ($self, $doc_id) = @_;

	my $deleted = $self->{'queue'}->add($doc_id);
	delete $self->{'table'}->{$deleted} if defined $deleted;
	$self->{'table'}->{$doc_id} = 1;
}

sub was_processed {
	my ($self, $doc_id) = @_;

	return 0 if $self->{disabled};
	return $self->{'table'}->{$doc_id};
}

sub log {
	my ($self, $doc_id) = @_;

	return if $self->{disabled};

	$self->_add($doc_id);
#	print "fn: ".$self->{'fh'}."\n";
	my $fh = $self->{'fh'};
	print $fh "$doc_id\n";
	$self->{dirty_writes} += 1;

	if ($self->{dirty_writes} > $self->{max_files}) {
		$self->cleanup;
		$self->{dirty_writes} = 0;
	}
}

sub cleanup {
	my ($self) = @_;

	return if $self->{disabled};

	my $path = $self->{'path'};
	$self->{'fh'}->close() if $self->{'fh'};
	my $fh = new FileHandle ">$path" or die "Can't open $path for writing: $!";
	my $queue = $self->{'queue'};

	for (my $i = 0; $i < $queue->len; $i++) {
		my $doc_id = $queue->get($i);
		print $fh "$doc_id\n";
	}
	undef $fh
}

1;
