# $RCSfile: Process.pm,v $; $Date: 2016-04-07 12:22:44 -0400 (Thu, 07 Apr 2016) $; $Revision: 3073 $; $Name:  $

package Feedfetcher::Process;

use strict;

# XML Team modules
use Core::Log;
use Core::XmlUtil;
use Core::Util;
use Core::RegisterXSLT;
use Core::ActiveFixtures;
use Feedfetcher::TransformPlugin;
use Feedfetcher::Util;

#CPAN modules
use XML::LibXML;
use XML::LibXSLT;
use File::Path;
use File::Basename;
use FileHandle;
use Carp;
use File::Copy qw(move);
use Digest::MD5 qw(md5 md5_hex md5_base64);

sub new {
    my ($class, $data) = @_;

	my $self = {};
	bless $self, $class;
	$self->initialize($data);
	return $self;
}

sub cfg { my ($self) = @_; return $self->{'cfg'}; }

sub log { 
	my ($self, $msg) = @_; 
	$self->{'log'}->log($msg) if $msg; 
	return $self->{'log'}; 
}

sub args {
	my ($self, $arg) = @_; 
	return $arg ? $self->{'args'}->{$arg} : $self->{'args'}; 
}

# Initialize objects and variables needed for processing
sub initialize {
	my ($self, $worker_data) = @_;

	my $data = $worker_data->data;
	$self->{args} = $data->{args};
	$self->{cfg} = $data->{cfg};
	$self->{log} = $data->{log};
	$self->{wd} = $worker_data;
#	$self->{last_dbstore_time} = $data->{last_dbstore_time};
#	$self->{db_store_finished} = $data->{db_store_finished};

	$self->check_transformations();

	# Set up logging
	$self->{process_log} = Core::Log->new({ 
		'log_path' => $self->cfg->required_param('log_process_path'), 
		'console' => 0 }) if $self->cfg->param('log_process_path');
	$self->{'validation_log'} = Core::Log->new({ 
		'log_path' => $self->cfg->required_param('log_validation_path'), 
		'console' => $self->args('print') }) if $self->cfg->param('validator');

	# Set up database storage, if required
	if ($self->use_database()) {
		require FeedfetcherDeluxe::DbStore;
		eval {
			$self->{'dbstore'} = new FeedfetcherDeluxe::DbStore($self->cfg, $self->log,
			  $self->args);
		};
		if ($@) {
			$self->{log}->log("[ERR] There was an error processing document: $@");
			$self->{wd}->set_abort;
		}
	}

	$self->{'active_docs_fixtures'} = 
		Core::ActiveFixtures::load($self->cfg, 'active-fixtures-docs');
	$self->{'active_fixtures_db'} = 
		Core::ActiveFixtures::load($self->cfg, 'active-fixtures-db');

	$self->init_validation();

	$self->{doc_logger} = $data->{doc_logger};
	$self->{push_queue} = $data->{push_queue};
	$self->{pull_queue} = $data->{pull_queue};
	$self->{queue_semaphore} = $data->{queue_semaphore};
	$self->{raw_table}  = $data->{raw_table};
	$self->{abort}  = $data->{abort};
}

sub launch {
	my ($self) = @_;

	until ($self->check_abort) { $self->process_queue; }
}

sub init_validation {
	my ($self) = @_;

	my $validator = $self->cfg->param('validator');
	return if !$validator;

	my $vdir = 'Feedfetcher::Validation';
	my $vdir_path = 'Feedfetcher/Validation';

	my $vpackage = $vdir . '::' . $validator;
	my $vpath = "$vdir_path/${validator}.pm";
	require $vpath;
	$self->{validation_module} = 
		${vpackage}->new($self->cfg, $self->{validation_log});
}

sub check_transformations {
  my ($self) = @_;
  
	# Names of the transformations
	my @transforms = $self->cfg->params('transform');

	# Where to dump the output
	my @transform_paths = $self->cfg->params('transform_dir');

	# What type to make the output file
	my @transform_suffixes = $self->cfg->params('transform_suffix');

	# What path layout scheme to use
	my @transform_layouts = $self->cfg->params('transform_layout');

	my $count = @transforms;
	for (my $i = 0; $i < $count; $i++) {
		my $transform_num = $i + 1;
		my $transform_chain = $transforms[$i] or 
			die "Transform $transform_num is empty";
		$transform_paths[$i] or die "No path given for $transform_chain";
		$transform_suffixes[$i] or die "No suffix supplied for $transform_chain";
		$transform_layouts[$i] or die "No path_layout supplied for $transform_chain";
	}
}

# Return and cache a stylesheet object for a given transformation
sub get_stylesheet {
	my ($self, $transform) = @_;

	return $self->{'xslts'}->{$transform} if $self->{'xslts'}->{$transform};
	my $base_path = $self->cfg->param('stylesheet_dir') or 
		die "Stylesheet_dir not defined in config file";

	my $xslt = XML::LibXSLT->new();
	my $path = "$base_path/${transform}.xsl";
	die "No such transform '$transform', $path doesn't exist" if
		!-f $path;
	my $style_ref = $xslt->parse_stylesheet_file($path) or 
		die "Can't load stylesheet for $transform: $!";
	$self->{'xslts'}->{$transform} = $style_ref;

	return $style_ref;
}

sub process_queue {
	my ($self) = @_;

	my $queue_semaphore = $self->{queue_semaphore};
#	lock($queue_semaphore);
	$queue_semaphore->down;

	my $xml = $self->{push_queue}->dequeue_nb;
	$xml = $self->{pull_queue}->dequeue_nb if !$xml;

	return if $self->check_abort;	

	die "Both queues empty, semaphore not empty, snh" if !$xml;

	$self->log("[DBG] Processing beginning...");

	my $doc_logger = $self->{doc_logger};
	my ($path, $doc, $raw);
	my $lock = $self->{wd}->main_lock;
	$self->log("[DBG] Acquiring process lock...");
#	my $last_dbstore_time = time;
#	$self->{last_dbstore_time} = \$last_dbstore_time;
#	$self->log("[DBG] last_dbstore_time: ".${$self->{last_dbstore_time}});
	lock($lock);	    # Safeguard parser
	$self->log("[DBG] Process lock acquired.");
	
    if ($self->check_abort) {
#        $self->log("[DBG] !!!!! check_abort retuned true.");
    	return;
    }   
	
#	return if $self->check_abort;	
	my $force_load = 0;
	my $force_generic = 0;
	my $doc_id = undef;

	eval {
		$doc = $self->{wd}->parse_string($xml);
		my $xtsTree = $doc;
		my $force_load_txt = attribute_xpath($doc, 
			'/sports-ff-wrapper/@force-load');
		my $force_generic_txt = attribute_xpath($doc, 
			'/sports-ff-wrapper/@force-generic');
		$force_load = $force_load_txt && $force_load_txt eq 'true';
		$force_generic = $force_generic_txt && $force_generic_txt eq 'true';
		my $root = solonode($doc, 
			"/sports-ff-wrapper/sports-content");
		$doc->setDocumentElement( $root );
		# Add xts namespace, but don't change the namespace of the node (the zero final arg)
		# See http://search.cpan.org/~pajas/XML-LibXML/LibXML.pod
		$doc->documentElement->setNamespace("http://www.xmlteam.com","xts", 0);
		$xml = $doc->toString(0);

		$doc_id = required_attribute_xpath($doc, 
			'/sports-content/sports-metadata/@doc-id');

		return if !$force_load && $doc_logger->was_processed($doc_id);

#		$self->log("[DBG] doc-id: $doc_id");
		$raw = $self->{raw_table}->{$doc_id};
		delete $self->{raw_table}->{$doc_id};	# free up memory

		if (!Core::ActiveFixtures::doc_lookup($self->{'active_docs_fixtures'},
			$doc, $self->cfg)) {
			$self->log("[ADV] Document not in active fixtures list - skipping.");
			return;
		}

		# Validate
		$self->{'validation_log'}->trap( sub {
			$self->{validation_module}->validate($doc, $xml)
		}) if $self->{validation_module};	

		# Store all relevant information from document in database
		if ($self->use_database()) {
			my $db_store_finished = 0;
			if (Core::ActiveFixtures::doc_lookup($self->{active_fixtures_db}, $doc, $self->cfg)) {
				$self->{db_store_finished} = \$db_store_finished;
				$self->{'dbstore'}->store_doc($doc, $xml, $force_load) 	
			}
			$db_store_finished = 1;
			$self->{db_store_finished} = \$db_store_finished;
		}
			
		

		# Write file to disk, with transformations
		$self->transform($doc, $xml, $raw, $force_generic);

        # Pipe through user-defined command, if necessary
        my $cmd = $self->cfg->param('postprocessor');

                my $pptype = $self->cfg->param('postprocessortype');
                if ($pptype =~ /docid/i) {
                        if ($cmd && $cmd !~ /^none$/i) {
                                $self->pipe_to_command($cmd, $doc_id);
                        }
                } else {
                        if ($cmd && $cmd !~ /^none$/i) {
                                $self->pipe_to_command($cmd, $xml);
                        }
                }

# NOTE: Changes as per client's suggestion
#		if ($cmd && $cmd !~ /^none$/i) {
#			$self->pipe_to_command($cmd, $xml);
#		}

		my $pending = $self->{push_queue}->pending + $self->{pull_queue}->pending;
		
		# Get MD5 hash of document
		my $digest = md5_hex($xml);
		
		$self->log("[HI ] Processed $doc_id digest $digest ($pending left)");
	};
	
	if ($@) {
		my $error_msg = $@;
		eval {
			$path = $self->error_path($doc);
		};
		if (!$@) {
			my $xml_path = "$path.xml";
			my $raw_path = "$path.raw";
			my $fh = new FileHandle ">$xml_path" or die "Can't open $xml_path for writing: $!";
			print $fh $xml;
			undef $fh;
			if ($raw) {
				$fh = new FileHandle ">$raw_path" or die "Can't open $raw_path for writing: $!";
				print $fh $raw;
				undef $fh;
			}
		}
		my $doc_id_txt = $doc_id ? $doc_id : "unknown doc-id";
		$self->log("[ERR] <${doc_id_txt}> $error_msg");
	} else {
		my $doc_id = required_attribute_xpath($doc, 
			'/sports-content/sports-metadata/@doc-id');
		$doc_logger->log($doc_id) if !$force_load;
	}

	return $doc;
}

sub pipe_to_command {
	my ($self, $cmd, $output) = @_;

	open POSTPROC, "| $cmd" or die "can't pipe to $cmd: $!";
	local $SIG{PIPE} = sub { die "postprocessing pipe broke" };
	print POSTPROC $output;
	close POSTPROC or die "couldn't pipe to $cmd: $! $?";
}

# Transform document with transform plugins or xslt stylesheets
sub transform {
	my ($self, $doc, $txt, $raw, $force_generic) = @_;

	my $transform_prefix = $force_generic ? 'generic_' : '';
	
	# Names of the transformations
	my @transforms = $self->cfg->params($transform_prefix . 'transform');

	# Where to dump the output
	my @transform_paths = $self->cfg->params($transform_prefix . 'transform_dir');

	# Where to dump the staging output
	my @transform_staging_paths = $self->cfg->params($transform_prefix . 'transform_staging_dir');

	# What type to make the output file
	my @transform_suffixes = $self->cfg->params($transform_prefix . 'transform_suffix');

	# What path layout scheme to use
	my @transform_layouts = $self->cfg->params($transform_prefix . 'transform_layout');

	# Which fixtures to transform for
	my @transform_active_fixture_paths = 
		$self->cfg->params($transform_prefix . 'transform_active_fixtures');
	my @transform_active_fixtures = map 
		{ Core::ActiveFixtures::load($self->cfg, $_) } @transform_active_fixture_paths;
	
	
	my $count = @transforms;
	for (my $i = 0; $i < $count; $i++) {
		my $transform_chain = $transforms[$i] or
			die "Transform is empty";
		my $transform_path = $transform_paths[$i] or
			die "No path given for $transform_chain";
		my $transform_staging_path = $transform_staging_paths[$i] or
			die "No staging directory supplied for $transform_chain";
		my $transform_suffix = $transform_suffixes[$i] or 
			die "No suffix supplied for $transform_chain";
		my $transform_layout = $transform_layouts[$i] or
			die "No path_layout supplied for $transform_chain";
		my $transform_active_fixtures = $transform_active_fixtures[$i] or
			die "No active_fixtures supplied for $transform_chain";

		if ($transform_active_fixtures &&
			!Core::ActiveFixtures::doc_lookup($transform_active_fixtures,
			$doc, $self->cfg)) {
			next;
		}
		
		my @chain = split(/\s*,\s*/, $transform_chain);
		my $input = $txt;
		foreach my $transform (@chain) {
			if (Feedfetcher::TransformPlugin::is_plugin($transform)) {
				# Transform with plugin
				# TEMP: HARDCODE html2text plugin params
				my $data = { 'nl' => "\n", 'endp' => "\n", 'indent' => 4 };
				$input = Feedfetcher::TransformPlugin::transform($transform, 
					$input, $data);
			} else {	
				# Transform with xsl file
				my $cur_doc = $self->{wd}->parse_string($input);
				my $stylesheet = $self->get_stylesheet($transform);
				die "No such stylesheet: $transform" if !$stylesheet;
				my $result = $stylesheet->transform($cur_doc);
				$input = $stylesheet->output_string($result);
			}
		}

		# Write out the output
		my $output = $input;
		if ($output !~ /<killme\s*\/>/) {
			mkpath($transform_staging_path);
			my $staging_path = $transform_staging_path . "/staging.tmp";
			my $path = $self->build_path($transform_path, $transform_layout, $doc);
#           print "path: $path\n";
			my $t_path = "$path.$transform_suffix";
#           print "staging_path: $staging_path\n";
#           print "t_path: $t_path\n";
			my $out_file = new FileHandle ">$staging_path" or die "$! ($staging_path)";
			print $out_file $output;
			undef $out_file;
#			rename $staging_path, $t_path or die $!;
            # Use the Copy module's capabilities rather than the rename call, since "rename" can result 
            # in failure to work across filesystem volumes
            move $staging_path, $t_path or die $!;

			# Write out raw
			if ($raw && @chain == 1 && $chain[0] eq 'identity') {
				my $raw_path = "$path.raw";
				my $raw_out_file = new FileHandle ">$raw_path" or die "$! ($raw_path)";
				print $raw_out_file $raw;
				undef $raw_out_file;
			}		
		}
	}
}

# Return the path to write the file to if there was an error
sub error_path {
	my ($self, $doc) = @_;

	$self->{'error_cnt'} += 1;	# Global variable
	my $base_path = $self->cfg->required_param('feed_error_dir');
	mkpath($base_path);

	# Find the doc-id
	my $metanode = solonode($doc, "/sports-content/sports-metadata");
	my $doc_id = $metanode ? attribute($metanode, "doc-id") : undef;
    $doc_id =~ s/\///g;

	# Get the current date/time
	my $timestamp = Core::Util::current_ISO8601_time();

	my $filename;
	my $error_cnt = $self->{'error_cnt'};
	if ($doc_id) {
		$filename = $doc_id . "_${timestamp}_$error_cnt";
	} else {
		$filename = "unclassified_${timestamp}_$error_cnt";
	}

	return $base_path . "/$filename";
}

# Return the path to write the file to if everything was OK
sub build_path {
	my ($self, $base_path, $path_layout, $doc) = @_;

	my @res;
	$path_layout = lc($path_layout);
	if ($path_layout =~ 'flat') { @res = $self->path_flat($doc); }
	elsif ($path_layout eq 'slug') { @res = $self->path_slug($doc); }
	elsif ($path_layout eq 'basic') { @res = $self->path_basic($doc); }
	elsif ($path_layout eq 'date_content_tree') { 
		@res = $self->path_date_content_tree($doc); 
	} elsif ($path_layout eq 'content_tree') { 
		@res = $self->path_content_tree($doc); 
	} elsif ($path_layout eq 'event') { 
		@res = $self->path_event($doc); 
	} elsif ($path_layout eq 'comp_games') {
		@res = $self->path_comp($doc);
	} else { 
		# substitution method
		@res = $self->path_sub($doc, $path_layout);
	}

	return if !@res;
	my ($rel_dir, $file) = @res;
	my $dir = $rel_dir ? "$base_path/$rel_dir" : $base_path;
	
	# Create directories for path if they don't exist
	mkpath($dir);
	
	return "$dir/$file";
}

# Subroutines that create specific path layouts
# ---------------------------------------------

# Path structure: doc-id.xml
sub path_flat {
	my ($self, $doc) = @_;

	my $metanode = 
		solonode($doc, "/sports-content/sports-metadata");
	my $doc_id = required_attribute($metanode, "doc-id");
    $doc_id =~ s/\///g;

	return (undef, "$doc_id");
}

# Path structure: /yyyy/mm/dd/doc-id.xml
sub path_basic {
	my ($self, $doc) = @_;

	my $metanode = 
		solonode($doc, "/sports-content/sports-metadata");
	my $doc_id = required_attribute($metanode, "doc-id");

	my $dir = $self->date_path($metanode);
	my $file = $doc_id;
	$file =~ s/\///g;
	
	
	return ($dir, $file);
}

# Path structure: /yyyy/mm/dd/sport-alias/league-key/document-class/fixture-key/doc-id.xml
sub path_date_content_tree {
	my ($self, $doc) = @_;

	my $metanode = 
		solonode($doc, "/sports-content/sports-metadata");

	my ($sub_path, $doc_id) = $self->path_content_tree($doc);
	my $date_path = $self->date_path($metanode);
	$doc_id =~ s/\///g;
	
	return ("$date_path/$sub_path", $doc_id);
}

sub get_slug {
	my ($self, $metanode) = @_;

	my $general_slug = "no_slug";
	my $slug_query = 
		'sports-content-codes/sports-content-code[@code-type="slug"]';
	my $slug_content_node = solonode($metanode, $slug_query, 1);
	return $general_slug if !$slug_content_node;
	my $slug = required_attribute($slug_content_node, "code-key");
	return $general_slug if !$slug;
	return $slug;
}

# slug.xml
sub path_slug {
	my ($self, $doc) = @_;

	my $metanode = 
		solonode($doc, "/sports-content/sports-metadata"); 
	my $doc_id = required_attribute($metanode, "doc-id"); 
	my $slug = $self->get_slug($metanode);

	$slug = substr($slug, index($slug, '^') + 4 - length($slug));
	$slug =~ s/\ /_/g;
	$slug =~ s/,/_/g;
	$slug = substr($slug, 0, index($slug, '<'));
	$slug = substr($slug, 0, index($slug, '&'));
#	$slug =~ s/\</_/g;
#	$slug =~ s/\>/_/g;

#	print "\n*** in slug: [$slug] ***\n";

	return (undef, "$slug");
}

sub path_sub {
	my ($self, $doc, $format) = @_;

	my $metanode = 
		solonode($doc, "/sports-content/sports-metadata");

	my $doc_id = required_attribute($metanode, "doc-id"); 
	my $doc_class = required_attribute($metanode, "document-class");
	my $fixture_key = required_attribute($metanode, "fixture-key");
	
	my $sport_name = $self->get_sport($metanode);
	my $league_key = $self->get_league($metanode);	
	my $team_code = $self->get_team_code($metanode) || 'undefined-team';
	$team_code = lc($team_code);

	my $event_query = "/sports-content/sports-event/event-metadata";
	my @event_nodes = nodes($doc, $event_query);
	my $event_key = @event_nodes == 1 ? required_attribute($event_nodes[0], "event-key") : 'undefined-event';

	my $date = $metanode->getAttribute("date-time");	
	my @date_arr = Feedfetcher::Util::parse_iso_date($date);
	if (!@date_arr) {
		die "Bad date format: $date";
	}
	my ($year, $month, $day) = @date_arr;
	$year = sprintf("%04d", $year);

	$format =~ s/\{doc-id\}/$doc_id/;
	$format =~ s/\{doc-class\}/$doc_class/;
	$format =~ s/\{fixture-key\}/$fixture_key/;
	$format =~ s/\{sport-name\}/$sport_name/;
	$format =~ s/\{league-key\}/$league_key/;
	$format =~ s/\{team-code\}/$team_code/;
	$format =~ s/\{event-key\}/$event_key/;
	$format =~ s/\{year\}/$year/;

	my ($base, $dir) = fileparse($format);
	return ($dir, $base);
}

sub get_sport {
	my ($self, $metanode) = @_;

	my $general_sport = "general";
	my $sport_query = 
		'sports-content-codes/sports-content-code[@code-type="sport"]';
	my $sport_content_node = solonode($metanode, $sport_query, 1);
	return $general_sport if !$sport_content_node;
	my $sport;
	$sport = required_attribute($sport_content_node, "code-key") if 
		$sport_content_node;
	my $sport_name;
	$sport_name = $sport_content_node->getAttribute("code-name");
	return $general_sport if !defined $sport_name || length($sport_name) == 0;
	return $self->lcdash($sport_name);
}

sub get_league {
	my ($self, $metanode) = @_;

	my $general_league = "l.sportsml.com.general";
	my $league_query = 
		'sports-content-codes/sports-content-code[@code-type="league"][1]';
	my $league_content_node = solonode($metanode, $league_query, 1);
	if (! $league_content_node) {
		print "no league found\n";
	}
	return $general_league if !$league_content_node;
	my $league = required_attribute($league_content_node, "code-key");
	return $general_league if !$league;
	return $league;
}

sub get_team {
	my ($self, $metanode) = @_;

	my $team_query = 
		'sports-content-codes/sports-content-code[@code-type="team"]';
	my $team_content_node = firstnode($metanode, $team_query);
	return if !$team_content_node;
	my $team = required_attribute($team_content_node, "code-key");
	return $team;
}

sub get_team_code {
	my ($self, $metanode) = @_;

	my $team_key = $self->get_team($metanode);
	my $dbstore = $self->{dbstore};
	if ($dbstore) {
		return $dbstore->get_team_code($team_key);
	}
	return undef;
}

# Path structure: sport-alias/league-key/document-class/fixture-key/doc-id.xml
sub path_content_tree {
	my ($self, $doc) = @_;

	my $metanode = 
		solonode($doc, "/sports-content/sports-metadata");
	my $doc_id = required_attribute($metanode, "doc-id"); 
	my $doc_class = required_attribute($metanode, "document-class"); 
	my $fixture_key = required_attribute($metanode, "fixture-key"); 
	
	my $sport_name = $self->get_sport($metanode);
	my $league = $self->get_league($metanode);

	return ("$sport_name/$league/$doc_class/$fixture_key", $doc_id);
}

# Path structure: 
# /sport-alias/league-key/document-class/fixture-key.xml
# /sport-alias/league-key/events/event-key/
#				pre-event-coverage.xml
#				post-event-coverage.xml
#				event-stats.xml
#				event-score.xml
sub path_event {
	my ($self, $doc) = @_;

	my $metanode = 
		solonode($doc, "/sports-content/sports-metadata"); 
	my $doc_id = required_attribute($metanode, "doc-id"); 
	my $doc_class = required_attribute($metanode, "document-class");
	my $fixture_key = required_attribute($metanode, "fixture-key");
	
	my $sport_name = $self->get_sport($metanode);
	my $league_key = $self->get_league($metanode);	
	my $team_key = $self->get_team($metanode);

	my ($sub_dir, $sub_file) = 
		$self->build_path_event_subpath($doc, $doc_class, $fixture_key, $team_key);

	return ("$sport_name/$league_key/$sub_dir", $sub_file);
}

# /sport-category/sport/competition/doc-id.xml
sub path_comp {
	my ($self, $doc) = @_;

	my $metanode = 
		solonode($doc, "/sports-content/sports-metadata"); 
	my $doc_id = required_attribute($metanode, "doc-id"); 
	my $content_codes = $self->get_content_codes($doc);
	my $sport_cat = $content_codes->{'sport-category'} || 'sc.sportsml.com.general';
	my $sport = $content_codes->{'sport'} || 's.sportsml.com.general';
	my $comp = $content_codes->{'competition'} || 'c.sportsml.com.general';
	
	return ("$sport_cat/$sport/$comp", $doc_id);
}

sub get_content_codes {
	my ($self, $doc) = @_;

	my @sport_content_nodes = 
		$doc->findnodes('/sports-content/sports-metadata/' .
		'sports-content-codes/sports-content-code');

	my %codes = ();
	foreach my $node (@sport_content_nodes) {
		my $code_type = $node->getAttribute('code-type');
		my $code_key = $node->getAttribute('code-key');

		next if (!$code_type || !$code_key);

		$codes{$code_type} = $code_key;
	}

	return \%codes;
}

sub build_path_event_subpath {
     my ($self, $doc, $doc_class, $fixture_key, $team_key) = @_;

     if ($fixture_key eq 'pre-event-coverage' ||
          $fixture_key eq 'post-event-coverage' ||
          $fixture_key eq 'event-stats' ||
          $fixture_key eq 'event-score' ||
          $fixture_key eq 'event-stats-progressive' ||
          $fixture_key eq 'event-stats-progressive-shifts' ||
          $fixture_key eq 'event-stats-progressive-splits' ||
          $fixture_key eq 'play' ||
          $fixture_key eq 'play-summary') {

          my $event_query = "/sports-content/sports-event/event-metadata";
          my $event_node = solonode($doc, $event_query);
          my $event_key = required_attribute($event_node, "event-key");

          return ("events/$event_key", $fixture_key);
     } else {
          my $file = $team_key ? "${fixture_key}-${team_key}" : $fixture_key;
          return ($doc_class, $file);
     }
}

# Return year/month/day based on date-time in document
sub date_path {
	my ($self, $metanode) = @_;

	my $date = $metanode->getAttribute("date-time");
	if (!defined $date) {
		die "No date-time";
		return;		
	}
	
	return Feedfetcher::Util::date_path($date);
}

# Returns whether this program is configured to store documents to the database
sub use_database {
	my ($self) = @_;
	
	my $use_db = $self->cfg->param('use_database');
	return $use_db && $use_db =~ /true/i && !$self->args('suppress_db');
}

# Convert text to lowercase, with dashes replacing spaces
sub lcdash {
	my ($self, $txt) = @_;
	$txt =~ s/ /-/g;
	$txt = lc($txt);
	return $txt;
}

sub check_abort {
	my ($self) = @_;

	return $self->{wd}->check_abort;
}

1;