# $RCSfile$; $Date: 2010-07-30 15:35:58 -0700 (Fri, 30 Jul 2010) $; $Revision: 2799 $; $Name$

package Feedfetcher::Pull;

use strict;

#use Time::HiRes qw(usleep gettimeofday);

$| = 1;

# XML Team modules
use Core::XmlUtil;
use Core::Util;
use Core::ActiveFixtures;

#CPAN modules
use XML::LibXML;
use XML::LibXSLT;
use File::Spec;
use File::Slurp;
use File::Path;
use File::Temp;
use FileHandle;
use Carp;
use LWP::Simple;

sub new {
    my ($class, $data) = @_;

	my $self = {};
	bless $self, $class;
	$self->initialize($data);
	return $self;
}

sub interval { my ($self) = @_; return $self->{'interval'}; }
sub cfg { my ($self) = @_; return $self->{'cfg'}; }

sub log { 
	my ($self, $msg) = @_; 
	$self->{'log'}->log($msg) if $msg; 
	return $self->{'log'}; 
}

sub args {
	my ($self, $arg) = @_; 
	return $arg ? $self->{'args'}->{$arg} : $self->{'args'}; 
}

sub launch {
	my ($self) = @_;

	my $t0 = time;		# Start time in seconds since epoch
	my $interval = $self->{interval};
	my $args = $self->{args};

	until ($self->{wd}->check_abort) { 
		# Catch exceptions
		eval {
			$self->process();	# Main processing code
		};
		if ($@) {
			$self->log("[ERR] $@");
		} 

		my $et = time - $t0;	# Elapsed seconds
		# Sleep for remaining time, or exit loop if $interval isn't set or if in test mode
		if (!$interval || $args->{test}) {
			$self->{wd}->wait_till_empty if !$self->{wd}->check_abort;;
			$self->{wd}->set_abort;
			return;
		}
		
		$t0 = time if $et >= $interval;
		while (time - $t0 < $interval) {
			last if ($self->{wd}->check_abort);
			sleep 1;
		}
	}

	$self->log("[MED] Pull was shut down");
}

# Initialize objects and variables needed for processing
sub initialize {
	my ($self, $work_data) = @_;

	my $data = $work_data->data;
	$self->{cfg} = $data->{cfg};
	$self->{log} = $data->{log};
	$self->{debug_log} = $data->{debug_log};	
	$self->{doc_logger} = $data->{doc_logger};
	$self->{push_queue} = $data->{push_queue};
	$self->{pull_queue} = $data->{pull_queue};
	$self->{use_database} = (($self->{cfg}->param('use_database')) =~ /true/i);
#	if ($self->{use_database}) {
#		$self->{db_storage_warning_threshold} = $self->{cfg}->param('db_storage_warning_threshold') || 240;
#		$self->{last_dbstore_time} = $data->{last_dbstore_time};
#		$self->{last_fetched_time} = $data->{last_fetched_time};
#		$self->{db_store_finished} = $data->{db_store_finished};
#	}
	$self->{queue_semaphore} = $data->{queue_semaphore};
	$self->{raw_table}  = $data->{raw_table};
	$self->{abort}  = $data->{abort};
	$self->{args}  = $data->{args};
	$self->{wd} = $work_data;

	my $interval = $self->cfg->param('pull_interval');
	$interval *= 60 if $interval;	# Convert to seconds (from minutes)
	$self->{'interval'} = $interval;

	$self->{'active_docs_fixtures'} = 
		Core::ActiveFixtures::load($self->cfg, 'active-fixtures-docs');

	$self->init_plugins();
}

sub init_plugins {
	my ($self) = @_;

	my $plugin_dir = 'Feedfetcher::Plugins';
	my $plugin_dir_path = 'Feedfetcher/Plugins';
	my @requested_plugins = $self->cfg->required_params('source');
	my @plugins = ();
	my $check_abort = sub { return $self->{wd}->check_abort; };

	my $filter_module_name = $self->args('filter_module') || 
		$self->cfg->param('local_default_plugin');
	my $filter_module = undef;
	if ($filter_module_name) {
		my $plugin = $plugin_dir . '::' . $filter_module_name;
		my $plugin_path = "$plugin_dir_path/${filter_module_name}.pm";
		require $plugin_path;
		$filter_module = ${plugin}->new(
			$check_abort, $self->args, $self->cfg, $self->log, $self->{wd});
	}
	
	if ($self->args('local')) {
		require Feedfetcher::Plugins::Local;
		push @plugins, 
			new Feedfetcher::Plugins::Local($check_abort, $self->args, 
			$self->cfg, $self->log, $self->{wd}, $filter_module);
	} else {
		foreach my $requested_plugin (@requested_plugins) {
			my $plugin = $plugin_dir . '::' . $requested_plugin;
			my $plugin_path = "$plugin_dir_path/${requested_plugin}.pm";
			require $plugin_path;
			push @plugins, ${plugin}->new($check_abort, $self->args, 
				$self->cfg, $self->log, $self->{wd}, $filter_module);
		}
	}

	 $self->{'plugins'} = \@plugins;
}

# Return and cache a stylesheet object for a given transformation
sub get_stylesheet {
	my ($self, $name) = @_;

	return $self->{'xslts'}->{$name} if $self->{'xslts'}->{$name};
	my $base_path = $self->cfg->param('stylesheet_dir') or 
		die "Stylesheet_dir not defined in config file";

	my $xslt = XML::LibXSLT->new();
	my $path = $self->cfg->param("${name}_xslt") or return undef;
	die "No transform for '$name', $path doesn't exist" if
		!-f $path;
	my $style_ref = $xslt->parse_stylesheet_file($path) or 
		die "Can't load stylesheet for $name: $!";
	$self->{'xslts'}->{$name} = $style_ref;

	return $style_ref;
}

# Process all unprocessed feeds
sub process {
	my ($self) = @_;

	$self->check_vet;			# Check if feed fetcher is clamped

#	$self->log("[DBG] BEGINNING A PROCESSING RUN FOR PULL");
#	if ($self->{'use_database'}) {
#		my $time_since_last_dbstore = (${$self->{last_fetched_time}} - ${$self->{last_dbstore_time}});
##		$self->log("[DBG] last fetched time: ".${$self->{last_fetched_time}});
##		$self->log("[DBG] last_dbstore_time from pull: ".${$self->{last_dbstore_time}});
##		$self->log("[DBG] db_store_finished: ".${$self->{db_store_finished}});
#		
##		$self->log("[DBG] DIFF: ".$time_since_last_dbstore);
#		if (($time_since_last_dbstore - 0) > $self->{db_storage_warning_threshold}) {
##			$self->log("[DBG] last_dbstore_time from pull: ".${$self->{last_dbstore_time}});
#			$self->log("[ERR] Pull DB_STORAGE_WARNING. No completed document storage in ".$time_since_last_dbstore." seconds.");
#		}
#		my $last_fetched_time = time;
#		$self->{last_fetched_time} = \$last_fetched_time;
#	}

	my $plugins = $self->{'plugins'};
	foreach my $plugin (@$plugins) {
		return if $self->{wd}->check_abort;		# Check for shutdown
		my $plugin_name = $plugin->name;
		my @files = undef;
		$self->log("[MED] Via Pull: Retrieving file list for $plugin_name...");
		eval {
			@files = $plugin->get_list();
		};
		if ($@) {
			$self->log("[ERR] Couldn't process $plugin_name: $@");
			next;
		}

		# Only list files that haven't been processed
		@files = grep { 
			!$self->{doc_logger}->was_processed($plugin->path2doc_id($_)) 
		} @files if $plugin->log_docs && !$self->args('all');

		return if $self->{wd}->check_abort;	# Check for shutdown
		my $num_files = @files;

		# Clamp feed fetcher if overloaded
		return if $self->vet($num_files, $self->args);	

		if ($num_files) {
			my $file_s = $num_files == 1 ? 'file' : 'files';
			$self->log("[MED] Via Pull: Processing $num_files $file_s for $plugin_name...");
		} else {
			
			my $last_dbstore_time = time;
			$self->{last_dbstore_time} = \$last_dbstore_time;

			$self->log("[MED] Via Pull: No new files, processing complete.");
		}
		my $count = 1;
		foreach my $file (@files) {
			return if $self->{wd}->check_abort;	# Check for shutdown
			eval {
				$self->download($plugin, $file, $count, $num_files);
			};
			$self->log("[ERR] $@") if $@;
			$count++;
		}

		$self->log("[MED] Via Pull: Processing completed for $plugin_name.") if 
			$num_files;

		$plugin->cleanup();
	}
}

sub download {
	my ($self, $plugin, $file, $count, $num_files) = @_;
	my $download_txt = (lc($plugin->code) eq 'txt');
	
#	print "download_txt: $download_txt\n";
	
	if ($download_txt eq '1') {
		$self->download_txt($plugin, $file, $count, $num_files);
	}
	elsif ($self->args('raw')) {
		$self->download_raw($plugin, $file, $count, $num_files);
	}
	elsif ($self->cfg->param('download_blob_only') eq '1') {
		$self->download_blob($plugin, $file, $count, $num_files);
	} else {
		$self->download_cooked($plugin, $file, $count, $num_files);
	}
}

sub download_raw {
	my ($self, $plugin, $file, $count, $num_files) = @_;

	my $doc = $plugin->retrieve($file);

	my $lock = $self->{wd}->main_lock;
	lock($lock);	    # Safeguard parser
	return if $self->{wd}->check_abort;		# Check for shutdown
	
#	$self->log("[DBG] Pre-filter");
	my $result = $plugin->filter_doc($doc);
	my $xml = $result->{'raw'};
	my $dir = $self->cfg->required_param('raw_dir');
	mkpath($dir);
	$file =~ s/\W//g;
	my $path = "$dir/$file.raw";
	my $fh = new FileHandle ">$path" or die "Can't open $path for writing: $!";
	print $fh $xml;
	$plugin->post_retrieve()
}

# Download a single file. This works for both local and remote files.
# Perform any requested operations on the file.
sub download_cooked {
	my ($self, $plugin, $file, $count, $num_files) = @_;
		
	my $xml = $plugin->retrieve($file);

	my $lock = $self->{wd}->main_lock;
	lock($lock);	    # Safeguard parser
	return if $self->{wd}->check_abort;		# Check for shutdown

	my $doc = $self->{wd}->parse_string($xml);	
	$doc = $self->apply_xslt($plugin, $doc);
	$xml = $doc->toString(0);
	my $result = $plugin->filter_doc($xml);
	my $raw = $result->{'raw'};
	my @xmls = @{$result->{'filtered'}};
	foreach my $xml (@xmls) {
		$self->process_doc($xml, $raw, $count, $num_files, $plugin, $plugin->forced, $plugin->generic);
	}
	$plugin->post_retrieve();
}

sub download_blob {
	my ($self, $plugin, $file, $count, $num_files) = @_;

	my $doc = $plugin->retrieve($file);

	my $lock = $self->{wd}->main_lock;
	lock($lock);	    # Safeguard parser
	return if $self->{wd}->check_abort;		# Check for shutdown
	
	my $dir = $self->cfg->required_param('raw_dir');
	mkpath($dir);

	my $fname = $file;
	my @parts = split('/', $fname);
	$fname = pop @parts;
    $fname =~ s/\?.*$//;

    my $blob_ext =  $self->cfg->param('pull_new_blob_ext');
    if (defined $blob_ext) {
        $fname =~ s/(.*)\..*$/$1\.$blob_ext/;
    }
#
#    print "download blob for file: $fname\n";
	
	$self->log("[LOW] Via Pull: ($count/$num_files) $fname");
	my $path = "$dir/$fname";
	my $fh = new FileHandle ">$path" or die "Can't open $path for writing: $!";
	if (($fname =~ /.*gif$/i) || ($fname =~ /.*jp.*g$/i)) {
	   binmode $fh;
	}
	print $fh $doc;
	$plugin->post_retrieve()
}

# Download (and save) files that are not xml
sub download_txt {
    my ($self, $plugin, $file, $count, $num_files) = @_;
        
    my $txt = $plugin->retrieve($file);
	
#	print "file: $file\n";
	my $fname = $file;
	my @parts = split('/', $fname);
	$fname = pop @parts;
#	print "fname: $fname\n";
	
    my $lock = $self->{wd}->main_lock;
    lock($lock);        # Safeguard parser
    return if $self->{wd}->check_abort;     # Check for shutdown

    my $result = $plugin->filter_doc($txt);
    my $raw = $result->{'raw'};
    my @xmls = @{$result->{'filtered'}};

#	my $time_stamp = $self->get_timestamp(1);
	my $dir = $self->cfg->required_param('txt_dir');
	# Create dir if it's not there
	mkpath($dir);
	$file =~ s/\W//g;
#	my $path = "$dir/$file.txt";
	# Use timestamp for now for filename prefix
	my $path = "$dir/${fname}.txt";
	my $fh = new FileHandle ">$path" or die "Can't open $path for writing: $!";
	print $fh $txt;
	undef $fh;
	$self->log("[DBG] Saving TXT plugin file $fname to $dir");
#    $plugin->post_retrieve();
}

sub apply_xslt {
	my ($self, $plugin, $doc) = @_;
	my $plugin_name = $plugin->code;

	# Transform with xsl file
	my $stylesheet = $self->get_stylesheet($plugin_name);
	return $doc if !$stylesheet;

	my $result = $stylesheet->transform($doc) or die "Couldn't transform document";
	return $result;
}

sub process_doc {
	my ($self, $xml, $raw, $count, $num_files, $plugin, $force_load, $force_generic) = @_;

	my $doc;
	eval {
#		print "process_doc\n";
		$doc = $self->{wd}->parse_string($xml);	
		my $doc_id = required_attribute_xpath($doc, 
			'/sports-content/sports-metadata/@doc-id');
		$self->{raw_table}->{$doc_id} = $raw;
		$self->log("[LOW] Via Pull: ($count/$num_files) $doc_id");
		my $queue = $self->{pull_queue};
		my $force_load_txt = $force_load ? "true" : "false";
		my $force_generic_txt = $force_generic ? "true" : "false";
		$xml =~ s/\<\?[^\?]*\?\>//;
		$xml =~ s/(\<\!DOCTYPE[^\>]*\>)//;
		$xml = qq{<?xml version="1.0" encoding="UTF-8"?>$1<sports-ff-wrapper force-load="$force_load_txt" force-generic="$force_generic_txt">} .
			qq{$xml</sports-ff-wrapper>};
		$queue->enqueue($xml);
		$self->{queue_semaphore}->up;
	};		
	if ($@) {
		my $error_msg = $@;
		my $base_path = $self->cfg->required_param('feed_error_dir');
		mkpath($base_path);

		my $tmp = new File::Temp( DIR => $base_path, SUFFIX => '.xml', 
			UNLINK => 0);
		print $tmp "$xml";
		$self->log("[ERR] Can't parse xml ($tmp): $error_msg");
		undef $tmp;
	}
}

sub vet {
	my ($self, $num_files) = @_;

	return if $self->args('wink');
	my $stop_threshold = $self->cfg->param('pull_stop_threshold');
	if ($stop_threshold && $num_files > $stop_threshold) {
		my $cfg_dir = $self->cfg->get_home_dir();
		my $fh = FileHandle->new( '>' . 
			File::Spec->catfile($cfg_dir, '.cancel_ff') );
		$fh->close();
		my $filename = "stop_alarm_". time;
		$self->drop_file($filename, 'stop_file');
		my $msg = "There were $num_files new files to be downloaded, " . 
			"but threshold set to $stop_threshold!";
		$self->raise_alarm("ABORTED: $msg");
		$self->{wd}->set_abort;

		return 1;
	}

	my $warn_threshold = $self->cfg->param('pull_warn_threshold');
	if ($warn_threshold && $num_files > $warn_threshold) {
		my $filename = "warning_alarm_". time;
		$self->drop_file($filename, 'warn_file');
		my $msg = "There were $num_files new files to be downloaded, " . 
			"but warn threshold set to $warn_threshold!";
		$self->raise_alarm("WARNING: $msg");
		$self->log("[ERR] *WARNING*: $msg\n");
	}

	return 0;
}

# Drops a diagnostic file into the feed directory
sub drop_file {
	my ($self, $out_file, $source_file) = @_;

	eval {
		my $feed_path = $self->cfg->param('feed_error_dir') or die "Feed_error_dir not defined";
		mkpath($feed_path);

		my $drop_file = $self->cfg->param($source_file) or die "$source_file not defined";
		
		my $input_fh = new FileHandle $drop_file or die "Can't open $drop_file: $!";
		my @input = <$input_fh>;
		$input_fh->close();
		my $input = join("", @input);
		my $suffix = $self->cfg->param('alert_suffix') || 'txt';
		my $out_path = "$feed_path/$out_file.$suffix";
		my $output_fh = new FileHandle ">$out_path" or 
			die "Couldn't output $out_path: $!";
		print $output_fh $input;
		$output_fh->close();
	};
	$self->log("[ERR] $@") if $@;
}

sub check_vet {
	my ($self) = @_;
	my $home_dir = $self->cfg->get_home_dir();
	my $wink = $self->args('wink');
	if (!$wink && -f File::Spec->catfile($home_dir, '.cancel_ff')) {
		$self->log("Feed Fetcher is clamped! See error log for cause. " .
			"Use restore.pl utility to restore operation.");
		$self->{wd}->set_abort;
	}
}

# Raises alarm by calling alarm.pl on server
sub raise_alarm {
	my ($self, $alarm) = @_;

	my $login = $self->cfg->param('username');
	my $to_list = $self->cfg->param('alert_to');
	my $to_clause = $to_list ? "&to=$to_list" : "";

	my @servers = $self->cfg->params('alarm_server');
	foreach my $server (@servers) {
		eval {
			get("http://$server/feedfetcher/alarm.pl" . 
				"?login=$login&alarm=$alarm$to_clause") or 
				die "Couldn't connect to alarm server $server: $!";
		};
		$self->log("[ERR] $@") if $@;
	}
	$self->log("[ERR] Couldn't connect to any alarm server") if $@;
}

#sub get_timestamp {
#    my ($self, $use_usec) = @_;
#    
#    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime();
#    my ($s, $usec) = gettimeofday();
#    $year += 1900;
#    $mon += 1;
#    $mon = sprintf("%02d", $mon);
#    $mday = sprintf("%02d", $mday);
#    $hour = sprintf("%02d", $hour);
#    $min = sprintf("%02d", $min);
#    $sec = sprintf("%02d", $sec);
#    $usec = sprintf("%06.6s", $usec);
#    my $time_str = "${year}-${mon}-${mday}T${hour}${min}${sec}";
#    $time_str = "${time_str}.${usec}" if ($use_usec eq '1');
#
#    $time_str =~ s/^\s*(.*?)\s*$/$1/;
#    
#    return $time_str;
#}

1;