# $Id: DbStore.pm 3074 2016-04-07 19:51:30Z ian $

package FeedfetcherDeluxe::DbStore;

use strict;

use DBI();
use File::Path;
use File::Basename;
use File::Spec;

use Core::Configurator;
use Core::DbUtil;
use Core::XmlUtil;
use Core::Util;
use FeedfetcherDeluxe::SpecificStats;
use FeedfetcherDeluxe::PbpStore;
use Core::ActiveFixtures;
use FeedfetcherDeluxe::XtossUtil;
use FeedfetcherDeluxe::Entities::Season;
use FeedfetcherDeluxe::Entities::Publisher;
use FeedfetcherDeluxe::Entities::Display_name;
use FeedfetcherDeluxe::Entities::Affiliation;
use FeedfetcherDeluxe::Entities::Affiliation_phase;
use FeedfetcherDeluxe::Entities::Affiliations_event;
use FeedfetcherDeluxe::Entities::Person;
use FeedfetcherDeluxe::Entities::Team;
use FeedfetcherDeluxe::Entities::Team_phase;
use FeedfetcherDeluxe::Entities::Position;
use FeedfetcherDeluxe::Entities::SportsmlDocument;
use FeedfetcherDeluxe::Entities::Wagering_moneyline;
use FeedfetcherDeluxe::Entities::Wagering_total_score_line;
use FeedfetcherDeluxe::Entities::Wagering_straight_spread_line;
use FeedfetcherDeluxe::Entities::Participants_event;
use FeedfetcherDeluxe::Entities::Period;
use FeedfetcherDeluxe::Entities::Weather_condition;
use FeedfetcherDeluxe::Entities::Event;
use FeedfetcherDeluxe::Entities::Ranking;
use FeedfetcherDeluxe::Entities::Award;
use FeedfetcherDeluxe::Entities::Record;
use FeedfetcherDeluxe::Entities::Site;
use FeedfetcherDeluxe::Entities::Standing;
use FeedfetcherDeluxe::Entities::Injury_phase;
use FeedfetcherDeluxe::Entities::Events_sub_season;
use FeedfetcherDeluxe::Entities::Sports_property;
use FeedfetcherDeluxe::Entities::QueryCache;
use Conf::ActiveFixtures;
use Util::DateUtils;
use IO::Compress::Gzip qw(gzip $GzipError) ;

#my $loadflexsport = 0;
#my $loadxtoss = 1;

my $xtoss_table_prefix = '';

my %supported_code_types = undef;

my @event_fields = ('event_id', 'event_key', 'publisher_key', 'site_key',
	'last_update', 'start_date_time', 'start_date_time_local', 'site_alignment',
	'event_status', 'season_type', 'season_key', 'event_name'
);


sub new {
    my ($class, $cfg, $log, $args) = @_;
	my $self = {};
	bless $self, $class;
	$self->{'cfg'} = $cfg;
	$self->{'log'} = $log;
	$self->{'args'} = $args;
	$self->initialize();
	return $self;
}

sub log {
	my ($self, $msg) = @_;
	if ($msg) {
		$self->{'log'}->log($msg);
	} else {
		return $self->{'log'};
	}
}

sub reconnect {
	my ($self) = @_;
	$self->init_db() if !$self->{'dbh'};
}

sub disconnect {
	my ($self) = @_;
	$self->{'dbh'}->disconnect if $self->{'dbh'};
	$self->{'dbh'} = undef;
}

sub restart {
	my ($self) = @_;
	$self->disconnect;
	$self->reconnect;
}

sub version {
	my ($cfg) = @_;
	# Connect to the database.
	my $dbh = Core::DbUtil::connect($cfg);

	# flexsql: meta / select
	my $sql = q{ SELECT db_version FROM meta WHERE active=1 };
	my $arr_ref = $dbh->selectrow_arrayref($sql);
	return undef if !$arr_ref;
	return $arr_ref->[0];
}

sub initialize {
	my ($self) = @_;
	my $cfg = $self->{'cfg'};
	my $log = $self->{'log'};
	$self->{xtutil};

#	print "initializing... \n";

	$self->{date_utils} = Util::DateUtils->new();
	$self->load_supported_code_types();
	# $self->{'dbh'} is available after this call
	$self->init_db;
	
	$self->{'active_events_fixtures'} =
		Core::ActiveFixtures::load($cfg, 'active-fixtures-events');

    $self->{'active_participants_event'} =
        Core::ActiveFixtures::load($cfg, 'active-fixtures-participants-events');

	$self->{'active_map_events_fixtures'} =
		Core::ActiveFixtures::load($cfg, 'active-map-events-fixtures');

	$self->{'active_players'} =
		Core::ActiveFixtures::load($cfg, 'active-fixtures-players');

	$self->{'active_injuries'} =
		Core::ActiveFixtures::load($cfg, 'active-fixtures-injuries');

	# AlanLD
	$self->{'active_league_directory'} =
		Core::ActiveFixtures::load($cfg, 'active-fixtures-league-directory');

	$self->{'active_award'} =
		Core::ActiveFixtures::load($cfg, 'active-fixtures-award');

    $self->{'active_ranking'} =
        Core::ActiveFixtures::load($cfg, 'active-fixtures-ranking');

    $self->{'active_record'} =
        Core::ActiveFixtures::load($cfg, 'active-fixtures-record');

    $self->{'active_standing'} =
        Core::ActiveFixtures::load($cfg, 'active-fixtures-standing');

	$self->{'active_result'} =
		Core::ActiveFixtures::load($cfg, 'active-fixtures-result');

    # Though subscore is handled in an entity anyway, not through this afaik
	$self->{'active_subscore'} =
		Core::ActiveFixtures::load($cfg, 'active-fixtures-subscore');

	$self->{'active_wagering'} =
		Core::ActiveFixtures::load($cfg, 'active-fixtures-wagering');

	$self->{'active_weather'} =
		Core::ActiveFixtures::load($cfg, 'active-fixtures-weather');

	$self->{'event_level_lookup'} =
		Core::ActiveFixtures::load($cfg, 'event-level-lookup');
}

sub check_reconnect {
	my ($self) = @_;

	my $elapsed = time - $self->{t0};
	$elapsed /= 60;		# in minutes
	my $target_elapsed = $self->{cfg}->param('db_connection_time') || 10;

	if ($elapsed > $target_elapsed) {
		$self->{log}->log('[DBG] Connection stale, reconnecting to db...');
		my $table_entity = FeedfetcherDeluxe::Entities::Entity->new($self->{'dbh'}, $self->{log});
		$table_entity->reset();
		$self->restart;
		$self->{log}->log('[DBG] Reconnected to database');
	}
}

sub init_db {
	my ($self) = @_;

	my $cfg = $self->{'cfg'};
	my $log = $self->{'log'};
	$self->{t0} = time;

	# Connect to the database.
	my $dbh = Core::DbUtil::connect($cfg);
	$self->{'dbh'} = $dbh;

	$self->{'stats'} = new FeedfetcherDeluxe::SpecificStats($dbh, $cfg, $log) if
		$cfg->param('db_do_stats');

	$self->{'pbp'} = FeedfetcherDeluxe::PbpStore->new($cfg, $log);
	$self->{xtutil} = new FeedfetcherDeluxe::XtossUtil($self->{dbh}, $self->{log});


	#######################################################################################
# TODO - remove - not used anywhere
	$self->{'xtoss_person_event_metadata_exists'} = $dbh->prepare(q{
		SELECT id
		FROM person_event_metadata
		WHERE person_id=?
		AND event_id=?
	});

# TODO - remove - not used anywhere
	$self->{'xtoss_person_event_metadata_insert'} = $dbh->prepare(q{
		INSERT INTO person_event_metadata (person_id, event_id, status, weight, role_id, position_id, team_id)
		VALUES (?, ?, ?, ?, ?, ?, ?)
	});

# TODO - remove - not used anywhere
	$self->{'xtoss_position_exists'} = $dbh->prepare(q{
		SELECT id
		FROM positions
		WHERE affiliation_id=?
	});

# TODO - remove - not used anywhere
	$self->{'xtoss_position_insert'} = $dbh->prepare(q{
		INSERT INTO positions (affiliation_id, abbreviation)
		VALUES (?, ?)
	});

	# xtossql: publishers / insert
	$self->{'xtoss_publisher_insert'} = $dbh->prepare(q{
		INSERT INTO publishers (
			publisher_key,
			publisher_name
		)
		VALUES (?, ?)
	});

	# xtossql: document_fixtures / insert
	$self->{'xtoss_document_fixtures_insert'} = $dbh->prepare(q{
		INSERT INTO document_fixtures (
			fixture_key,
			publisher_id,
			name,
			document_class_id
		)
		VALUES (?, ?, ?, ?)
	});

	# xtossql: documents / insert
	$self->{'xtoss_doc_insert'} = $dbh->prepare(q{
		INSERT INTO documents (
			doc_id,
			date_time,
			title,
			language,
			publisher_id,
			source_id,
			document_fixture_id,
			revision_id,
			priority,
			stats_coverage,
			db_loading_date_time
		)
		VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, CURRENT_TIMESTAMP)
		});

	# xtossql: contents / insert
	$self->{'xtoss_document_contents_insert'} = $dbh->prepare(q{
		INSERT INTO document_contents
		(document_id, abstract, abstract_blob, sportsml, sportsml_blob) VALUES (?, ?, ?, ?, ?)
		});

	# xtossql: document_classes / get ID of most recently created row for xtoss_document_classes
	$self->{'xtoss_latest_document_classes'} = $dbh->prepare(q{
		SELECT max(id) AS id from document_classes
		});

# TODO - remove - not used anywhere
	# xtossql: events / insert
	$self->{'xtoss_event_insert'} = $dbh->prepare(q{
		INSERT INTO events (
			event_key,
			publisher_id,
			start_date_time,
			site_id,
			site_alignment,
			event_status,
			duration,
			attendance,
			last_update
		)
		VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)
		});

	# xtossql: events / update
	$self->{'xtoss_event_update'} = $dbh->prepare(q{
		UPDATE events SET
			start_date_time=?,
			site_id=?,
			site_alignment=?,
			event_status=?,
			duration=?,
			attendance=?,
			last_update=?
		WHERE id=?
		});

	# xtossql: events_documents / insert
	$self->{'xtoss_event_document_insert'} = $dbh->prepare(q{
		INSERT INTO events_documents (event_id,document_id)
		VALUES (?, ?)
		});

	# xtossql: events_documents / delete
	$self->{'xtoss_event_document_delete'} = $dbh->prepare(q{
		DELETE FROM events_documents
		where document_id = ?
		});

# TODO - remove - not used anywhere (this function moved to Event.pm)
	# xtossql: events / select with recency
	$self->{xtoss_later_events_select} = $dbh->prepare(q{
		SELECT id
		FROM events
		WHERE id=?
		AND last_update > ?
		});

# TODO - remove - not used anywhere (this function moved to Event.pm)
	# xtossql: events / id select
	$self->{'xtoss_event_exists'} = $dbh->prepare(q{
		SELECT id
		FROM events
		WHERE event_key=?
		AND publisher_id=?
		});

# TODO - remove - not used anywhere
	# xtossql: events / select
	$self->{'xtoss_event_select'} = $dbh->prepare(q{
		SELECT
			id,
			event_key,
			publisher_id,
			start_date_time,
			site_id,
			site_alignment,
			event_status,
			duration, attendance
		FROM events
		WHERE id=?
		});

# TODO - remove - not used anywhere
	# xtossql: site / select
	$self->{'xtoss_site_exists'} = $dbh->prepare(q{
		SELECT id
		FROM sites
		WHERE site_key=?
		AND publisher_id=?
		});

	# xtossql: events_documents /id select
	$self->{'xtoss_event_document_exists'} = $dbh->prepare(q{
		SELECT event_id
		FROM events_documents
		WHERE event_id=?
		AND document_id=?
		});

# TODO - remove - not used anywhere
	# xtossql: sites / insert
	$self->{'xtoss_site_insert'} = $dbh->prepare(q{
		INSERT INTO sites (site_key, publisher_id, location_id)
		VALUES (?, ?, ?)
		});

	# xtossql: documents / update
	$self->{'xtoss_doc_update'} = $dbh->prepare(q{
		UPDATE documents
		SET date_time=?,
		title=?,
		doc_id=?,
		publisher_id=?,
		source_id=?,
		revision_id=?,
		document_fixture_id=?,
		language=?,
		priority=?,
		stats_coverage=?,
		db_loading_date_time= CURRENT_TIMESTAMP
		WHERE id=?
		});

	# xtossql: document_contents / update
	$self->{'xtoss_document_contents_update'} = $dbh->prepare(q{
		UPDATE document_contents SET abstract=?, abstract_blob=?, sportsml=?, sportsml_blob=?
		WHERE document_id=?
		});

# TODO - switch to using and remove this
	# xtossql: publishers / update
	$self->{'xtoss_publisher_exists'} = $dbh->prepare(q{
		SELECT id
		FROM publishers WHERE publisher_key=?
		});

	# xtossql: document_fixtures / select
	$self->{'xtoss_document_fixture_exists'} = $dbh->prepare(q{
		SELECT id
		FROM document_fixtures WHERE publisher_id=? AND fixture_key=?
		});

	# xtossql: document_classes, document_fixtures / select
	$self->{'xtoss_document_class_exists'} = $dbh->prepare(q{
		SELECT document_classes.id
		FROM document_classes, document_fixtures
		WHERE document_fixtures.publisher_id=?
		AND document_classes.name=?
		AND document_fixtures.document_class_id=document_classes.id
		});

	# xtossql: document_classes / insert
	$self->{'xtoss_document_class_insert'} = $dbh->prepare(q{
		INSERT INTO document_classes (name)
		VALUES (?)
		});

	# xtossql: documents / select
	$self->{'xtoss_doc_exists'} = $dbh->prepare(q{
		SELECT id
		FROM documents WHERE doc_id=?
		});

	# xtossql: latest_revisions / delete
	$self->{'xtoss_latest_revisions_delete'} = $dbh->prepare(q{
		DELETE FROM latest_revisions
		WHERE latest_document_id=?
		});

	# xtossql: latest_revisions / select
	$self->{'xtoss_latest_revisions_exists'} = $dbh->prepare(q{
		SELECT revision_id
		FROM latest_revisions
		WHERE revision_id=?
		});

	# xtossql: latest_revisions / insert
	$self->{'xtoss_latest_revisions_insert'} = $dbh->prepare(q{
		INSERT INTO latest_revisions (revision_id, latest_document_id)
		VALUES (?, ?)
		});

	# xtossql: latest_revisions / update
	$self->{'xtoss_latest_revisions_update'} = $dbh->prepare(q{
		UPDATE latest_revisions
		SET latest_document_id=?
		WHERE revision_id=?
		});

	# xtossql: document_fixtures_events / select
	$self->{later_xtoss_document_fixtures_events_select} = $dbh->prepare(q{
		SELECT id
		FROM document_fixtures_events
		WHERE event_id=? AND document_fixture_id=? AND last_update > ?
		});

	# xtossql: document_fixtures_events / insert
	$self->{'xtoss_document_fixtures_events_insert'} = $dbh->prepare(q{
		INSERT INTO document_fixtures_events
		(event_id, document_fixture_id, latest_document_id, last_update)
		VALUES (?, ?, ?, ?)
		});

	# xtossql: document_fixtures_events / delete
	$self->{'xtoss_document_fixtures_events_delete'} = $dbh->prepare(q{
		DELETE FROM document_fixtures_events
		WHERE event_id=? AND document_fixture_id=?
		});

    # TODO - delete sub and reference to it - flexsport specific
	$self->prepare_doc_sports_content_codes($dbh);
	$self->prepare_event_content($dbh);
}

sub prepare_doc_sports_content_codes {
	my ($self, $dbh) = @_;

	# xtossql: teams_documents / delete
	$self->{'xtoss_teams_documents_delete'} = $dbh->prepare(q{
		DELETE FROM teams_documents
		WHERE document_id = ?
		});

	# xtossql: teams / select
	$self->{'xtoss_teams_select'} = $dbh->prepare(q{
		SELECT id
		FROM teams
		WHERE team_key=? AND publisher_id=?
		});

	# xtossql: persons_documents / delete
	$self->{'xtoss_persons_documents_delete'} = $dbh->prepare(q{
		DELETE FROM persons_documents
		WHERE document_id = ?
		});

	# xtossql: persons / select
	$self->{'xtoss_persons_select'} = $dbh->prepare(q{
		SELECT id
		FROM persons
		WHERE person_key=? AND publisher_id=?
		});

	# xtossql: persons_documents / insert
	$self->{'xtoss_persons_documents_insert'} = $dbh->prepare(q{
		INSERT INTO persons_documents (person_id, document_id)
		VALUES (?, ?)
		});

	# xtossql: persons / insert
	$self->{'xtoss_persons_insert'} = $dbh->prepare(q{
		INSERT INTO persons (person_key, publisher_id)
		VALUES (?, ?)
		});

	# xtossql: persons / select
	$self->{'xtoss_latest_persons'} = $dbh->prepare(q{
		SELECT max(id) from persons
		});

	# xtossql: teams / select -- get ID of most recently created row
	$self->{'xtoss_latest_teams'} = $dbh->prepare(q{
		SELECT max(id) from teams
		});

	# fixit later, also add league_id, home_site_id
	# xtossql: teams / insert
	$self->{'xtoss_teams_insert'} = $dbh->prepare(q{
		INSERT INTO teams (team_key, publisher_id)
		VALUES (?, ?)
		});

# TODO - move to entity
	# xtossql: display_names / insert
	$self->{'xtoss_display_names_insert'} = $dbh->prepare(q{
		INSERT INTO display_names (language, entity_id, entity_type, full_name, first_name, last_name)
		VALUES (?, ?, ?, ?, ?, ?)
		});

	# xtossql: teams_documents / insert
	$self->{'xtoss_teams_documents_insert'} = $dbh->prepare(q{
		INSERT INTO teams_documents (team_id, document_id)
		VALUES (?, ?)
		});

	# xtossql: affiliations / select -- get ID of most recently created row
	$self->{'xtoss_latest_affiliations'} = $dbh->prepare(q{
		SELECT max(id) from affiliations
		});

# TODO - move to entity
	# fixit later, also add league_id, home_site_id
	# xtossql: affiliations / insert
	$self->{'xtoss_affiliations_insert'} = $dbh->prepare(q{
		INSERT INTO affiliations (affiliation_key, affiliation_type, publisher_id)
		VALUES (?, ?, ?)
		});

# TODO - move to entity
	# xtossql: affiliations_documents / delete
	$self->{'xtoss_affiliations_documents_delete'} = $dbh->prepare(q{
		DELETE FROM affiliations_documents
		WHERE document_id = ?
		});

# TODO - move to entity
	# xtossql: affiliations / select
	$self->{'xtoss_affiliations_select'} = $dbh->prepare(q{
		SELECT id
		FROM affiliations
		WHERE affiliation_type=? AND affiliation_key=? AND publisher_id=?
		});

# TODO - move to entity
	# xtossql: affiliations_documents / insert
	$self->{'xtoss_affiliations_documents_insert'} = $dbh->prepare(q{
		INSERT INTO affiliations_documents (affiliation_id, document_id)
		VALUES (?, ?)
		});
}

sub prepare_event_content {
	my ($self, $dbh) = @_;

# TODO - remove - not used
	# xtossql: participants_events / select
	$self->{'xtoss_participants_events_exists'}=$dbh->prepare(q{
		SELECT id
		FROM participants_events
		WHERE event_id=?
		AND participant_type=?
		AND participant_id=?
		});

# TODO - remove - not used
	# xtossql: participants_events / update
	$self->{'participants_events_update'}=$dbh->prepare(q{
		UPDATE participants_events
		SET alignment=?,
		score=?,
		event_outcome=?
		WHERE id=?
		});

# TODO - remove - not used
	# xtossql: participants_events / insert
	$self->{'participants_events_insert'}=$dbh->prepare(q{
		INSERT INTO participants_events
		(participant_type, participant_id, event_id, alignment, score, event_outcome)
		VALUES
		(?, ?, ?, ?, ?, ?)
		});

# TODO - remove - not used
	# xtossql: participants_events / delete
	$self->{'participants_events_delete'}=$dbh->prepare(q{
		DELETE FROM participants_events
		WHERE event_id=?
		});

# TODO - remove - not used
	# xtossql: periods / select
	$self->{'xtoss_period_exists'}=$dbh->prepare(q{
		SELECT id
		FROM periods
		WHERE participant_event_id=?
		AND period_value=?
		});

# TODO - remove - not used
	# xtossql: periods / insert
	$self->{'xtoss_period_insert'}=$dbh->prepare(q{
		INSERT INTO periods
		(participant_event_id, period_value, score)
		VALUES (?, ?, ?)
		});

# TODO - remove - not used
	# xtossql: periods / update
	$self->{'xtoss_period_update'}=$dbh->prepare(q{
		UPDATE periods
		SET participant_event_id=?,
		period_value=?,
		score=?
		WHERE id=?
		});

# TODO - remove - not used
	# xtossql: periods / delete
	$self->{'xtoss_periods_delete'}=$dbh->prepare(q{
		DELETE FROM periods
		WHERE participant_event_id=?
		});

	# xtossql: affiliations_events / select
	$self->{'xtoss_affiliations_events_select'} = $dbh->prepare(q{
		SELECT event_id
		FROM affiliations_events
		WHERE event_id=? AND affiliation_id=?
		});

	# xtossql: affiliations_events / insert
	$self->{'xtoss_affiliations_events_insert'} = $dbh->prepare(q{
		INSERT INTO affiliations_events (affiliation_id, event_id)
		VALUES (?, ?)
		});

	# xtossql: affiliations_events / delete
	$self->{'xtoss_affiliations_events_delete'} = $dbh->prepare(q{
		DELETE FROM affiliations_events
		WHERE event_id = ?
		});

	# get league_id(s) for this event (can be more than one!)
	# xtossql: affiliations, affiliations_events / select
	$self->{'xtoss_event_leagues_select'} = $dbh->prepare(q{
		SELECT affiliations.id
		FROM affiliations, affiliations_events
		WHERE affiliations_events.event_id=?
		AND affiliations_events.affiliation_id=affiliations.id
		AND affiliations.affiliation_type='league'
		});

# TODO - Remove this after dealing with old way of doing things in method this is used in
	# xtossql: seasons / select
	$self->{'xtoss_season_select'} = $dbh->prepare(q{
		SELECT id
		FROM seasons
		WHERE season_key=?
		AND publisher_id=?
		AND league_id=?
		});

	# xtossql: seasons / insert
	$self->{'xtoss_season_insert'} = $dbh->prepare(q{
		INSERT INTO seasons
		(season_key, publisher_id, league_id)
		VALUES (?, ?, ?)
		});

	# xtossql: sub_seasons / select
	$self->{'xtoss_sub_season_select'} = $dbh->prepare(q{
		SELECT id
		FROM sub_seasons
		WHERE sub_season_key=?
		AND season_id=?
		});

	# xtossql: sub_seasons / insert
	$self->{'xtoss_sub_season_insert'} = $dbh->prepare(q{
		INSERT INTO sub_seasons
		(sub_season_key, season_id, sub_season_type)
		VALUES (?, ?, ?)
		});

	# xtossql: events_sub_seasons / delete
	$self->{'xtoss_events_sub_seasons_delete'} = $dbh->prepare(q{
		DELETE FROM events_sub_seasons
		WHERE event_id=?
		});

    $self->{'xtoss_events_sub_seasons_select'} = $dbh->prepare(q{
        DELETE FROM events_sub_seasons
        WHERE event_id=? AND sub_season_id=?
        });

	# xtossql: events_sub_seasons / insert
	$self->{'xtoss_events_sub_seasons_insert'} = $dbh->prepare(q{
		INSERT INTO events_sub_seasons
		(event_id, sub_season_id)
		VALUES (?, ?)
		});

	# xtossql: seasons / select -- get latest ID
	$self->{'xtoss_latest_season'} = $dbh->prepare(q{
		SELECT max(id) FROM seasons
		});

	# xtossql: sub_seasons / select -- get latest ID
	$self->{'xtoss_latest_sub_season'} = $dbh->prepare(q{
		SELECT max(id) from sub_seasons
		});

}

sub load_supported_code_types {
	my ($self) = @_;

	my $codes = $self->{'cfg'}->required_param('hierarchy_levels');
	my @codes = split(/,/, $codes);
	%supported_code_types = map { Core::Util::trim($_) => 1 } @codes;
}

sub store_doc {
	my ($self, $doc, $xml, $force_load) = @_;

	# NOTE: force_load is discovered in Process.pm
	# Not really used. More relevant is the FF arg for ignore updates
#		my $force_load_txt = attribute_xpath($doc, 
#			'/sports-ff-wrapper/@force-load');
#		my $force_generic_txt = attribute_xpath($doc, 
#			'/sports-ff-wrapper/@force-generic');
#		$force_load = $force_load_txt && $force_load_txt eq 'true';

	my $ignore_last_update = ($force_load || $self->{args}->{ignore_last_update});

    my $query_cache = FeedfetcherDeluxe::Entities::QueryCache->new($self->{dbstore_module});
    
	my $active_fixtures = Conf::ActiveFixtures->new($doc);
    my $found = $active_fixtures->doc_lookup('active-fixtures-injuries');

	# Set to 0. Will be set to 1 if ignore last update flag set or a more recently stored
	# version of an event is found.
    my $more_recent_event_flag = 0;

	$self->check_reconnect;
	my $basic_data = $self->get_basic_data($doc);
	my $cfg = $self->{'cfg'};

	$self->set_active_flags($doc, $cfg);
	# mef 	setup for $basic_db_data

	my $basic_db_data = $self->store_doc_table($basic_data, $doc, $xml);

	my $document_id;
	my $publisher_id;
	my $document_fixture_id;

	$document_id = $basic_db_data->{'document_id'};
	$self->{document_id} = $document_id;
	$publisher_id = $basic_db_data->{'publisher_id'};
	$document_fixture_id = $basic_db_data->{'document_fixture_id'};

    my $event_entity = FeedfetcherDeluxe::Entities::Event->new($self->{dbh}, $self->{log});
    
    # Check events for recency conflicts (unless FF is running in a mode where last update is ignored)
    if ( $self->{'active'}->{'event'} && ! defined $ignore_last_update ) {
        $more_recent_event_flag = $event_entity->check_for_old( $doc, $basic_db_data->{'publisher_id'});
        if ($more_recent_event_flag eq '1') {
			$self->{log}->log("[ADV] More recent document for an event was found. Exiting processing for this document.");
            return $more_recent_event_flag;
        }
    }

    # Store just person information, since it will be needed later
    my $person_entity = FeedfetcherDeluxe::Entities::Person->new($self->{dbh}, $self->{log});
    $person_entity->store_persons($doc, $publisher_id);#if $self->{'active'}->{'player'};

    # Store team information
    # store_team ends up calling a team entity, which does not impinge on events, so events can be processed after this.
    $self->store_team($basic_data, $doc, $publisher_id);

    # Store events information
    # Do this after storing persons and teams, since their storage is assumed in store_event_table
    # $more_recent_event_flag should return 0 since this was already checked.
    # NOTE we pass along ignore_last_update in order to allow inserting to occur but not updating
    # unless ignore_last_update has been set.
#    my $map_event = $self->{'active'}->{'map_event'};
#    print "map event: $map_event\n";
    $more_recent_event_flag = $self->store_event_table(
    	$basic_data, $basic_db_data, $doc, $self->{'active'}->{'map_event'}, $ignore_last_update
   	) if $self->{'active'}->{'event'};
   	
    if ($self->{'active'}->{'event'}) {
    	$event_entity->store_events_sub_seasons($doc, $publisher_id);
    }

    # Call store_player, which calls Person entity player storage, including associated person tables and the events table
    $self->store_player($basic_data, $doc, $publisher_id) if $self->{'active'}->{'player'};
    
    # Call participants_events storage now that teams and players processed
    if ($self->{'active'}->{'participants_event'}) {
        my $participants_event_entity = FeedfetcherDeluxe::Entities::Participants_event->new($self->{dbh}, $self->{log});
        $participants_event_entity->store_all_from_doc($doc, $publisher_id);
    }
    
    $self->call_pbp_store($doc);

	# Try to find season data outside of event
	if ($self->{'active'}->{'event'}) {
		$self->store_season_from_sports_metadata($doc, $publisher_id);
	}

	$self->{'stats'}->store_stats($doc, $basic_data->{'fixture_key'},
		$basic_data->{'league_key'},
		$basic_data->{'publisher_key'},
		$document_id,$publisher_id,$document_fixture_id) if $self->{'stats'};

    $self->update_league_directory($basic_data, $publisher_id, $doc) if $self->{'active'}->{'league_directory'};

    # Store awards data
    # Need to add .dat file if check
    
    if ($self->{'active'}->{'award'}) {
        my $award_entity = FeedfetcherDeluxe::Entities::Award->new($self->{dbh}, $self->{log});
        $award_entity->store_all_from_doc($doc, $publisher_id);
    }
    
    if ($self->{'active'}->{'record'}) {
        my $record_entity = FeedfetcherDeluxe::Entities::Record->new($self->{dbh}, $self->{log});
        $record_entity->store_all_from_doc($doc, $publisher_id);
    }

    if ($self->{'active'}->{'standing'}) {
        my $standing_entity = FeedfetcherDeluxe::Entities::Standing->new($self->{dbh}, $self->{log});
        $standing_entity->store_all_from_doc($doc, $publisher_id);
    }

    if ($self->{'active'}->{'wagering'}) {
        my $wagering_moneyline_entity = FeedfetcherDeluxe::Entities::Wagering_moneyline->new($self->{dbh}, $self->{log});
      	$wagering_moneyline_entity->store_all_from_doc($doc, $publisher_id);

        my $wagering_straight_spread_line = FeedfetcherDeluxe::Entities::Wagering_straight_spread_line->new($self->{dbh}, $self->{log});
        $wagering_straight_spread_line->store_all_from_doc($doc, $publisher_id);

        my $wagering_total_score_line = FeedfetcherDeluxe::Entities::Wagering_total_score_line->new($self->{dbh}, $self->{log});
        $wagering_total_score_line->store_all_from_doc($doc, $publisher_id);
    }

    if ($self->{'active'}->{'ranking'}) {
        my $ranking_entity = FeedfetcherDeluxe::Entities::Ranking->new($self->{dbh}, $self->{log});
      	$ranking_entity->store_all_from_doc($doc, $publisher_id);
    }
    

	# Weather handling
	if ($self->{'active'}->{'weather'}) {
		my $weather_condition_entity = FeedfetcherDeluxe::Entities::Weather_condition->new($self->{dbh}, $self->{log});
		$weather_condition_entity->store_all_from_doc($doc, $publisher_id);
	}
	
	# Question: Do we want to update for all event-documents? If so, need to ditch the force_load etc. stuff.
    if ( $self->{'active'}->{'event'} && ! defined $ignore_last_update ) {
		# NOTE: This code won't be reached if the document has old event data
		# AND if the ignore_updates flag has not been set, so we can safely
		# just update the value now.

		# Save event last_update times now for all events in the document
		# So that those monitoring this field won't be trying to access other
		# event-related data won't be misled into thinking that an event's data
		# is available. Change made at request of fantasyplanet - ticket 1902
		# Call the event entity to get it to store the last_update times the document's events, based on
		# this document's date-time value
		$event_entity->store_event_last_updates($doc, $publisher_id);
    }
    
    # Use Event entity to scan for events to clear out
    my $event_entity = FeedfetcherDeluxe::Entities::Event->new($self->{dbh}, $self->{log});
    $event_entity->scan_for_delete($doc, $publisher_id);
    
    
	# All modules under the perl/FeedFetcherDeluxe/Entities directory have a query cache.
	# This cache needs to be invalidated after each document processing.
    $query_cache->invalidate_all();
    return $more_recent_event_flag;
}

sub set_active_flags {
	my ($self, $doc, $cfg) = @_;

	$self->{'active'}->{'event'} = Core::ActiveFixtures::doc_lookup(
		$self->{'active_events_fixtures'}, $doc, $cfg);

    $self->{'active'}->{'participants_event'} = Core::ActiveFixtures::doc_lookup(
        $self->{'active_participants_event'}, $doc, $cfg);

	$self->{'active'}->{'map_event'} = Core::ActiveFixtures::doc_lookup(
		$self->{'active_map_events_fixtures'}, $doc, $cfg);

	$self->{'active'}->{'ranking'} = Core::ActiveFixtures::doc_lookup($self->{'active_ranking'}, $doc, $cfg);
	$self->{'active'}->{'award'} = Core::ActiveFixtures::doc_lookup($self->{'active_award'}, $doc, $cfg);
    $self->{'active'}->{'record'} = Core::ActiveFixtures::doc_lookup($self->{'active_record'}, $doc, $cfg);
    $self->{'active'}->{'standing'} = Core::ActiveFixtures::doc_lookup($self->{'active_standing'}, $doc, $cfg);

	my $hi_level_list = Core::ActiveFixtures::doc_lookup($self->{'active_result'}, $doc, $cfg);

	my @hi_levels = split(/, ?/, $hi_level_list);
	foreach my $hi_level (@hi_levels) {
		$self->{active}->{result}->{$hi_level} = 1;
	}

	my $sub_hi_level_list = Core::ActiveFixtures::doc_lookup(	$self->{'active_subscore'}, $doc, $cfg);

	my @sub_hi_levels = split(/, ?/, $sub_hi_level_list);
	foreach my $sub_hi_level (@sub_hi_levels) {
		$self->{active}->{subscore}->{$sub_hi_level} = 1;
	}

	$self->{'active'}->{'wagering'} = Core::ActiveFixtures::doc_lookup(
		$self->{'active_wagering'}, $doc, $cfg);

	$self->{'active'}->{'player'} =	Core::ActiveFixtures::doc_lookup(
		$self->{'active_players'}, $doc, $cfg);

	$self->{'active'}->{'injuries'} =	Core::ActiveFixtures::doc_lookup(
		$self->{'active_injuries'}, $doc, $cfg);

	# AlanLD
	$self->{'active'}->{'league_directory'} =	Core::ActiveFixtures::doc_lookup(
		$self->{'active_league_directory'}, $doc, $cfg);

	$self->{'active'}->{'weather'} =	Core::ActiveFixtures::doc_lookup(
		$self->{'active_weather'}, $doc, $cfg);
}

sub get_basic_data {
	my ($self, $doc) = @_;


#	my ($sportsml_field_name, $xtoss_key_value) = $self->{xtutil}->get_xtoss_column_info('team-in-possession-idref', 'l.nfl.com-t.1', 1);
	my $xtoss_name = $self->{xtutil}->get_xtoss_name('team-in-possession-idref');
#	$self->log("[DBG] xtoss_name: ".$xtoss_name);

#	my ($new_column_name, $xtoss_key_value) = $self->{xtutil}->get_xtoss_column_info($xtoss_name, 1, 1);

# TEST - to find a person using the Entity infrastructure
#	my $person_entity = FeedfetcherDeluxe::Entities::Person->new($self->{dbh}, $self->{log});
#	my $test_key = $person_entity->retrieve_key_by_id(1);
#	$self->log("[DBG] test_key: ".$test_key);

	my $data = {};
    my $sports_content_node = solonode($doc, "/sports-content");
    my $path_id = required_attribute($sports_content_node, "path-id");
    my $sport_name = $path_id;
#    print "path id: $path_id\n";
    my @parts = split "\/", $sport_name;
    $sport_name = $parts[0];
#    print "sport name: $sport_name\n";

	my $metanode = $data->{'metanode'} = solonode($doc,
		"/sports-content/sports-metadata");

	my $content_codes_node = $data->{'content_codes_node'} = solonode($metanode,
		"sports-content-codes");

	$data->{'doc_id'} = required_attribute($metanode, "doc-id");
	$data->{sport_name} = $sport_name;
	$self->{doc_id} = $data->{'doc_id'};

	# OLD content code fetch
	#my $content_codes = $self->get_content_codes($content_codes_node);
	# NEW content code fetch
	my $content_code_data = $self->get_content_codes($content_codes_node);
	my $content_codes = $content_code_data->{'codes'};
	my $code_names = $content_code_data->{'code_names'};
	# $self->log("[DBG] content: $content_codes");


	$data->{'content_codes'} = $content_codes;
	$data->{'code_names'} = $code_names;

	$data->{'fixture_key'} = required_attribute($metanode, "fixture-key");

	# xtoss
	$data->{'fixture_name'} = required_attribute($metanode, "fixture-name");
	$data->{'document_class'} = required_attribute($metanode, "document-class");

	my $publisher_keys = $content_codes->{'publisher'};
	die "No publisher-key in document" if !$publisher_keys || @$publisher_keys == 0;
	$data->{'publisher_key'} = @{$publisher_keys}[0];

	# xtoss
	my $source_keys = $content_codes->{'source'};
	$data->{'source_key'} = @{$source_keys}[0];

	$self->{'log'}->log(
		"[ADV] More than one publisher in sports-content-codes. " .
		"Using first one. (" . $self->{doc_id} . ")") if
		@$publisher_keys > 1;

	$data->{'league_key'} = get_league($metanode);
#	print "league key: $data->{'league_key'}\n";
	$data->{date_time} = required_attribute($metanode, 'date-time');

	return $data;
}

sub get_league {
	my ($metanode) = @_;

	my $general_league = "l.sportsml.com.general";
	my $league_query =
		'sports-content-codes/sports-content-code[@code-type="league"][1]';

	my $league_content_node = solonode($metanode, $league_query, 1);
    if (! $league_content_node) {
        print "no league found for document\n";
    }

	return $general_league if !$league_content_node;

	my $league = required_attribute($league_content_node, "code-key");
#	print "LEAGUE!: $league\n";

	return $general_league if !$league;

	return $league;
}

sub store_doc_table {
	my ($self, $basic_data, $doc, $xml) = @_;

	# Default archiving sportsml to 1 (yes) if not defined, otherwise
	# still default to 1 (yes) unless it's been set to 0 (zero) specifically
	my $archive_sportsml = '1';
	# undef will get evaluated to zero, so only check against zero if defined
	if (defined $self->{'cfg'}->param('archive_sportsml')) {
		$archive_sportsml = ($self->{'cfg'}->param('archive_sportsml') eq '0') ? '0': '1';
	}
	my $archive_sportsml = ($self->{'cfg'}->param('archive_sportsml') eq '0') ? '0': '1';
	# Default storing sportsml and abstract in DB to 0 (no)
    my $store_sportsml_in_db = ($self->{'cfg'}->param('store_sportsml_in_db') eq '1') ? '1': '0';
	my $sportsml_blob = ($store_sportsml_in_db eq '1') ? $xml: undef;
	my $abstract_blob = ($store_sportsml_in_db eq '1') ? $self->extract_abstract($doc): undef;
	$abstract_blob = ($abstract_blob =~ /^\s*$/) ? undef: $abstract_blob;
	
    my $sport_name = undef;
    if (defined $self->{'cfg'}->param('use_sport_name_in_archive_path')) {
#        my @parts = split "\/", $basic_data->{sport_name};
#        $sport_name = $parts[0];
        $sport_name = $basic_data->{sport_name};
    }

	my $metanode = $basic_data->{'metanode'};
	my $doc_id = $basic_data->{'doc_id'};
	my $revision_id = $metanode->getAttribute('revision-id');
	$revision_id = $doc_id if !$revision_id;
	my $fixture_key = $basic_data->{'fixture_key'};

	# xtoss
	my $fixture_name = $basic_data->{'fixture_name'};
	my $document_class = $basic_data->{'document_class'};

	my $date_time = $basic_data->{date_time};
	my $sports_title_node = solonode($metanode, "sports-title");
	my $sports_title = $sports_title_node->textContent();
	$sports_title = substr $sports_title, 0, 512;
	my $language = attribute($metanode, "language");
	my $document_class = required_attribute($metanode, "document-class");
	my $stats_coverage = $metanode->getAttribute("stats-coverage");
	my $league = $basic_data->{'league_key'};
	my $abstract;
	
	if ($archive_sportsml eq '1') {
		unless ($store_sportsml_in_db eq '1') {
			$abstract = $self->write_abstract($date_time, $sport_name, $league, $document_class, $doc_id, $abstract_blob);
		}
	}

	my $sportsml;
	if ($archive_sportsml eq '1') {
		unless ($store_sportsml_in_db eq '1') {
			$sportsml = $self->write_sportsml($date_time, $sport_name, $league, $document_class, $doc_id, $xml);
		}
	}
	my $content_codes = $basic_data->{'content_codes'};
	my $code_names = $basic_data->{'code_names'};

	my $priority = get_content_code($content_codes, 'priority');
	my $publisher_key = $basic_data->{'publisher_key'};

	my $db_data = {};

	# This source stuff seems to be NOT USED
	# xtoss source-key
	my $source_key = $basic_data->{'source_key'};
	my $source_id;
	if ($source_key) {
		$source_id = $self->find_publisher($source_key);
	}


	# xtoss -- if pubkey is there already, should we update publisher-name anyway? likely. Good for corrections.
	# We could get an array of fields back, including the pub-name,
	# and hence check whether the pubname is up-to-date, without having to take the
	# hit of another sql query?

	my $publisher_id = $self->find_publisher($publisher_key);

	my $document_fixture_id = $self->find_document_fixture($publisher_id, $fixture_key, $fixture_name, $document_class);


	# xtoss -- calling of sql for document insert
	# Temporarily removed: $fixture_key,
	my $document_id = '';

	# Note: revision_id is a string from the SportsML, and is NOT an autoincrement link to xtoss_latest_revisions
    
    my $date_utils = Util::DateUtils->new();
    
    my $date_time_stamp = $date_utils->ISO8601_date2time($date_time);
    my $date_time_local = $date_utils->epoch2ISO8601($date_time_stamp);    
    my $current_time_stamp = time();
    my $current_date_time = $date_utils->get_ISO8601_date($current_time_stamp);
    my $lag = $current_time_stamp-$date_time_stamp;
#    $self->{log}->log("[ADV] $date_time:$date_time_stamp $current_date_time:$current_time_stamp");
    
    $self->{log}->log("[HI ] $basic_data->{sport_name}/$league/$document_class/$fixture_key/$date_time_local/:now=$current_date_time:lag=$lag");
    
    if ($document_id = $self->find_document_id($doc_id)) {
        $self->{'xtoss_doc_update'}->execute(
            sportsml_date2sql_date($date_time),
            $sports_title,
            $doc_id,
            $publisher_id,
            $source_id,
            $revision_id,
            $document_fixture_id,
            $language,
            $priority,
            $stats_coverage,
            $document_id
        );
        $self->{'xtoss_doc_update'}->finish;
        $self->{'xtoss_document_contents_update'}->execute($abstract, $abstract_blob, $sportsml, $sportsml_blob, $document_id);
        $self->{'xtoss_document_contents_update'}->finish;
    } else {
        $self->{'xtoss_doc_insert'}->execute(
            $doc_id,
            sportsml_date2sql_date($date_time),
            $sports_title,
            $language,
            $publisher_id,
            $source_id,
            $document_fixture_id,
            $revision_id,
            $priority,
            $stats_coverage
            );
            $self->{'xtoss_doc_insert'}->finish;
        $document_id = $self->find_document_id($doc_id);
        $self->{'xtoss_document_contents_insert'}->execute($document_id, $abstract, $abstract_blob, $sportsml, $sportsml_blob);
        $self->{'xtoss_document_contents_insert'}->finish;
    }


    my $sports_property_entity = FeedfetcherDeluxe::Entities::Sports_property->new($self->{dbh}, $self->{log});
    $sports_property_entity->process_all($metanode, 'documents', $document_id);

	$self->store_latest_revisions($revision_id, $document_id);

	$self->store_doc_content_codes($doc, $document_id, $content_codes, $code_names, $publisher_id);

	$db_data->{'document_id'} = $document_id;
	$db_data->{'publisher_id'} = $publisher_id;
	$db_data->{'document_fixture_id'} = $document_fixture_id;
	return $db_data;
}

sub write_abstract {
	my ($self, $date_time, $sport_name, $league, $doc_class, $doc_id, $content) = @_;

	return undef if !$content;
	my $path = $self->build_db_path($date_time, $sport_name, $league, $doc_class);
	$path = "$path/$doc_id-abstract.txt";
	return $self->write_content($content, $path);
}

sub write_sportsml {
	my ($self, $date_time, $sport_name, $league, $doc_class, $doc_id, $content) = @_;
    my $sport = undef;

	my $path = $self->build_db_path($date_time, $sport_name, $league, $doc_class);
#    print "path: $path sport_name: $sport_name\n";
	$path = "$path/$doc_id.xml";
#    print "path: $path\n";
	return $self->write_content($content, $path);
}

sub build_db_path {
	my ($self, $date_time, $sport_name, $league, $doc_class) = @_;

	my $date_path = Feedfetcher::Util::date_path($date_time);
#	print "$date_path s: $sport_name l: $league c: $doc_class\n";
	return ((defined $sport_name) ? "$date_path/$sport_name/$league/$doc_class": "$date_path/$league/$doc_class");
}

sub write_content {
	my ($self, $content, $path) = @_;

    my $gzip = (lc $self->{'cfg'}->param('compress_archive_files') =~ /^\s*[1]\s*$|^\s*true\s*$/i) ? 1 : 0;
	my $base = $self->{'cfg'}->required_param('base_db_dir');
	my $full_path = File::Spec->catfile($base, $path);
	my $dir = dirname($full_path);
	mkpath($dir) if !-d $dir;
	my $fh = new FileHandle ">$full_path" or die "Can't write to $full_path: $!";
	print $fh $content;
	undef $fh;
    # See http://perldoc.perl.org/IO/Compress/Gzip.html#Examples
    if ($gzip eq '1' && $full_path =~ /.*.xml/i) {
    	my $buffer;
    	gzip \$content => \$buffer or die "gzip failed: $GzipError\n";
        my $gz_fh = new FileHandle ">$full_path.gz" or die "Can't write to $full_path: $!";
        print $gz_fh $buffer;
#        gzip "<$full_path>" => "<$full_path.gz>" or die "Compression of file $full_path failed: $GzipError\n";
        unlink "$full_path";
        # Add extension to path so that when value saved in DB it will be accurate
        # Need to test this to ensure it works as expected, though it should.
        $path = $path . ".gz";
    }
	$path =~ s/\\/\//g;
	return $path;
}

# note that $document_id is an autoincrement ID (FK to xtoss_documents table),
# whereas revision_id is a string from the SportsML
sub store_latest_revisions {
	my ($self, $revision_id, $document_id) = @_;

    # TODO-mapping - verify logic equivalent
    my $latest_revisions_record_mapper = FeedfetcherDeluxe::Entities::MetaRecordMapper->new($self->{dbh}, $self->{log}, 'latest_revisions');
#    $latest_revisions_record_mapper->retrieve(
#    {
#        'latest_document_id'=>$document_id
#    }
#    );
#    $latest_revisions_record_mapper->set(
#    {
#        'revision_id'=>$revision_id
#    }
#    );
    $latest_revisions_record_mapper->retrieve(
    {
        'revision_id'=>$revision_id
    }
    );
    $latest_revisions_record_mapper->set(
    {
        'latest_document_id'=>$document_id
    }
    );

    $latest_revisions_record_mapper->save();

#	# First delete all (if any) records for this doc-id
#	$self->{'xtoss_latest_revisions_delete'}->execute($document_id);
#	my $exists = $self->{'dbh'}->selectrow_arrayref($self->{'xtoss_latest_revisions_exists'},
#		undef, $revision_id);
#	if ($exists) {
#		$self->{'xtoss_latest_revisions_update'}->execute($document_id, $revision_id);
#	} else {
#		$self->{'xtoss_latest_revisions_insert'}->execute($revision_id, $document_id);
#	}
}


# TODO - move this to Publisher entity
# xtoss -- subroutine to see if publisher exists. If so, returns its ID. If nfind_publisherfor publisher.
sub find_publisher {
	my ($self, $publisher_key) = @_;

	$self->{'xtoss_publisher_exists'}->execute($publisher_key);
	my @arr = $self->{'xtoss_publisher_exists'}->fetchrow_array();
	$self->{'xtoss_publisher_exists'}->finish;

	if ($arr[0]) {
		return @arr ? $arr[0] : undef;
	}
	else {
		# FIXIT: inserting generic publisher-name below
		$self->{'xtoss_publisher_insert'}->execute(
			$publisher_key,
			'publisher-name'
		);
		$self->{'xtoss_publisher_exists'}->execute($publisher_key);

		# my $lastid = $self->{'xtoss_publisher_exists'}->execute($publisher_key)->insertid('xtoss_publisher_exists');


		@arr = $self->{'xtoss_publisher_exists'}->fetchrow_array();
		$self->{'xtoss_publisher_exists'}->finish;

		return @arr ? $arr[0] : undef;
	}
}

# xtoss -- subroutine to see if fixture-publisher combo exists
# NOTES: Should update the fixture-name and document-class if fixture-pub combo exists already?
sub find_document_fixture {
	my ($self, $publisher_id, $fixture_key, $fixture_name, $document_class) = @_;

	$self->{'xtoss_document_fixture_exists'}->execute($publisher_id, $fixture_key);
	my @arr = $self->{'xtoss_document_fixture_exists'}->fetchrow_array();
	$self->{'xtoss_document_fixture_exists'}->finish;

	if ($arr[0]) {
		return @arr ? $arr[0] : undef;
	}
	else {
		my $document_class_id = $self->find_document_class($publisher_id, $document_class);

		$self->{'xtoss_document_fixtures_insert'}->execute(
			$fixture_key,
			$publisher_id,
			$fixture_name,
			$document_class_id
		);
		$self->{'xtoss_document_fixture_exists'}->execute($publisher_id, $fixture_key);
		@arr = $self->{'xtoss_document_fixture_exists'}->fetchrow_array();
		$self->{'xtoss_document_fixture_exists'}->finish;
		return @arr ? $arr[0] : undef;
	}
}


# xtoss -- subroutine to see if document-class exists
sub find_document_class {
	my ($self, $publisher_key, $document_class) = @_;

	$self->{'xtoss_document_class_exists'}->execute($publisher_key, $document_class);
	my @arr = $self->{'xtoss_document_class_exists'}->fetchrow_array();
	$self->{'xtoss_document_class_exists'}->finish;

	if ($arr[0]) {
		return @arr ? $arr[0] : undef;
	}
	else {
		$self->{'xtoss_document_class_insert'}->execute(
			$document_class
		);

		# NOTE: The method below for getting the id of the newly created
		# row for document_class may not be the best way. It uses max(id).
		# I do believe it's cross-platform, at least.
		$self->{'xtoss_latest_document_classes'}->execute();
		@arr = $self->{'xtoss_latest_document_classes'}->fetchrow_array();
		$self->{'xtoss_latest_document_classes'}->finish;
		return @arr ? $arr[0] : undef;
	}
}


# xtoss -- subroutine to see if doc exists, and return its ID
sub find_document_id {
	my ($self, $doc_id) = @_;

	$self->{'xtoss_doc_exists'}->execute($doc_id);
	my @arr = $self->{'xtoss_doc_exists'}->fetchrow_array();
	$self->{'xtoss_doc_exists'}->finish;
	return @arr ? $arr[0] : undef;
}

sub extract_abstract {
	my ($self, $doc) = @_;

	my @articles =
		$doc->findnodes('/sports-content/article[@format="nitf"]/nitf/body/body.head/abstract');
	if (@articles) {
		my $article = shift @articles;
		my $content = $article->textContent;
		$content =~ s/^\s+//;
		$content =~ s/\s+$//;
		
		return $content;
	} else {
		return "";
	}
}

sub get_content_codes {
	my ($self, @content_codes_nodes) = @_;

	my @sport_content_nodes = ();
	foreach my $content_codes_node (@content_codes_nodes) {
		my @current_sport_content_nodes =
			$content_codes_node->getChildrenByTagName("sports-content-code");
		push @sport_content_nodes, @current_sport_content_nodes;
	}

	my %codes = ();
	my %code_names = ();
	my $doc_id = $self->{doc_id};

	foreach my $node (@sport_content_nodes) {
		my $code_type = $node->getAttribute('code-type');
		# xtoss
		my $code_name = $node->getAttribute('code-name');
		my $code_key = $node->getAttribute('code-key');
		$self->{'log'}->log("[ADV] Required attribute 'code-type' " .
			"not defined for 'sports-content-code' ($doc_id)") if !$code_type;
		next if (!$code_type || !$code_key);
		next if ($code_type =~ /^\s*$/ || $code_key =~ /^\s*$/);
		my $list = $codes{$code_type};
		$list = [] if !defined $list;
		push @$list, $code_key;
		$codes{$code_type} = $list;

		# imarsman 2/09/07
		my $code_name_key = "$code_type.$code_key";

		$code_names{$code_name_key} = $code_name;
#		 $self->log("[DBG] code-key: $code_key");
#		 $self->log("[DBG] code-name: $code_name");
#		 $self->log("[DBG] code-name key: $code_name_key");

		# xtoss - need code here to record as part of this hash the
		# content code NAME in addition to its KEY and TYPE
		# (needed for publisher-name, in addition to other things)
	}

	# OLD retval - just the codes
	# return /%codes;
	# NEW retval - content codes plus a hash of content code names
	my %ret_val = ();
	$ret_val{'codes'}=\%codes;
	$ret_val{'code_names'}=\%code_names;

	return \%ret_val;
}

sub get_content_code {
	my ($content_codes, $type) = @_;

	return $content_codes->{$type} &&
		@{$content_codes->{$type}} ?
		lc($content_codes->{$type}->[0]) : undef;
}

# TODO - remove once data structures used in this method are traced and removed if they can be
#sub store_doc_content_codes {
#	my ($self, $doc_id, $codes) = @_;
#
#	foreach my $key (keys %$codes) {
#		next if !$supported_code_types{$key};
#		my $vals_ref = $codes->{$key};
#
#		foreach my $val (@$vals_ref) {
#			if (!$self->{'dbh'}->selectrow_arrayref($self->{'map_doc_level_select'},
#				undef, $doc_id, $key, $val) ) {
#				$self->{'map_doc_level_insert'}->execute($doc_id, $key, $val);
#				$self->{'map_doc_level_insert'}->finish;
#			}
#		}
#	}
#}

sub find_first_last_name() {
	my ($self, $doc, $context, $key) = @_;

	my ($first_name, $last_name);
	# Grab the first instance of a player's metadata. Career stats files
	# e.g. xt.fss.20070702-l.nfl.com-p.1539-player-stats-career.xml
	# can have duplicate player metadata entries, one per career year
	# We assume here that the first one is fine, which assumes year remains constant across years
	my $path = "//${context}/${context}-metadata[${context}-key='${key}']/name";
	my $name_node = firstnode($doc, $path);
	if ($name_node) {
		$first_name = attribute($name_node, 'first');
		$last_name = attribute($name_node, 'last');
	}
	return ($first_name, $last_name);
}

sub store_doc_content_codes {
	my ($self, $doc, $document_id, $codes, $code_names, $publisher_id) = @_;

    # TODO mapping - Decide if logic can be revised to just insert if not there rather than
    # clearing all then re-adding
	# clear out all old keys
	$self->{'xtoss_affiliations_documents_delete'}->execute($document_id);
	$self->{'xtoss_teams_documents_delete'}->execute($document_id);
	$self->{'xtoss_event_document_delete'}->execute($document_id);
	# pdoc	clear out old player keys
	$self->{'xtoss_persons_documents_delete'}->execute($document_id);

	foreach my $key (keys %$codes) {
		next if !$supported_code_types{$key};
		my $vals_ref = $codes->{$key};

		foreach my $val (@$vals_ref) {
			# imarsman 2/09/07
			my $code_name_key = "$key.$val";
			my $name = $code_names->{$code_name_key};
			# should insert a new affiliation if this affiliation is not present
# $self->{log}->log("[DBG] key: $key code_name_key: $code_name_key");
			if ($key eq 'team') {

                # TODO - redo this using mapper
				$self->{'xtoss_teams_select'}->execute($val, $publisher_id);
				my @arr = $self->{'xtoss_teams_select'}->fetchrow_array();
				$self->{'xtoss_teams_select'}->finish;

				if ($arr[0]) {
					$self->{'xtoss_teams_documents_insert'}->execute($arr[0], $document_id);
				}
				else {
					# insert new rows into xtoss_teams and xtoss_display_names

					$self->{'xtoss_teams_insert'}->execute($val, $publisher_id);

#					$self->{'xtoss_latest_teams'}->execute();
#					my @new_team_id = $self->{'xtoss_latest_teams'}->fetchrow_array();
#					$self->{'xtoss_latest_teams'}->finish;
                    my $new_team_id = Core::DbUtil::retrieve_max_id($self->{dbh}, "teams");
                    
                    
					my ($first_name, $last_name) = $self->find_first_last_name($doc, 'team', $val);
                    my $display_name_record_mapper = FeedfetcherDeluxe::Entities::MetaRecordMapper->new($self->{dbh}, $self->{log}, 'display_names');
                    $display_name_record_mapper->retrieve(
                    {
                        'language'=>'en-US',
                        'entity_type'=>'teams',
                        'entity_id'=>$new_team_id
                    }
                    );
                    $display_name_record_mapper->set(
                    {
                        'full_name'=>$name,
                        'first_name'=>$first_name,
                        'last_name'=>$last_name
                    }
                    );
                    $display_name_record_mapper->save();
#					$self->{'xtoss_display_names_insert'}->execute('en-US', @new_team_id->[0], 'teams', $name, $first_name, $last_name);

					# then try again
					# TODO mapper insert/update logic fix
#                    my $teams_documents_record_mapper = FeedfetcherDeluxe::Entities::MetaRecordMapper->new($self->{dbh}, $self->{log}, 'teams_documents');
#                    $teams_documents_record_mapper->retrieve(
#                    {
#                        'team_id'=>@new_team_id->[0],
#                        'document_id'=>$document_id
#                    }
#                    );
#                    $teams_documents_record_mapper->save();
#					$self->{'xtoss_teams_documents_insert'}->execute(@new_team_id->[0], $document_id);
				}
			}
			# pdoc	test for s-c-c of type "player"
			elsif ($key eq 'player') {

				$self->{'xtoss_persons_select'}->execute($val, $publisher_id);
				my @arr = $self->{'xtoss_persons_select'}->fetchrow_array();
				$self->{'xtoss_persons_select'}->finish;

				if ($arr[0]) {
					$self->{'xtoss_persons_documents_insert'}->execute($arr[0], $document_id);
				}
				else {
					# insert new rows into xtoss_persons and xtoss_display_names
					$self->{'xtoss_persons_insert'}->execute($val, $publisher_id);

#					$self->{'xtoss_latest_persons'}->execute();
#					my @new_person_id = $self->{'xtoss_latest_persons'}->fetchrow_array();
#					$self->{'xtoss_latest_persons'}->finish;
                    my $new_person_id = Core::DbUtil::retrieve_max_id($self->{dbh}, "persons");

					my ($first_name, $last_name) = $self->find_first_last_name($doc, 'player', $val);
					# fixit: add full code-name, to provide proper name of person
					$self->{'xtoss_display_names_insert'}->execute('en-US', $new_person_id, 'persons', $name, $first_name, $last_name);

					# then try again
					$self->{'xtoss_persons_documents_insert'}->execute($new_person_id, $document_id);
				}
			}
			else {				
				$self->{'xtoss_affiliations_select'}->execute($key, $val, $publisher_id);
				my @arr = $self->{'xtoss_affiliations_select'}->fetchrow_array();
				$self->{'xtoss_affiliations_select'}->finish;

				if ($arr[0]) {
					$self->{'xtoss_affiliations_documents_insert'}->execute($arr[0], $document_id);
				}
				else {
					# insert new rows into xtoss_affiliations and xtoss_display_names

					$self->{'xtoss_affiliations_insert'}->execute($val, $key, $publisher_id);

#					$self->{'xtoss_latest_affiliations'}->execute();
#					my @new_affiliation_id = $self->{'xtoss_latest_affiliations'}->fetchrow_array();
#					$self->{'xtoss_latest_affiliations'}->finish;
                    my $new_affiliation_id = Core::DbUtil::retrieve_max_id($self->{dbh}, "affiliations");

					$self->{'xtoss_display_names_insert'}->execute('en-US', $new_affiliation_id, 'affiliations', $name, undef, undef);

					# then try again
					$self->{'xtoss_affiliations_documents_insert'}->execute($new_affiliation_id, $document_id);
				}
			}
		}
	}
}


sub delete_event {
	my ($self, $event_id, $sports_event_node) = @_;

	if ($self->{'active'}->{'subscore'} && firstnode($sports_event_node,
		'descendant::player-stats/sub-score[@score]|' .
		'descendant::team-stats/sub-score[@score]')) {
		$self->{'map_event_subscore_delete'}->execute($event_id);
	}

	if ($self->{'active'}->{'award'} && firstnode($sports_event_node,
		'descendant::award')) {
		$self->{'map_event_award_delete'}->execute($event_id);
	}

	# Delete spurious old teams from map_event_result table
	if ($self->{active}->{result}) {
		my @team_keys = attributes_xpath($sports_event_node,
			'team/team-metadata/@team-key');

	# flexsql: map_event_result / delete
		if (@team_keys) {
			my $where_clause = "event_id = ? AND hi_level = ? AND " .
				join(" AND ", map { 'hi_level_key != ?' } @team_keys);
			my $sql = qq{DELETE FROM map_event_result WHERE $where_clause};
			$self->{dbh}->do($sql, undef, $event_id, 'teams', @team_keys);
		}
	}
}

sub call_pbp_store {
    my ($self, $doc) = @_;
	
    my @sports_event_nodes =
        $doc->findnodes("/sports-content//sports-event");
	
    foreach my $node (@sports_event_nodes) {
        my $event_metanode = solonode($node, "event-metadata");
        my $event_key = attribute($event_metanode, 'event-key');
        next unless defined $event_key;
        next if $event_key =~ /^\s*$/;

        $self->{pbp}->store($self->{dbh}, $node, $event_key, $self->{document_id});
    }
}

sub store_event_table {
	# mef	passing basic_db_data around, here and below
	my ($self, $basic_data, $basic_db_data, $doc, $map_fixtures, $force_load) = @_;

    my $more_recent_event_flag = 0;
	my @sports_event_nodes =
		$doc->findnodes("/sports-content//sports-event");

# Find event OR tournament nodes - Need to be ready in process_event to try both for metadata node
#    my @sports_event_nodes =
#        $doc->findnodes("/sports-content//sports-event | /sports-content//tournament");

	my $event_level_xpaths = $self->prepare_event_xpaths($doc);

	# Process each event
	foreach my $node (@sports_event_nodes) {
	   $self->trap( sub {
            $more_recent_event_flag = $self->process_event($node, $basic_data, $basic_db_data,
            $event_level_xpaths, $map_fixtures, $force_load);
		});
        # This more_recent_event logic is meant to allow the discovery that a document contains
        # older event data than that already stored to bubble up and allow the end of processing
        # for event data for the document.
        # This is the second level for this logic		
		if ($more_recent_event_flag eq '1') {
            $self->log("[ADV] more recent event found");
            return $more_recent_event_flag;
		}
	}
	# NOTE
	# Store all participants_events data AFTER storing all events, since as participants are spun through in multiple event
	# document they mey be tied to events not yet stored, leading to errors and failures.
	my $publisher_id = $basic_db_data->{'publisher_id'};
	
	# Unnecessary
#	if ($self->{'active'}->{'result'}) {
#		$self->log("[DBG] active_result");
#	   my $participants_event_entity = FeedfetcherDeluxe::Entities::Participants_event->new($self->{dbh}, $self->{log});
#	   $participants_event_entity->store_all_from_doc($doc, $publisher_id);
#	}

#    # Weather support
#    my $weather_condition_entity = FeedfetcherDeluxe::Entities::Weather_condition->new($self->{dbh}, $self->{log});
#    $weather_condition_entity->store_all_from_doc($doc, $publisher_id);
    
    return $more_recent_event_flag;
}

sub find_event_document {
	my ($self, $event_id, $document_id) = @_;

	$self->{'xtoss_event_document_exists'}->execute($event_id, $document_id);
	my @arr = $self->{'xtoss_event_document_exists'}->fetchrow_array();
	$self->{'xtoss_event_document_exists'}->finish;
	return @arr ? 1 : undef;
}

sub add_event_document {
	my ($self, $event_id, $document_id) = @_;

	$self->{'xtoss_event_document_insert'}->execute($event_id, $document_id);
}

# TODO - remove - dead code
# Find a site id by site_key
# Insert if not found
#sub xtoss_find_site {
#	my ($self, $site_key, $publisher_id) = @_;
#
##	my $site_id = undef;
##	$self->{'xtoss_site_exists'}->execute($site_key, $publisher_id);
##	my @arr = $self->{'xtoss_site_exists'}->fetchrow_array();
##	$self->{'xtoss_site_exists'}->finish;
#
##	return @arr ? @arr->[0] : undef;
#    return undef;
#}

# TODO - remove - dead code
## Add a site by key and publisher id and
## return the id of the new record or zero if it didn't work.
#sub xtoss_add_site {
#	my ($self, $site_key, $publisher_id, $location_id) = @_;
#
## TODO FIXME re-add this logic
##	$self->{'xtoss_site_insert'}->execute($site_key, $publisher_id, $location_id);
##	my $site_id = $self->xtoss_find_site($site_key, $publisher_id);
##	return $site_id || 0;
#    return undef;
#}

## TODO - REMOVE - done now in Events entity
## Test for the existence of a more recent xtoss event
## The only problem is that at the moment there's no field to test
## recency by, so at the moment it tests for the uniqueness of
## an event/publisher combination
#sub xtoss_more_recent_event_updates {
#	my ($self, $event_id, $date_time) = @_;
#
#	my $sql_date_time = sportsml_date2sql_date($date_time);
#	my $ref = $self->{dbh}->selectrow_arrayref($self->{'xtoss_later_events_select'},
#		undef, $event_id, $sql_date_time
#	);
#	return $ref ? 1 : undef;
#}

# TODO - remove - not used any more in favour of entities
## Retrieve the id for an event to test whether or not it exists
#sub xtoss_find_event {
#	my ($self, $event_key, $publisher_id) = @_;
#
#	$self->{'xtoss_event_exists'}->execute($event_key, $publisher_id);
#	my @arr = $self->{'xtoss_event_exists'}->fetchrow_array();
#	$self->{'xtoss_event_exists'}->finish;
#	return @arr ? @arr->[0] : undef;
#}

# mef	sub more_recent_event_fixture_updates
sub more_recent_event_fixture_updates {
	my ($self, $event_id, $fixture_key, $date_time) = @_;

	my $sql_date_time = sportsml_date2sql_date($date_time);
	my $ref = $self->{dbh}->selectrow_arrayref($self->{'later_event_fixtures_select'},
		undef, $event_id, $fixture_key, $sql_date_time);
	return $ref ? 1 : undef;
}

sub more_recent_xtoss_document_fixtures_events_updates {
	my ($self, $xtoss_event_id, $document_fixture_id, $date_time) = @_;

	my $sql_date_time = sportsml_date2sql_date($date_time);
	my $ref = $self->{dbh}->selectrow_arrayref($self->{'later_xtoss_document_fixtures_events_select'},
		undef, $xtoss_event_id, $document_fixture_id, $sql_date_time);
	return $ref ? 1 : undef;
}

# TODO - modify to handle tournament node related metadata
sub process_event {
	# mef	added basic_db_data
	my ($self, $node, $basic_data, $basic_db_data, $event_level_xpaths, $map_fixtures, $ignore_last_update) = @_;

   my $event_entity = FeedfetcherDeluxe::Entities::Event->new($self->{dbh}, $self->{log});
 
    my $more_recent_event_flag = 0;
	my $metanode = $basic_data->{'metanode'};
	my $doc_id = $basic_data->{'doc_id'};

	my $publisher_key = $basic_data->{'publisher_key'};
	my $fixture_key = $basic_data->{'fixture_key'};
	my $event_row = {};
	my $event_metanode = solonode_optional($node, "event-metadata");
	# Look in tournament-metadata node as well - this may fail - needs testing
	if (! defined $event_metanode) {
		$event_metanode = solonode_optional($node, "tournament-metadata");
	}
	
	# Alternate - tournament-key
	my $event_key = attribute($event_metanode, 'event-key');
    my $series_index = attribute($event_metanode, 'series-index');
	return 0 unless defined $event_key;
    return 0 if $event_key =~ /^\s*$/;
	
    # Alternate - tournament-name
	my $event_name = $event_metanode->getAttribute('event-name') || '';

	# Added for xtoss
    my $event_number = $event_metanode->getAttribute('event-number');
    my $round_number = $event_metanode->getAttribute('xts:round-number') || $event_metanode->getAttribute('round-number');
    my $broadcast_listing = $event_metanode->getAttribute('xts:broadcast-listing') || $event_metanode->getAttribute('broadcast-listing');

    # tournament-status ?
	my $event_status = $event_metanode->getAttribute('event-status') || '';
	my $site_alignment = $event_metanode->getAttribute('site-alignment');
	# TODO - Find out if the attribute search should fail over to xts namespace
	my $time_certainty = $event_metanode->getAttribute('time-certainty');
	my $site_key = $event_metanode->getAttribute('site-key');
	my $start_date_time = $event_metanode->getAttribute('start-date-time');
    my $end_date_time = $event_metanode->getAttribute('end-date-time');
	my $start_date_time_local = attribute($event_metanode, 'xts:start-date-time-local');
	    
	my $date_time = $basic_data->{date_time};

	# blank locals from basic_db_data
	my $document_id;
	my $publisher_id;
	my $document_fixture_id;
	my $event_id;

	# db_data_lookups
	$document_id = $basic_db_data->{'document_id'};
	$publisher_id = $basic_db_data->{'publisher_id'};
	$document_fixture_id = $basic_db_data->{'document_fixture_id'};

#    $event_id = $event_entity->retrieve_id_by_key($event_key, $publisher_id);

	# FIXME - xtoss - fetch or find valid location_id, attendance, duration (a character field)
	my $location_id = 0;
    my $duration = $event_metanode->getAttribute('duration');
	my $attendance;
	my $site_node = solonode_optional($event_metanode, "descendant::site");
	if (defined $site_node) {
		my $site_statistics_node = firstnode($site_node, 'site-stats[@attendance]');
		if (defined $site_statistics_node) {
			$attendance = $site_statistics_node->getAttribute('attendance');
		}
	}

	my $xtoss_site_id = undef;

    my $site_entity = FeedfetcherDeluxe::Entities::Site->new( $self->{dbh}, $self->{log} );
    $xtoss_site_id = $site_entity->store_from_doc( $event_metanode, $publisher_id );

	# Try to set start_date_time in valid sql form and log any failure to do so.
	eval {
		if (defined $start_date_time) {
			$start_date_time = $self->{date_utils}->ISO8601_date2SQL_date($start_date_time);
		}
	};
	if ($@) {
		$self->{log}->log("[ADV] Invalid start_date_time: $start_date_time for event $event_key. Clearing this value to avoid error.");
		$start_date_time = undef;
	}
    # Try to set end_date_time in valid sql form and log any failure to do so.
    eval {
        if (defined $end_date_time) {
            $end_date_time = $self->{date_utils}->ISO8601_date2SQL_date($end_date_time);
        }
    };
    if ($@) {
        $self->{log}->log("[ADV] Invalid end_date_time: $end_date_time for event $event_key. Clearing this value to avoid error.");
        $end_date_time = undef;
    }
	eval {
		if (defined $start_date_time_local) {
			$start_date_time_local = $self->{date_utils}->ISO8601_date2SQL_date($start_date_time_local, 1);
		}
	};
	if ($@) {
		$self->{log}->log("[ADV] Invalid start_date_time_local: $start_date_time_local for event $event_key. Ignoring this value.");
		$start_date_time_local = undef;
	}
    my $event_record_mapper = FeedfetcherDeluxe::Entities::MetaRecordMapper->new($self->{dbh}, $self->{log}, 'events');
    $event_record_mapper->retrieve(
    {
        'event_key'=>$event_key,
        'publisher_id'=>$publisher_id
    }
    );

    my $hr = {
    # TODO-mapping - Decide if start_date_time_local should be set all the time
       'start_date_time_local' => $start_date_time_local,
       'site_id' => $xtoss_site_id,
       'site_alignment' => $site_alignment,
#       'event_status' => $event_status,
       'duration' => $duration,
       'attendance' => $attendance,
       'event_number' => $event_number,
       'round_number' => $round_number,                
   # last_update is now handled after the document has been updated
   #                'last_update' => sportsml_date2sql_date($date_time),
       'time_certainty' => $time_certainty,
       'broadcast_listing' => $broadcast_listing,
       'series_index' => $series_index,
       'end_date_time' => $end_date_time
    };
    my $existing_status = $event_record_mapper->get_field_value('event_status');
    if (! ($existing_status eq 'deleted'))  {
#    	$self->{log}->log("[DBG] Status: $existing_status"); 
        $hr->{'event_status'} = $event_status;
    }
    else {
#        $self->{log}->log("[DBG] Status: $existing_status"); 
    }
    # start_date_time has been set earlier to either be a valid sql date or undef
    if (defined $start_date_time)  {
        $hr->{'start_date_time'} = $start_date_time;
    }
    else {
        $self->{log}->log("[ADV] No valid start_date_time found for event key $event_key. Other event data will be updated."); 
    }
    $event_record_mapper->set($hr);
    $event_id = $event_record_mapper->save();

    my $sports_property_entity = FeedfetcherDeluxe::Entities::Sports_property->new($self->{dbh}, $self->{log});
    $sports_property_entity->process_all($event_metanode, 'events', $event_id);

#	if (! $event_id) {
#        $event_id = $event_entity->create({
#        	'event_key' => $event_key,
#        	'publisher_id' => $publisher_id,
#        	'start_date_time' => $start_date_time,
#        	'start_date_time_local' => $start_date_time_local,
#        	'site_id' => $xtoss_site_id,
#        	'site_alignment' => $site_alignment,
#        	'event_status' => $event_status,
#        	'duration' => $duration,
#        	'attendance' => $attendance,
#        	'event_number' => $event_number,
#        	'round_number' => $round_number,
## last_update is now handled after the document has been updated
##        	'last_update' => sportsml_date2sql_date($date_time),
#        	'time_certainty' => $time_certainty,
#        	'broadcast_listing' => $broadcast_listing
#        });
#	}
#	else {
#		my $more_recent;
#		unless (defined $ignore_last_update) {
#			my $hr = {
#	                'site_id' => $xtoss_site_id,
#	                'site_alignment' => $site_alignment,
#	                'event_status' => $event_status,
#	                'duration' => $duration,
#	                'attendance' => $attendance,
#	                'event_number' => $event_number,
#	                'round_number' => $round_number,                
#	# last_update is now handled after the document has been updated
#	#                'last_update' => sportsml_date2sql_date($date_time),
#	                'time_certainty' => $time_certainty,
#		        	'broadcast_listing' => $broadcast_listing
#	            };
#	        # start_date_time has been set earlier to either be a valid sql date or undef
#	        if (defined $start_date_time)  {
#	            $hr->{'start_date_time'} = $start_date_time;
#	        }
#	        else {
#	            $self->{log}->log("[ADV] No valid start_date_time found for event key $event_key. Other event data will be updated."); 
#	        }
#	        $event_entity->update($hr, {'id' => $event_id});
#		}
#	}

    # TODO-mapping - Use mapper here instead of entity
    # Deal with event name in display_names if available
    my $event_name = $event_metanode->getAttribute('event-name');
    	
    if (defined $event_name) {
    	my $display_name_entity = FeedfetcherDeluxe::Entities::Display_name->new( $self->{dbh}, $self->{log} );
    	my $display_name_id = $display_name_entity->retrieve_id_by_key(
        {
            'entity_type' => 'events',
            'entity_id'   => $event_id
        });
        if ( ! defined $display_name_id ) {
           # Create display name record to store more detailed naming information about the league, division, etc.
            $display_name_entity->create(
                {
                    'entity_type' => 'events',
                    'entity_id'   => $event_id,
                    'language'    => 'en-US',
                    'full_name'   => $event_name
                }
            );
        }
        else {
            $display_name_entity->update(
                {
                    'entity_type' => 'events',
                    'entity_id'   => $event_id,
                    'language'    => 'en-US',
                    'full_name'   => $event_name
                },
                { 
                	'id' => $display_name_id 
                }
            );
        }
    }
	# mef	calling sub xtoss_document_fixtures_events

    my $more_recent_document_fixtures_events_update = $self->more_recent_xtoss_document_fixtures_events_updates($event_id, $document_fixture_id, $date_time);
#    my $logic_test = ($map_fixtures && !$self->more_recent_xtoss_document_fixtures_events_updates($event_id, $document_fixture_id, $date_time));;

#    print "map fixtures: $map_fixtures more_recent_document_fixtures_events_update: $more_recent_document_fixtures_events_update\n";
	$self->document_fixtures_events($event_id, $document_fixture_id, $date_time, $document_id) if
		($ignore_last_update ||
		($map_fixtures &&
		! defined $more_recent_document_fixtures_events_update)); #$self->more_recent_xtoss_document_fixtures_events_updates($event_id, $document_fixture_id, $date_time));

	# Add events_documents table record if needed
	my $xtoss_event_document_test = $self->find_event_document($event_id, $document_id);
	if (! defined $xtoss_event_document_test) {
		$self->add_event_document($event_id, $document_id);
	}

    # NOTE
    # This code is lifted from process_event_content
    # // BEGIN lifted code from process_event_content (now deleted)
    $self->{'xtoss_affiliations_events_delete'}->execute($event_id);
    # Process content codes like league, season
    my $content_codes = $self->process_content_codes($event_key, $node, $event_id, $publisher_id);
    # Add extra hierarchy levels (players, teams handled elsewhere) from event itself
    $self->process_extra_levels($event_key, $node, $event_level_xpaths, $event_id, $publisher_id);
    # // END lifted code from process_event_content (now deleted)

    return $more_recent_event_flag;
}

sub store_team {
    my ($self, $basic_data, $doc, $publisher_id) = @_;

    my $team_entity = FeedfetcherDeluxe::Entities::Team->new($self->{dbh}, $self->{log});
    $team_entity->store_all_from_doc($doc, $publisher_id);
}

sub store_player {
	my ($self, $basic_data, $doc, $publisher_id) = @_;

    my $person_entity = FeedfetcherDeluxe::Entities::Person->new($self->{dbh}, $self->{log});
    $person_entity->store_all_from_doc($doc, $publisher_id);
}


# AlanLD
sub update_league_directory {
	my ($self, $basic_data, $publisher_id, $doc) = @_;
	

	my $affiliation_phase_entity = FeedfetcherDeluxe::Entities::Affiliation_phase->new($self->{dbh}, $self->{log});
	
	my @group_nodes = nodes($doc, '//statistic/group');

	foreach my $group_node (@group_nodes) {
		$affiliation_phase_entity->process_league_directory_group($group_node, $publisher_id)
	}
}

sub _get_name {
	my ($self, $name_node) = @_;

	my ($first, $middle, $last, $full, $nickname) = (
		$name_node->getAttribute('first'),
		$name_node->getAttribute('middle'),
		$name_node->getAttribute('last'),
		$name_node->getAttribute('full'),
		$name_node->getAttribute('nickname')
	);

	if (!$full) {
		my @names = ();
		push @names, $first if $first;
		push @names, $middle if $middle;
		push @names, $last if $last;
		$full = join(" ", @names);
	}

	return $full ? ( $first, $middle, $last, $full, $nickname ) : undef;
}

# TEMP
sub convert2sportsmldate {
	my ($date) = @_;

	$date =~ m{(\d+)/(\d+)/(\d+)} or return $date;
	my ($month, $day, $year) = ($1, $2, $3);
	return sprintf("%04d%02d%02dT000000+0000", $year, $month, $day);
}

sub _player_exists {
	my ($self, $player_key, $publisher_key) = @_;

	my $arr_ref = $self->{'dbh'}->selectrow_arrayref($self->{'refdyn_player_exists'},
		undef, $player_key, $publisher_key);
	return $arr_ref;
}

# mef	sub map_event_fixture
# TODO - remove if obsolete
sub map_event_fixture {
	my ($self, $event_id, $fixture_key, $date_time, $doc_id) = @_;

	my $sql_date_time = sportsml_date2sql_date($date_time);
	$self->{"map_event_fixture_delete"}->execute($event_id, $fixture_key);
	$self->{"map_event_fixture_insert"}->execute($event_id, $fixture_key,
		$doc_id, $sql_date_time);
}

sub document_fixtures_events {
	my ($self, $event_id, $document_fixture_id, $date_time, $document_id) = @_;

	my $sql_date_time = sportsml_date2sql_date($date_time);
#	$self->{"xtoss_document_fixtures_events_delete"}->execute($event_id, $document_fixture_id);
#	$self->{"xtoss_document_fixtures_events_insert"}->execute($event_id, $document_fixture_id, $document_id, $sql_date_time);
#    print "dfe!\n";

    my $document_fixtures_events_record_mapper = FeedfetcherDeluxe::Entities::MetaRecordMapper->new($self->{dbh}, $self->{log}, 'document_fixtures_events');
        $document_fixtures_events_record_mapper->retrieve(
        {
            'event_id'=>$event_id,
            'document_fixture_id'=>$document_fixture_id
        }
        );
        $document_fixtures_events_record_mapper->set(
        {
            'latest_document_id'=>$document_id,
            'last_update'=>$sql_date_time
        }
    );
    $document_fixtures_events_record_mapper->save();

}

# TODO Move this into season entity
sub store_season {
	my ($self, $node, $publisher_id, $season_key, $sub_season_key) = @_;

    my $affiliation_entity = FeedfetcherDeluxe::Entities::Affiliation->new($self->{dbh}, $self->{log});

    if (!$season_key) {
    	$self->{log}->log("[DBG] No season key available, leaving store_season in DbStore.pm");
    	return;
    }
    if ($season_key !~ /\d+/) {
        $self->{log}->log("[DBG] Season key not a valid integer, leaving store_season in DbStore.pm");
        return;
    }
    

	my $league_key;
	my $league_id;
	my $league_node = firstnode($node, "//sports-content-code[\@code-type='league']");

	return if ! $league_node;
	$league_key = attribute($league_node, 'code-key');
	
	$league_id = $affiliation_entity->retrieve_id_by_key({'affiliation_key'=>$league_key, 'affiliation_type'=>'league', 'publisher_id'=>$publisher_id});

	$sub_season_key =~ s/-/_/g;
	my $sub_season_type = $sub_season_key;

	$sub_season_key = $season_key."_".$sub_season_type;

	my $local_season_id= $self->{xtutil}->get_season_id($season_key, $publisher_id, $league_id);
	my $local_sub_season_id;

	if ($local_season_id) {
		$local_sub_season_id = $self->{xtutil}->get_sub_season_id($sub_season_key, $local_season_id);

		if (! $local_sub_season_id) {
			# insert subseason
			$self->{'xtoss_sub_season_insert'}->execute($sub_season_key, $local_season_id, $sub_season_type);

			$local_sub_season_id = $self->{xtutil}->get_sub_season_id($sub_season_key, $local_season_id);
		}
	}
	else {
		# insert season
		$self->{'xtoss_season_insert'}->execute($season_key, $publisher_id, $league_id);

		$local_season_id= $self->{xtutil}->get_season_id($season_key, $publisher_id, $league_id);

		# insert sub_season
		$self->{'xtoss_sub_season_insert'}->execute($sub_season_key, $local_season_id, $sub_season_type);

		$local_sub_season_id = $self->{xtutil}->get_sub_season_id($sub_season_key, $local_season_id);
	}
}

# TODO Move to Season entity
sub store_season_from_sports_metadata {
	my ($self, $node, $publisher_id) = @_;


	my $league_key;
	my $league_id;
	my $season_key;
	my $sub_season_key;
	my $sub_season_type;

##	my $league_key;
#	my $league_node = firstnode($node, "//sports-content-codes/sports-content-code[\@code-type='league']");
	my $sportsml_document_entity = FeedfetcherDeluxe::Entities::SportsmlDocument->new($node, $self->{log});	
	my $league_key = $sportsml_document_entity->get_content_code_key('league');
#	print "league key: $league_key\n";
#	return if ! $league_node;
	return if ! $league_key;

#	$league_key = attribute($league_node, 'code-key');

	$league_id = $self->{xtutil}->get_league_id($league_key, $publisher_id);
	my $metadata_node = firstnode($node, "//statistic/statistic-metadata");
	return if ! $metadata_node;

	my $date_coverage_type = attribute($metadata_node, 'date-coverage-value');
	if ($date_coverage_type) {
		$season_key = $date_coverage_type;
		$sub_season_type = attribute($metadata_node, 'date-coverage-type');
		$sub_season_key = $sub_season_type;
		# Normalize 'regular' to 'seson-regular' but leave things like 'post-season'
		$sub_season_key = ($sub_season_key =~ /season/) ? $sub_season_key: "season-$sub_season_key";
	}

	$self->store_season($node, $publisher_id, $season_key, $sub_season_key);
}

# Store sports-content-codes for event
sub process_content_codes {
	my ($self, $event_id, $event_node, $xtoss_event_id, $publisher_id) = @_;

	my $query = q{ancestor-or-self::*/*/sports-content-codes};
	my @sports_content_codes = nodes($event_node, $query);

	my $content_code_data = $self->get_content_codes(@sports_content_codes);
	my $content = $content_code_data->{'codes'};
	my $content_names = $content_code_data->{'code_names'};


		# mel 	xtoss version: iterate thru content codes, and write to map_event_level

        my $affiliations_event_entity = FeedfetcherDeluxe::Entities::Affiliations_event->new($self->{dbh}, $self->{log});

		foreach my $hi_level (keys %$content) {
			next if !$supported_code_types{$hi_level};

			# pdoc	fix to avoid adding new affiliations for players and teams
			next if ($hi_level eq 'team');
			next if ($hi_level eq 'player');

			my $values_ref = $content->{$hi_level};

			foreach my $val (@$values_ref) {
#				$self->log("[DBG] KEY: $val");
				# imarsman 2/09/07
				my $code_name_key = "$hi_level.$val";

                my $affiliation_entity = FeedfetcherDeluxe::Entities::Affiliation->new($self->{dbh}, $self->{log});
                my $affiliation_id = $affiliation_entity->retrieve_id_by_key({'affiliation_key'=>$val, 'affiliation_type'=>$hi_level, 'publisher_id'=>$publisher_id});

#                my $affiliations_event_entity = FeedfetcherDeluxe::Entities::Affiliations_event->new($self->{dbh}, $self->{log});
                
				if ($affiliation_id) {
                    my $hashref = $affiliations_event_entity->retrieve_all_fields_by_key({'affiliation_id'=>$affiliation_id, 'event_id'=>$xtoss_event_id});
                    if (! $hashref) {
                    	$affiliations_event_entity->create({'affiliation_id'=>$affiliation_id, 'event_id'=>$xtoss_event_id});
                    }
				}
				else {
					# insert new rows into xtoss_affiliations and xtoss_display_names

                    $affiliation_id =  $affiliation_entity->create({'affiliation_key'=>$val, 'affiliation_type'=>$hi_level, 'publisher_id'=>$publisher_id});        

					# imarsman 2/09/07
					my $name = $content_names->{$code_name_key};
					$self->{'xtoss_display_names_insert'}->execute('en-US', $affiliation_id, 'affiliations', $name, undef, undef);

					# then try again
                    my $hashref = $affiliations_event_entity->retrieve_all_fields_by_key({'affiliation_id'=>$affiliation_id, 'event_id'=>$xtoss_event_id});
                    if (! $hashref) {
                        $affiliations_event_entity->create({'affiliation_id'=>$affiliation_id, 'event_id'=>$xtoss_event_id});
                    }
				}
			}
		}


		# seez	xtoss_seasons and xtoss_sub_seasons

		# get the league_id for that event
		# if season_id already exists, then get its id
		#	if sub_season exists, then get its id
		#	else insert a row for this new sub_season and get its new id
		# else insert a row for this new season and get its new id
		#	insert sub_season, and get its id
		# update events_sub_seasons for this xtoss_event_id

#		my $sub_season_type;
#
#		my $season_key = get_content_code($content, 'season');

#       NOTE: The code in the if ($season_key) block assumes that an event will be tied to the s-c-c season information, which
#             is pretty safe to do, but this is a duplication of the code called earlier with $event_entity->store_events_sub_seasons
#             which loops through the event nodes and stores season affiliations for those.
#
#		# NOTE: If there's a season_key deal with the season, else skip
#		if ($season_key) {
#
#			$sub_season_type = get_content_code($content, 'season-type') if ! $sub_season_type;
#			$sub_season_type = 'regular' if ($sub_season_type =~ /^\s*$/);
#			# Normalize sub season key so that 'regular' becomes 'season-regular' but 'post-season' stays that way
#			$sub_season_type = ($sub_season_type =~ /season/) ? $sub_season_type: "season-$sub_season_type";
#
#			# Need:
#			# $publisher_id - OK
#			# $league_id -
#			# $season_key - OK
#			# $sub_season_key - OK (sub_season_type)
#
#
#			my $sub_season_key = $season_key . "_" . $sub_season_type;
#            # Make all underscores to agree with the way things are done in Season.pm
#            $sub_season_key =~ s/-/_/g;
#
#			$self->{'xtoss_event_leagues_select'}->execute($xtoss_event_id);
#			my @event_league_list = $self->{'xtoss_event_leagues_select'}->fetchrow_array();
#			$self->{'xtoss_event_leagues_select'}->finish;
#
#			# fixit: should really iterate over set of event_leagues, instead of just assuming one
#			# fixit: may want to dump the sub_season_key value, and just use season.id + sub_season.sub_season_type
#			# TODO: Ensure that we can skip this season and sub-season handling if a sub-season is not available
#			#             E.G. Baseball event-stats-progressive.
#			#             When sub-season type not available, insertion of sub-seasons fails (which it should)
#			if (@event_league_list->[0] && $sub_season_type) {
#				my $local_event_league_id = @event_league_list->[0];
#
#				$self->{'xtoss_season_select'}->execute($season_key,$publisher_id,$local_event_league_id);
#				my @season_id_list = $self->{'xtoss_season_select'}->fetchrow_array();
#				$self->{'xtoss_season_select'}->finish;
#
#				my $local_season_id;
#				my $local_sub_season_id;
#
#				if (@season_id_list->[0]) {
#					$local_season_id = @season_id_list->[0];
#					$self->{'xtoss_sub_season_select'}->execute($sub_season_key,$local_season_id);
#					my @sub_season_id_list = $self->{'xtoss_sub_season_select'}->fetchrow_array();
#					$self->{'xtoss_sub_season_select'}->finish;
#
#					if (@sub_season_id_list->[0]) {
#						$local_sub_season_id = @sub_season_id_list->[0];
#					}
#					else {
#						# insert subseason
#						$self->{'xtoss_sub_season_insert'}->execute($sub_season_key, $local_season_id, $sub_season_type);
#
#						$self->{'xtoss_latest_sub_season'}->execute();
#						my @new_sub_season_id_list = $self->{'xtoss_latest_sub_season'}->fetchrow_array();
#						$self->{'xtoss_latest_sub_season'}->finish;
#
#						$local_sub_season_id = @new_sub_season_id_list->[0];
#					}
#				}
#				else {
#					# insert season
#					$self->{'xtoss_season_insert'}->execute($season_key, $publisher_id, $local_event_league_id);
#
#					$self->{'xtoss_latest_season'}->execute();
#					my @new_season_id_list = $self->{'xtoss_latest_season'}->fetchrow_array();
#					$self->{'xtoss_latest_season'}->finish;
#
#					$local_season_id = @new_season_id_list->[0];
#
#					# insert sub_season
#					$self->{'xtoss_sub_season_insert'}->execute($sub_season_key, $local_season_id, $sub_season_type);
#
#					$self->{'xtoss_latest_sub_season'}->execute();
#					my @new_sub_season_id_list = $self->{'xtoss_latest_sub_season'}->fetchrow_array();
#					$self->{'xtoss_latest_sub_season'}->finish;
#
#					$local_sub_season_id = @new_sub_season_id_list->[0];
#				}
##                my $events_sub_season_record_mapper = FeedfetcherDeluxe::Entities::MetaRecordMapper->new($self->{dbh}, $self->{log}, 'events_sub_seasons');
##                $events_sub_season_record_mapper->retrieve({'event_id'=>$xtoss_event_id, 'sub_season_id'=>$local_sub_season_id});
##                $events_sub_season_record_mapper->save();
#
#                my $events_sub_season_entity = FeedfetcherDeluxe::Entities::Events_sub_season->new( $self->{dbh}, $self->{log} );
#                my $events_sub_seasons_count = $events_sub_season_entity->retrieve_row_count( {'event_id'=>$xtoss_event_id, 'sub_season_id'=>$local_sub_season_id} );
#                $events_sub_season_entity->create( {'event_id'=>$xtoss_event_id, 'sub_season_id'=>$local_sub_season_id} ) if ($events_sub_seasons_count eq '0');
#
#                # TODO replace this delete/insert
#				# now have $local_sub_season_id, $xtoss_event_id
#				# delete any old events_sub_seasons for this $xtoss_event_id
##				$self->{'xtoss_events_sub_seasons_delete'}->execute($xtoss_event_id);
#				# make a new row
##				$self->{'xtoss_events_sub_seasons_insert'}->execute($xtoss_event_id,$local_sub_season_id);
#			}
#		}

	return $content;
}

sub prepare_event_xpaths {
	my ($self, $doc) = @_;

	my $val = Core::ActiveFixtures::doc_lookup($self->{'event_level_lookup'}, $doc,
		$self->{'cfg'});

	my @levels = map { trim($_) } split(/,/, $val);
	my $xpaths = [];

	foreach my $level (@levels) {
		die "Incorrectly formatted cfg line: $val - no semicolon" if
			$level !~ /;/;
		my ($hi_level, $xpath) = map { trim($_) } split(/;/, $val);
		push @$xpaths, { 'hi_level' => $hi_level, 'xpath' => $xpath };
	}

	return $xpaths;
}

# fixnow: split into two subroutines?
sub process_extra_levels {
	my ($self, $event_id, $event_node, $event_level_xpaths, $xtoss_event_id, $publisher_id) = @_;

	my $levels = $event_level_xpaths;

	# mel 	sub process_extra_levels: xtoss populations for xtoss_affiliations_events table
	foreach my $level (@$levels) {
		my $xpath = $level->{'xpath'};
		my $hi_level = $level->{'hi_level'};
		my @hi_level_keys = attributes_xpath($event_node, $xpath);

		# pdoc	fix to avoid adding new affiliations for players and teams
		next if ($hi_level eq 'team');
		next if ($hi_level eq 'player');

        my $affiliations_event_entity = FeedfetcherDeluxe::Entities::Affiliations_event->new($self->{dbh}, $self->{log});

		foreach my $hi_level_key (@hi_level_keys) {

			$self->{'xtoss_affiliations_select'}->execute($hi_level, $hi_level_key, $publisher_id);
			my @arr = $self->{'xtoss_affiliations_select'}->fetchrow_array();
			$self->{'xtoss_affiliations_select'}->finish;

			if ($arr[0]) {
				# NOTE: should only do this if that same row's not there already
				my $ref = $self->{'dbh'}->selectrow_arrayref($self->{'xtoss_affiliations_events_select'},
					undef, $xtoss_event_id, $arr[0]);
				if (!$ref) {
					$self->{'xtoss_affiliations_events_insert'}->execute($arr[0], $xtoss_event_id);
				}
			}
			else {
				# insert new rows into affiliations and display_names

				$self->{'xtoss_affiliations_insert'}->execute($hi_level_key, $hi_level, $publisher_id);

#				$self->{'xtoss_latest_affiliations'}->execute();
#				my @new_affiliation_id = $self->{'xtoss_latest_affiliations'}->fetchrow_array();
#				$self->{'xtoss_latest_affiliations'}->finish;
                my $new_affiliation_id = Core::DbUtil::retrieve_max_id($self->{dbh}, "affiliations");

                my $display_name_record_mapper = FeedfetcherDeluxe::Entities::MetaRecordMapper->new($self->{dbh}, $self->{log}, 'display_names');
                $display_name_record_mapper->retrieve(
                {
                    'language'=>'en-US',
                    'entity_type'=>'affiliations',
                    'entity_id'=>$new_affiliation_id
                }
                );
                $display_name_record_mapper->set(
                {
                    'full_name'=>'Full Affiliation Name'
                }
                );
                $display_name_record_mapper->save();
#				$self->{'xtoss_display_names_insert'}->execute('en-US', @new_affiliation_id->[0], 'affiliations', 'Full Affiliation Name', undef, undef);

                # then try again
                my $hashref = $affiliations_event_entity->retrieve_all_fields_by_key({'affiliation_id'=>$new_affiliation_id, 'event_id'=>$xtoss_event_id});
                if (! $hashref) {
                    $affiliations_event_entity->create({'affiliation_id'=>$new_affiliation_id, 'event_id'=>$xtoss_event_id});
                }

#				# then try again
#				$self->{'xtoss_affiliations_events_insert'}->execute(@new_affiliation_id->[0], $xtoss_event_id);
			}
		}
	}
}

use Carp;
use Date::Pcalc qw( Add_Delta_DHMS );

sub sportsml_date2sql_date {
	my ($date_time) = @_;

	return undef if !defined $date_time;
    my $date_utils = Util::DateUtils->new();
    my ($year, $mon, $day, $hour, $min, $sec) = $date_utils->get_UTC_time_parts($date_time);

#	$date_time =~ /^(\d\d\d\d)(\d\d)(\d\d)T(\d\d)(\d\d)(\d\d)((Z)|([+-])(\d\d)(\d\d))$/
#		or croak "date-time formatted incorrectly: $date_time";
#
#	my ($year, $mon, $day, $hour, $min, $sec, undef, $z, $plusminus, $zone_hr, $zone_min) =
#		($1, $2, $3, $4, $5, $6, $7, $8, $9, $10, $11);
#
#	if ($z eq 'Z') {
#		$plusminus = '+';
#		$zone_hr = 0;
#		$zone_min = 0;
#	}
#
#	if ($plusminus eq '+') { $zone_hr = -$zone_hr; $zone_min = -$zone_min; }
#	# Factor in timezone
#	($year, $mon, $day, $hour, $min, $sec) =
#		Add_Delta_DHMS($year, $mon, $day, $hour, $min, $sec,
#		0 ,$zone_hr ,$zone_min ,0);


	return sprintf("%04d-%02d-%02d %02d:%02d:%02d",
		$year, $mon, $day, $hour, $min, $sec);
}

sub sportsml_date2sql_date_local {
	my ($date_time) = @_;

    my $date_utils = Util::DateUtils->new();
    my ($year, $mon, $day, $hour, $min, $sec) = $date_utils->get_UTC_time_parts($date_time, 1);
#	$date_time =~ /^(\d\d\d\d)(\d\d)(\d\d)T(\d\d)(\d\d)(\d\d)/
#		or die "date-time formatted incorrectly: $date_time";
#	my ($year, $mon, $day, $hour, $min, $sec) =
#		($1, $2, $3, $4, $5, $6);
	return sprintf("%04d-%02d-%02d %02d:%02d:%02d",
		$year, $mon, $day, $hour, $min, $sec);
}

sub trap {
	my ($self, $sub, @params) = @_;

	$self->log->trap($sub, $self->{doc_id}, @params);
}

1;