# $Id: Entity.pm 3024 2013-09-20 16:38:31Z ian $

package FeedfetcherDeluxe::Entities::Entity;

use strict;

my $prepared_statements;

use Carp;
use Core::DbUtil;
use Date::Pcalc qw( Add_Delta_DHMS );
use DBI();
use DBI::Const::GetInfoType;
use Core::Log;
use Core::DbUtil;
use Core::XmlUtil;
use FeedfetcherDeluxe::Entities::SportsmlDocument;
use FeedfetcherDeluxe::Entities::QueryCache;
use Util::SharedData;
use Util::DateUtils;

=head1 NAME
Core::Entities::Entity

=head1 SYNOPSIS

	my $publisher_entity = Core::Entities::Publisher->new($self->{dbh}, $self->{log});
	my %test_h = ('publisher_key', 'test', 'publisher_name', 'test name');
	my $new_id = $publisher_entity->create(\%test_h);

	# Update with inline hash arguments
	$publisher_entity->update({'publisher_name' => 'bob'}, {'id' => 1});
	
	# Update with pre-prepared hash arguments
	my %values = ('publisher_name', 'bob');
	my %keys = ('id', 1);
	$publisher_entity->update(\%values, \%keys);

=head1 DESCRIPTION

Entity is the superclass for xtoss entities such as Publishers, Documents, Events, Teams, Players, etc.
The Entity class provides a number of basic operations common to most entities, such as retrieval
by key, retrieval by id, test for id, insert, update, and delete. Besides storage and retrieval issues,
entities inheriting from this class have available the original sportsml document, allowing them to
retrieve additional information as required.

Note that entities are not treated as full objects, but rather as handy encapsulations of actions to
take in the cotext of certain data. Entities do not have properties that get filled on create or updated
on update. They have create and update methods and any other methods useful in the context
of the entities (e.g. a Team entity might have a store_all_teams_in_document method). The goal of
this code is to encapsulate handling of database entities in a way that makes it easy to understand
and manage.

=head2 Methods

=item $object->setup_sql

Declare any sql prepared staments required in addition to the basic ones provided by the entity class.

=item $object->create(hash keys_values, string table)

Create a new record for the entity (e.g. a new Publisher), with keys matching table field names and values
being the values to store for each field. The table argument allows more flexibility for situations
such as when an Event object needs to be able to insert into an events_documents table.

=item $object->update(hash keys_values, hash update_keys_values,  string table)

Update an entity's record (e.g. the record for a Publisher) using the field_name-value hash and
update_keys_values hash (e.g. id=>1). The table argument allows updating of tables not having
the same name as the entity; for example, when a join table (e.g. events_documents) needs to
be updated.

=item $object->sub retrieve_id_by_key(key, publisher_id)

Retrieve a record for an entity by its key. The value returned will be a hash of field_name-value pairs

=item $object->sub retrieve_by_id(id)

Retrieve a record for an entity by its id. The value returned will be a hash of field_name-value pairs

=item $object->sub retrieve_id_by_id(id)

A simple call to retrieve an id (by id). This can be used as a test to see if a record exists based on its id.

=item $object->sub retrieve_id_by_key(key)

A simple call to retrieve an id by key. This can be used as a test to see if a record exists based on its key.

=item $object->sub retrieve_max_id()

Retrieve the maximum id value for a table. This is used to pass back the id for a newly created record
from the create method until a cross-db way to do so is found.

=head1 AUTHOR

Ian Marsman

=cut

sub new {
    my ($class, $dbh, $log, $table_name) = @_;

	my $self = {};
	bless $self, $class;

	$self->initialize($dbh, $log, $table_name);

	return $self;	
}

sub reset {
    my ($self) = @_;
    
#    $self->{log}->log("[DBG] Resetting prepared statements");
    $prepared_statements = {};
}

sub initialize {
	my ($self, $dbh, $log, $table_name) = @_;

	$self->{shared_data} = Util::SharedData->new();
	$self->{cfg} = $self->{shared_data}->get('cfg');
#	if (! defined $prepared_statements) {
#		$prepared_statements = {};
#	}
	
    # Set manually (allow override) to avoid looking for a new id
    # on creation for tables without id column. Should make this more
    # smarter like.
    $self->{has_id} = 1;
	unless (defined $table_name) {
        my @arr = split('::', ref($self));
        my $name = pop(@arr);
        $self->{name} = lc($name);
    	$self->{table_name} = lc($name)."s";
        # Allow tables with names like baseball_action_pitches to avoid having an 's' added
        $self->{table_name} = lc($name) if ($name =~ /es$/);
    }
    else {
    	$self->{table_name} = $table_name;
    }
    # TODO It would be good at this point to use DBI calls to verify that the table name is valid

	$self->{dbh} = $dbh;
#	$self->{sportsml_document} = $sportml_document;
	$self->{log} = $log;
#	print "$self->{table_name}\n";
    $self->initialize_local();

    $self->{query_cache} = FeedfetcherDeluxe::Entities::QueryCache->new($self->{name});

	$self->setup_common_sql();
	$self->setup_sql();
}

sub initialize_local {
    my ($self) = @_;
	
}

sub get_prepared_statement {
    my ($self, $stmt_name) = @_;
    
    unless (defined $prepared_statements->{$self->{table_name}}) {
    	$prepared_statements->{$self->{table_name}} = {};
    }
    unless (defined $prepared_statements->{$self->{table_name}}->{$stmt_name."_stmt"}) {
#    	print "creating prepared statment ".$stmt_name."_stmt for $self->{table_name}\n";
        $prepared_statements->{$self->{table_name}}->{$stmt_name."_stmt"} = $self->{dbh}->prepare($self->{$stmt_name});
    }

#    unless (defined $self->{$stmt_name."_stmt"}) {
#        $self->{$stmt_name."_stmt"} = $self->{dbh}->prepare($self->{$stmt_name});
#    }
    return $prepared_statements->{$self->{table_name}}->{$stmt_name."_stmt"};
}

sub setup_common_sql {
	my ($self) = @_;
	
    $self->{'retrieve_max_id'} = qq^
        SELECT max(id) AS id
        FROM $self->{table_name}
    ^;
    $self->{'retrieve_max_id_stmt'} = undef;
    $self->{'retrieve_id_by_id'} = qq^
        SELECT id
        FROM $self->{table_name} WHERE id=?
    ^;
    $self->{'retrieve_id_by_id_stmt'} = undef;
    # Note: Not viable for all tables
    $self->{'retrieve_id_by_key'} = qq^
        SELECT id
        FROM $self->{table_name} 
        WHERE $self->{name}_key=?
        AND publisher_id=?
    ^;
    $self->{'retrieve_key_by_key_stmt'} = undef;
    # Note: Not viable for all tables
    $self->{'retrieve_key_by_id'} = qq^
        SELECT $self->{name}_key
        FROM $self->{table_name} WHERE id=?
    ^;
    # Note: Not viable for all tables
    $self->{'retrieve_by_id'} = qq^
        SELECT *
        FROM $self->{table_name} WHERE id=?
    ^;
    $self->{'retrieve_by_id_stmt'} = undef;
    # Note: Not viable for all tables
    $self->{'retrieve_by_key'} = qq^
        SELECT *
        FROM $self->{table_name} WHERE $self->{name}_key=?
    ^; 
    $self->{'retrieve_by_key_stmt'} = undef;
#	$self->{'retrieve_max_id'} = $self->{dbh}->prepare(qq{
#		SELECT max(id) AS id
#		FROM $self->{table_name}
#	});
#	$self->{'retrieve_id_by_id'} = $self->{dbh}->prepare(qq{
#		SELECT id
#		FROM $self->{table_name} WHERE id=?
#	});
#	# Note: Not viable for all tables
#	$self->{'retrieve_id_by_key'} = $self->{dbh}->prepare(qq{
#		SELECT id
#		FROM $self->{table_name} 
#		WHERE $self->{name}_key=?
#		AND publisher_id=?
#	});
#	# Note: Not viable for all tables
#	$self->{'retrieve_key_by_id'} = $self->{dbh}->prepare(qq{
#		SELECT $self->{name}_key
#		FROM $self->{table_name} WHERE id=?
#	});
#	# Note: Not viable for all tables
#	$self->{'retrieve_by_id'} = $self->{dbh}->prepare(qq{
#		SELECT *
#		FROM $self->{table_name} WHERE id=?
#	});
#	# Note: Not viable for all tables
#	$self->{'retrieve_by_key'} = $self->{dbh}->prepare(qq{
#		SELECT *
#		FROM $self->{table_name} WHERE $self->{name}_key=?
#	});	
}

sub setup_sql {
	my ($self) = @_;
}

sub query_cache {
    my ($self) = @_;
    
    return $self->{query_cache};
}

# Ascertain the database type (e.g. sqlserver, mysql, postgresql)
sub db_type {
    my ($self) = @_;

    return $self->{dbh}->get_info( $GetInfoType{SQL_DBMS_NAME} )
      || $self->{dbh}->{Driver}->{Name};
}

# Test to see if the current database connection is to a MySQL database
sub mysql {
    my ($self) = @_;

    my $db_type = $self->db_type();
    return $db_type =~ /mysql/i;
}

# Test to see if the current database is PostgreSQL
# Function tested - 2007.01.11 - imarsman
sub postgresql {
    my ($self) = @_;

    my $db_type = $self->db_type();
    return $db_type =~ /PostgreSQL/i;
}

sub sqlserver {
    my ($self) = @_;

    my $db_type = $self->db_type();

    return $db_type =~ /sql ?server/i;
}

sub retrieve_max_id {
    my ($self) = @_;

    my $id;
    
    if ($self->{has_id}) {
    # perhaps for SQL Server:
    # http://coding.derkeiler.com/Archive/Perl/perl.dbi.users/2005-04/msg00387.html
    # select @@identity
        if ($self->mysql() || $self->postgresql()) {
            # http://cipri4ph.wordpress.com/2008/04/05/perl-dbi-last-insert-id/
    #       $id = $dbh->{q{mysql_insertid}};
            $id = $self->{dbh}->last_insert_id(undef, undef, $self->{table_name}, "id");
#            print "id: $id for table: $self->{table_name} in Entity.pm\n";
        }
        else {
            my $stmt = $self->{dbh}->prepare("SELECT max(id) as id FROM ".$self->{table_name});
            $stmt->execute();
            my $hashref = $stmt->fetchrow_hashref();
            $stmt->finish;
            $id = $hashref->{id};
        }
    }
    return $id
}

#sub retrieve_max_id {
#	my ($self) = @_;
#
#    return undef if $self->{has_id} != 1;
#    my $id;
#    
##    print "table name: ".$self->{table_name}."\n";
#	my $stmt_name = 'retrieve_max_id';
#	if ($self->{dbh}->{Driver}{Name} =~ /mysql/i) {
#		# http://cipri4ph.wordpress.com/2008/04/05/perl-dbi-last-insert-id/
##		$id = $dbh->{q{mysql_insertid}};
#        $id = $dbh->->last_insert_id();
#	}
#	else {
#    	$self->get_prepared_statement('retrieve_max_id')->execute();
#	   my $hashref = $self->get_prepared_statement($stmt_name)->fetchrow_hashref();
#	   $self->get_prepared_statement($stmt_name)->finish;
#	   $id = $hashref->{id};
#	}
#	return $id
#}

sub get_dbh {
	my ($self) = @_;
	
	return $self->{dbh};
}

sub get_log {
    my ($self) = @_;
    
    return $self->{log};
}

sub get_table_name {
    my ($self) = @_;
    
    return $self->{table_name};
}

sub set_table_name {
    my ($self, $table_name) = @_;
    
    $self->{table_name} = $table_name;
    # Prepared statements need to be re-done
    $self->setup_common_sql();
}

# Generic create method.
# Argument is a hash of fieldname-value pairs
# Field names must be valid table field names
sub create {
	my ($self, $keys_values, $table) = @_;

	$table ||= $self->{table_name};
	my @keys = keys %$keys_values;
#	print "keys: @keys\n";

	# Screen out those keys and values whose values are empty strings 
	#$self->{log}->log("[DBG] ***** keys: @keys");
	#my @test_values = values %$keys_values;
	#$self->{log}->log("[DBG] ***** values: @test_values");

	@keys = grep {$keys_values->{$_} !~ /^\s*$/} @keys;
	my @values = map {$keys_values->{$_}} @keys;

#	print "keys: @keys\n";
	# Original keys and values derivation
	#%$keys_values = grep {} ;
	#@keys = keys %$keys_values;
	#my @values = values %$keys_values;
	
	# TODO check if we should be inserting only those fields that have a value
	
	my $insert_fields_string = join(', ', @keys);
	my $insert_values_placeholders = join(', ', map { '?' } @values);
	my $sql = "INSERT INTO $table ($insert_fields_string) " . "VALUES ($insert_values_placeholders)";
#$self->{log}->log("[DBG] ***** keys: @keys");
#$self->{log}->log("[DBG] ***** values: @values");
#$self->{log}->log("[DBG] ***** SQL: $sql");
	my $sth = $self->{dbh}->prepare($sql);
    # Attem to insert the record. If the insert fails, recover and report the error trapped.
	eval {
	    $sth->execute(@values);
        $sth->finish;
	};
	if($@){
        $sth->finish;
        $self->{log}->log("[DBG] Failed insert: $@. Document processing continuing.");
        return undef;
    }
	
	return $self->retrieve_max_id();
}

# Generic update method
# Argumentss are a list of fieldname-value pairs, the fieldname to update on, and the value of that field
# Field names must be valid for table
sub update {
	my ($self, $kv, $ukv, $table) = @_;

	my %keys_values = %$kv;

	# Screen out those keys and values whose values are empty strings 
	my @key_value_keys = keys %$kv;
#	print "keys: @key_value_keys\n";
	@key_value_keys = grep {$kv->{$_} !~ /^\s*$/} @key_value_keys;
	my @key_value_values = map {$kv->{$_}} @key_value_keys;
#	print "keys: @key_value_keys\n";

	my %update_keys_values = %$ukv;
	
	$table ||= $self->{table_name};

#	my @key_value_keys = keys %keys_values;
#	my @key_value_values = values %keys_values;

	my @update_key_value_keys = keys %update_keys_values;
	my @update_key_value_values = values %update_keys_values;

	# Handle keys and values to be set
	# Note that keys_values_keys has already been screened to eliminate those whose values are empty strings
	# i.e. not undef
	my $set_sub_clause = join(", ", map { "$_=?" } grep { defined $keys_values{$_} } @key_value_keys);
	my @clause_values = grep { defined $_ } @key_value_values;
	
	# Handle AND clause values (e.g. AND event_key=1 AND publisher_key=1
	# Note: Not screening update keys and values at this time since that could mask situations where we should NOT
	# be proceeding with an update
	my $and_clause = join(" AND ", map { "$_=?" } grep { defined $update_keys_values{$_} } @update_key_value_keys);
	$and_clause = "AND ".$and_clause if (scalar @update_key_value_keys >= 1);
	
	my @args = (@clause_values, @update_key_value_values);
	
	my $sql = qq{ UPDATE $table SET $set_sub_clause WHERE 1=1 $and_clause };
#    $self->{log}->log("[DBG] $sql");
	
    my $sth = $self->{dbh}->prepare($sql);
	my $rows = undef;
	# Attem to update the record. If the update fails, recover and report the error trapped.
	eval {
	    $sth->execute(@args);
        $rows = $sth->rows;
        $sth->finish;
        $self->{query_cache}->invalidate();
        if (! defined $rows) {
            return -1;
        }
	};
	if($@){
        $sth->finish;
        $self->{log}->log("[DBG] Failed update: $@. Document Processing continuing.");
        return undef;
    }
	
    return $rows;
}

sub update_allow_null_values {
    my ($self, $kv, $ukv, $table) = @_;

	# NOTE
	# NOT screening out empty string keys/values for empty keys at this time
    my %keys_values = %$kv;
    my %update_keys_values = %$ukv;
    
    $table ||= $self->{table_name};

    my @key_value_keys = keys %keys_values;
    my @key_value_values = values %keys_values;

    my @update_key_value_keys = keys %update_keys_values;
    my @update_key_value_values = values %update_keys_values;

    # Handle keys and values to be set
    my $set_sub_clause = join(", ", map { "$_=?" } @key_value_keys);
    my @clause_values = @key_value_values;
    
    # Handle AND clause values (e.g. AND event_key=1 AND publisher_key=1
    my $and_clause = join(" AND ", map { "$_=?" } grep { defined $update_keys_values{$_} } @update_key_value_keys);
    $and_clause = "AND ".$and_clause if (scalar @update_key_value_keys >= 1);
    
    my @args = (@clause_values, @update_key_value_values);
    
    my $sql = qq{ UPDATE $table SET $set_sub_clause WHERE 1=1 $and_clause };
#    $self->{log}->log("[DBG] $sql");
    
    my $sth = $self->{dbh}->prepare($sql);
    $sth->execute(@args);
    my $rows = $sth->rows;
    $sth->finish;
    $self->{query_cache}->invalidate();
    if (! defined $rows) {
        return -1;
    }
    return $rows;
}

sub delete {
    my ($self, $dkv, $table) = @_;

    my %delete_keys_values = %$dkv;
    
    $table ||= $self->{table_name};

    my @delete_key_value_keys = keys %delete_keys_values;
    my @delete_key_value_values = values %delete_keys_values;
    @delete_key_value_values = grep defined, @delete_key_value_values;

    my $key_count = @delete_key_value_keys;
    my $value_count = @delete_key_value_values;
    my $keys_values_diff = $key_count - $value_count;
 
    # If a value is missing, don't bother, this call will fail (e.g. null player key can't be used to look up a player)
    return undef unless ($keys_values_diff == 0);

    # Don't allow effectively delete all from table
   return if @delete_key_value_keys == 0;

    # Handle keys and values to be set
#    my @clause_values = grep { defined $_ } @key_value_values;
    
    # Handle AND clause values (e.g. AND event_key=1 AND publisher_key=1
    my $and_clause = join(" AND ", map { "$_=?" } grep { defined $delete_keys_values{$_} } @delete_key_value_keys);
    $and_clause = "AND ".$and_clause if (scalar @delete_key_value_keys >= 1);
    
    my @args = (@delete_key_value_values);
    
    my $sql = qq{ DELETE FROM $table WHERE 1=1 $and_clause };
    
#    $self->{log}->log("[DBG] ***** delete_key_value_keys: @delete_key_value_keys");
#    $self->{log}->log("[DBG] ***** delete_key_value_values: @delete_key_value_values");
#    $self->{log}->log("[DBG] ***** SQL: $sql @args");

    my $sth = $self->{dbh}->prepare($sql);
    $sth->execute(@args);
    my $rows = $sth->rows;
    $sth->finish;
    $self->{query_cache}->invalidate();
    if (! defined $rows) {
        return -1;
    }
    return $rows;
}

sub retrieve_row_count {
    my ($self, $arg) = @_;

    if (! ref($arg) eq "HASH") {
	
    }

    my %keys_values = %$arg;
        
    my @key_value_keys = keys %keys_values;
    my @key_value_values = values %keys_values;
    @key_value_values = grep(defined, @key_value_values);
        
    my $key_count = @key_value_keys;
    my $value_count = @key_value_values;
    my $keys_values_diff = $key_count - $value_count;
 
    # If a value is missing, don't bother, this call will fail (e.g. null player key can't be used to look up a player)
    return undef unless ($keys_values_diff == 0);
        
#        $self->{log}->log("[DBG] keys count: $key_count values count: $value_count");
#        $self->{log}->log("[DBG] keys_values_diff: $keys_values_diff");
        
# The line below was causing problems when a value was '0' abut plain grep weeded it out. 'defined' sees it
#    my $and_clause = join(" AND ", map { "$_=?" } grep {$keys_values{$_} } @key_value_keys);
    my $and_clause = join(" AND ", map { "$_=?" } grep {defined $keys_values{$_} } @key_value_keys);
    
    
    $and_clause = "AND ".$and_clause if (scalar @key_value_keys >= 1);
    
    my @args = (@key_value_values);
    
        # If the table has no key, just get all fields
    my $sql = qq{ SELECT count(*) AS count FROM $self->{table_name} WHERE 1=1 $and_clause };
#        @key_value_values
#$self->{log}->log("[DBG] ***** key_value_keys: @key_value_keys");
#$self->{log}->log("[DBG] ***** key_value_values: @key_value_values");
#$self->{log}->log("[DBG] ***** SQL: $sql");
    my $sth = $self->{dbh}->prepare($sql);

    $sth->execute(@args);    
    my $hashref = $sth->fetchrow_hashref();
    $sth->finish;
    return 0 if ! $hashref;
#    print "hashref count: $hashref->{count}\n";
    my $ret_val = $hashref->{count};
    
    return $ret_val;
}


sub retrieve_id_by_key {
    my ($self, $arg) = @_;
    my $ret_val;
    
    my $hashref;
    
    if ($self->{has_id}  eq '1') {
#    	$self->{log}->log("[DBG] ref.arg: ".ref($arg));
        if (ref($arg) eq "HASH") {
             $hashref = $self->{query_cache}->get($arg);
             if (! defined $hashref) {
                $hashref = $self->retrieve_all_fields_by_key($arg);
                $self->{query_cache}->set($arg, $hashref) if defined $hashref;
             }
             else {
#                $self->{log}->log("[DBG] CACHED id: ".$hashref->{id});
                # Uncommenting this effectively turns off caching for this case 
#                $hashref = $self->retrieve_all_fields_by_key($arg, $publisher_id);
             }
        }
        else {
            $hashref = $self->retrieve_all_fields_by_key($arg);
        }
    	
        # $self->{log}->log("[DBG] Getting 'id'");
        if (defined $hashref && $self->{has_id}  eq '1') {
            $ret_val = $hashref->{id}
        }
#        my $cache_hash = {'id' => $ret_val};
    }
    
    return $ret_val;
}

sub retrieve_all_fields_by_key_allow_null_values {
    my ($self, $arg) = @_;
#    print "allow null values\n";
    $self->retrieve_all_fields_by_key($arg, 1);    
}

# This method is confusingly named, since it does double or triple duty
# It does a straight lookup by key when a scalar is passed as arg
# e.g. retrieve_id_by_key('l.mlb.p.1036', 1)

# It does a lookup by multiple keys when an array is passed 
#    (it is assumed that multiple keys are defined in an overriden query in this case)

# e.g. retrieve_id_by_key(['season_regular', 'l.nfl.com'], 1]);
# It does a dynamic query creation with lookup parameters defined by $arg when $arg is is a hash
# e.g. retrieve_id_by_key({'id' => 3423}}, 1]);
sub retrieve_all_fields_by_key {
	my ($self, $arg, $allow_null_values) = @_;
	
#	print "arg type: ".ref($arg)."\n";
	my $stmt_name = 'retrieve_id_by_key';
    # $arg can be something like ['2007', 1] if more than one lookup key is needed
    # This is for the very rare case where there is a custom 'retrieve_id_by_key' query
    # defined for the subclass and there are more than one lookup keys to match (with ?s)
	if (ref($arg) eq "ARRAY") {
        $self->{log}->log("[DBG] !!!!! retrieve_all_fields_by_key with deprecated ARRAY arg. Values: @$arg");
#		push(@$arg, $publisher_id) if $publisher_id;
        $self->get_prepared_statement($stmt_name)->execute(@$arg);
	}
	elsif (ref($arg) eq "HASH") {
	    my %keys_values = %$arg;
		
#		if ($publisher_id) {
#            $keys_values{'publisher_id'} = $publisher_id;
#		}
				
        my @key_value_keys = keys %keys_values;
        my @key_value_values = values %keys_values;
        
#        unless (defined $allow_null_values) {
            # Filter out null values
            @key_value_values = grep defined, @key_value_values;
#        }
        # Filter out empty values
        @key_value_values = grep(! /^\s*$/, @key_value_values);
        
        my $key_count = @key_value_keys;
        my $value_count = @key_value_values;
#        print "keys: @key_value_keys values: @key_value_values\n";
        my $keys_values_diff = $key_count - $value_count;
 
        # If a value is missing, don't bother, this call will fail (e.g. null player key can't be used to look up a player)
#        $self->{log}->log("[DBG] keys_values_diff: $keys_values_diff");
        if (! defined $allow_null_values) {
            return undef unless ($keys_values_diff == 0);
        }
        
#        $self->{log}->log("[DBG] keys count: $key_count values count: $value_count");
#        $self->{log}->log("[DBG] keys_values_diff: $keys_values_diff");
        
        # Went back to 'defined' version, since zero maps to nothing without it
#        my $and_clause = join(" AND ", map { "$_=?" } grep { $keys_values{$_} } @key_value_keys);
        my $and_clause = join(" AND ", map { "$_=?" } grep {defined $keys_values{$_} } @key_value_keys);
        $and_clause = "AND ".$and_clause if (scalar @key_value_keys >= 1);

        if ($_[2]) {
#        	print "allow null values: $_[2]\n";
        	my @total_keys_count = keys %keys_values;
        	my @non_null_values_count = grep defined, values %keys_values;
#        	print "total_keys_count: @total_keys_count non_null_values_count: @non_null_values_count\n";
        	if (@total_keys_count > @non_null_values_count) {
                my $is_null_clause = join(" AND ", map { "$_ IS NULL" } grep {! defined $keys_values{$_} } @key_value_keys);
                $and_clause = $and_clause." AND ".$is_null_clause if (scalar @key_value_keys >= 1);
#                print "and clause: $and_clause\n";
        	}
        }
        
        my @args = (@key_value_values);
    
        # If the table has no key, just get all fields
        my $sql = qq{ SELECT * FROM $self->{table_name} WHERE 1=1 $and_clause };
#        @key_value_values
#$self->{log}->log("[DBG] ***** key_value_keys: @key_value_keys");
#$self->{log}->log("[DBG] ***** key_value_values: @key_value_values");
#$self->{log}->log("[DBG] ***** SQL: $sql");
        my $sth = $self->{dbh}->prepare($sql);

        $sth->execute(@args);
        my $hashref = $sth->fetchrow_hashref();
        $sth->finish;
        # If the table has no key, return all columns for the record hash
        return $hashref;
	}
	else {
#        $self->{log}->log("[DBG] ***** arg: $arg");
        return undef unless defined $arg;
        $self->get_prepared_statement($stmt_name)->execute($arg); # if ! $publisher_id;
#    	$self->get_prepared_statement($stmt_name)->execute($arg, $publisher_id) if $publisher_id;
	}
	my $hashref = $self->get_prepared_statement($stmt_name)->fetchrow_hashref();
	$self->get_prepared_statement($stmt_name)->finish;
	return $hashref;
}

sub retrieve_key_by_id {
	my ($self, $arg) = @_;
	
	my $stmt_name = 'retrieve_key_by_id';
	$self->get_prepared_statement($stmt_name)->execute($arg);
	my $hashref = $self->get_prepared_statement($stmt_name)->fetchrow_hashref();
	$self->get_prepared_statement($stmt_name)->finish;
	my $key_name = $self->{name}."_key";
	return $hashref->{$key_name};
}

sub retrieve_id_by_id {
	my ($self, $arg) = @_;
	
	my $stmt_name = 'retrieve_id_by_id';
	$self->get_prepared_statement($stmt_name)->execute($arg);
	my $hashref = $self->get_prepared_statement($stmt_name)->fetchrow_hashref();
	$self->get_prepared_statement($stmt_name)->finish;
	return $hashref->{id};
}

sub retrieve_by_id {
	my ($self, $arg) = @_;
	
	my $stmt_name = 'retrieve_by_id';
	$self->get_prepared_statement($stmt_name)->execute($arg);
	my $hashref = $self->get_prepared_statement($stmt_name)->fetchrow_hashref();
	$self->get_prepared_statement($stmt_name)->finish;
	return $hashref;
}

sub retrieve_by_key {
	my ($self, $arg) = @_;

	my $stmt_name = 'retrieve_by_key';
	$self->get_prepared_statement($stmt_name)->execute($arg);
	my $hashref = $self->get_prepared_statement($stmt_name)->fetchrow_hashref();
	$self->get_prepared_statement($stmt_name)->finish;
	return $hashref;
}

# To be removed - Same function exists in DateUtils. Need to switched where it's used in various entities to DateUtils one. 
sub sportsml_date2sql_date {
    my ($self, $date_time) = @_;

    return undef if !defined $date_time;
    my $date_utils = Util::DateUtils->new();
    my ($year, $mon, $day, $hour, $min, $sec) = $date_utils->get_UTC_time_parts($date_time);

#    $date_time =~ /^(\d\d\d\d)(\d\d)(\d\d)T(\d\d)(\d\d)(\d\d)((Z)|([+-])(\d\d)(\d\d))$/
#        or croak "date-time formatted incorrectly: $date_time";
#
#    my ($year, $mon, $day, $hour, $min, $sec, undef, $z, $plusminus, $zone_hr, $zone_min) =
#        ($1, $2, $3, $4, $5, $6, $7, $8, $9, $10, $11);
#
#    if ($z eq 'Z') {
#        $plusminus = '+';
#        $zone_hr = 0;
#        $zone_min = 0;
#    }
#
#    if ($plusminus eq '+') { $zone_hr = -$zone_hr; $zone_min = -$zone_min; }
#    # Factor in timezone
#    ($year, $mon, $day, $hour, $min, $sec) =
#        Add_Delta_DHMS($year, $mon, $day, $hour, $min, $sec,
#        0 ,$zone_hr ,$zone_min ,0);
    return sprintf("%04d-%02d-%02d %02d:%02d:%02d",
        $year, $mon, $day, $hour, $min, $sec);
}

# To be removed - Same function exists in DateUtils. Need to switched where it's used in various entities to DateUtils one. 
sub sportsml_date2sql_date_local {
    my ($self, $date_time) = @_;

    my $date_utils = Util::DateUtils->new();
    my ($year, $mon, $day, $hour, $min, $sec) = $date_utils->get_UTC_time_parts($date_time, 1);

#    $date_time =~ /^(\d\d\d\d)(\d\d)(\d\d)T(\d\d)(\d\d)(\d\d)/
#        or die "date-time formatted incorrectly: $date_time";
#    my ($year, $mon, $day, $hour, $min, $sec) =
#        ($1, $2, $3, $4, $5, $6);
    return sprintf("%04d-%02d-%02d %02d:%02d:%02d",
        $year, $mon, $day, $hour, $min, $sec);
}
1;