# $Id: MetaRecordMapper.pm 2977 2012-07-19 22:33:06Z ian $

# This package is meant only for use with the SportsDB database in FeedFetcher. It makes
# assumptions that would not be good to make otherwise, such as that tables have an 'id'
# field that is to be used when updating.

package FeedfetcherDeluxe::Entities::MetaRecordMapper;

my $tables = {};

use strict;

use Carp;
use DBI();
use Core::Log;
use Core::DbUtil;
use Core::XmlUtil;
use Util::SharedData;
use FeedfetcherDeluxe::Entities::Generic_table;
#use Benchmark ':hireswallclock';

use constant {
    TRUE => 1,
    FALSE => 0,
    CHANGED => 1,
    UNCHANGED => 0,
    INSERT => 'INSERT',
    UPDATE => 'UPDATE'
};


sub new {
    my ($class, $dbh, $log, $table_name) = @_;

    my $self = {};
    bless $self, $class;

    $self->initialize($dbh, $log, $table_name);

    return $self;   
}

sub initialize {
    my ($self, $dbh, $log, $table_name) = @_;

    $self->{shared_data} = Util::SharedData->new();
    $self->{cfg} = $self->{shared_data}->get('cfg');
    $self->{dbh} = $dbh;
    $self->{log} = $log;
    $self->{tables} = $tables;
    if (defined $table_name) {
        $self->initialize_entity($table_name);
    }
    $self->reset();
}

sub initialize_entity {
   my ($self, $table_name) = @_;

    $self->{entity} = FeedfetcherDeluxe::Entities::Generic_table->new($self->{dbh}, $self->{log}, $table_name);
    $self->{table_name} = $self->{entity}->get_table_name();
#    $self->{entity}->reset();
}

sub reset {
   my ($self, $table_name) = @_;
    
    if (defined $table_name) {
    	$self->initialize_entity($table_name);
        $self->{table_name} = $table_name;
        $self->{entity}->set_table_name($table_name);
#        $self->{entity}->reset();
    }
    $self->{record_hash} = {};
    $self->{record_lookup_fields_hash} = {};
    $self->{record_change_hash} = {};
    $self->{record_update_force_null_hash} = {};
    $self->{action} = undef;
    $self->{overwrite_nulls} = FALSE;
    $self->{record_retrieved} = FALSE;
    $self->{id} = undef;
}

#sub set_as_unchanged {
#   my ($self) = @_;
#    
##    $self->{record_change_hash} = {};
##    $self->{record_update_force_null_hash} = {};
##    $self->{action} = UPDATE;
#}

sub can_save {
    my ($self) = @_;

    return (defined $self->{action}) ? 1: undef;
}

sub get_action {
   my ($self) = @_;
   
   if ($self->{action} eq UPDATE) {
        return UPDATE;
   }
   elsif ($self->{action} eq INSERT) {
        return INSERT;
   }
   return undef;
}

sub get_id {
    my ($self) = @_;

    return $self->{id};
}

sub get_field_value {
    my ($self, $field) = @_;

    return $self->{record_hash}->{$field};
}

sub set_cache_table_def {
    my ($self) = @_;
    
    $tables->{$self->{table_name}} = {};
    # Use DBI's column_info call to get a list of columns for the table
    my $sth = $self->{entity}->get_dbh->column_info( undef, undef, $self->{table_name}, '%' );
    my $collection = $sth->fetchall_hashref('COLUMN_NAME');
    $sth->finish();
        
    # Run through array returned and set each key name to undef in record_hash
    while (my ($key, $value) = each (%$collection)){
        $tables->{$self->{table_name}}->{$key} = undef;
        # Set the IS_NULLABLE flag to be used later to avoid inserting or updating when a NOT NULL field
        # is set to be updated or inserted with an undefined (NULL) value
        # Default to FALSE unless YES is found
        $tables->{$self->{table_name}}->{$key}->{IS_NULLABLE} = (($value->{IS_NULLABLE} eq 'YES') ? TRUE: FALSE);
    }
}

# Set existing values retrieved from the table to NULL (undef) so that they will get overwritten as null
# when saving data.
sub set_existing_as_null {
    my ($self) = @_;
    
    # Don't do this unless this record is set to be updated
    return unless ($self->{action} eq UPDATE);
    # Don't do this unless the record has been retrieved already
    return unless ($self->{record_retrieved} eq TRUE);
    
#    $self->{log}->log("[DBG] Setting existing as null");
    while (my ($key, $value) = each (%{$self->{record_hash}})) {
        # Run through all fields not in the lookup list or 'id'
        unless (($self->{record_lookup_fields_hash}->{$key} eq TRUE) || ($key eq 'id')) {
            my $val = $self->{record_lookup_fields_hash}->{$key};
            # Set field to undef (null)
            $self->{record_hash}->{$key} = undef;
            # Set field as changed
            $self->{record_change_hash}->{$key} = CHANGED;
            # Setting this field to have it saved as null
            $self->{record_update_force_null_hash}->{$key} = TRUE;
    	}
    }
}

sub retrieve {
    my ($self, $keys_values, $allow_null_values) = @_;

    unless (ref($keys_values) eq "HASH") {
        $self->{log}->log("[DBG] retrieve_all_fields_by_key requires a hash as argument");
        return undef;
    }
    $self->reset();

    # Get the table definition (i.e. field list) from module cache if possible
    # If this is not done there will be one select to get the table field name list
    # plus one select per field to get type etc. information for EVERY record not found
    # for every table. Not good.
    unless (exists $tables->{$self->{table_name}}) {
        $self->set_cache_table_def();
    }

    # Get a single hash representing the row found (or undef if nothing found)
    my $ret_val;
    if (defined $allow_null_values) {
#    	print "Allowing nulls\n";
        $ret_val = $self->{entity}->retrieve_all_fields_by_key_allow_null_values($keys_values);
    }
    else {
        $ret_val = $self->{entity}->retrieve_all_fields_by_key($keys_values, undef);
    }
    $self->{record_retrieved} = TRUE;
    
    # TODO - Add a check for valid field names
    # Handle case where data exists and the action will be UPDATE
    if (defined $ret_val) {
        # Create a new tables cache for current table if it doesn't exist based on return value of query
    	unless (exists $tables->{$self->{table_name}}) {
    		$tables->{$self->{table_name}} = {};
            while (my ($key, $value) = each (%$ret_val)) {
                $tables->{$self->{table_name}}->{$key} = undef;
#                $self->{record_lookup_fields_hash}->{$key} = TRUE;
            }
    	}
        # Flag all incoming fields as lookup fields
        while (my ($key, $value) = each (%$keys_values)) {
            # Flag incoming field as lookup field
            $self->{record_lookup_fields_hash}->{$key} = TRUE;
        }
        $self->{record_hash} = $ret_val;
        
        $self->{id} = $self->{record_hash}->{id};
        unless ($self->{id}) {
            $self->{log}->log("[DBG] No id field found for table: ".$self->{table_name});
        }
        else {
            $self->{action} = UPDATE;
        }
    }
    # Handle case where no match found and the action will be INSERT.
    else {
        # Get the table definition (i.e. field list) from module cache if possible
        # If this is not done there will be one select to get the table field name list
        # plus one select per field to get type etc. information for EVERY record not found
        # for every table. Not good.
#        unless (exists $tables->{$self->{table_name}}) {
#            $self->set_cache_table_def();
#        }

        while (my ($key, $value) = each (%{$tables->{$self->{table_name}}})) {
            $self->{record_hash}->{$key} = undef;
        }

        # column_info doesn't get the id field, so add it for consistency
        # (will have to remove it when saving)
#        $self->{record_hash}->{id} = undef;
        # Set action to "INSERT" so that when "save" is called we can do that properly
        $self->{action} = INSERT;
        # Populate the various hashes
        while (my ($key, $value) = each (%$keys_values)) {
           if (exists $self->{record_hash}->{$key}) {
                # Set incoming field value in the record_hash hash
                $self->{record_hash}->{$key} = $value;
                # Flag incoming field as lookup field
                $self->{record_lookup_fields_hash}->{$key} = TRUE;
                # Flag incoming field as changed so it gets saved
                $self->{record_change_hash}->{$key} = CHANGED;
            }
        }
    }
    
}

sub set {
    my ($self, $keys_values, $force_null) = @_;
    
    my $ret_val = -1;

    my $ref = ref($keys_values);
    
    if (! defined $ref) {
        $self->{log}->log("[DBG] No valid value passed to 'set'.");
    }
    unless (ref($keys_values) eq "HASH") {
        $self->{log}->log("[DBG] Argument 1 to method 'set' must be a hash.");
    }
    else {
    	# Loop through key/value set passed in, setting any that are not part of the
    	# initial lookup field list or the id field, and that are non-empty and different
    	# from the starting value for the field.
        while (my ($key, $value) = each (%$keys_values)) {
        	next if (! defined $key);
        	next if $key =~ /^\s*$/;
        	next if $value =~ /^\s*$/;
            # Lowercase, just in case - another of the many sportsdb-specific things done
            $key = lc($key);
            $key =~ s/-/_/g;
            # Don't overwrite id field since it's the primary id for tables
            next if $key eq 'id';
            # Don't overwrite a field used in initial lookup
            next if exists $self->{record_lookup_fields_hash}->{$key};
            # Only set incoming fields if they are defined for the table
            if (exists $self->{record_hash}->{$key}) {
                # If a value is empty, skip setting it unless force_null has been set
                if ((! defined $value) ||( $value =~ /^\s*$/)) {
                    if ((defined $force_null) && (! defined $value)) {
                        # Set to undef (null) rather than undef or emtpy string
                        $self->{record_hash}->{$key} = undef;
                        $self->{record_change_hash}->{$key} = CHANGED;
                        $self->{record_update_force_null_hash}->{$key} = TRUE;
                    }
                    else {
                        next;
                    }
                }
            	# Set a field if it has changed in some way
                unless ($self->{record_hash}->{$key} eq $value) {
                    $self->{record_hash}->{$key} = $value;
                    $self->{record_change_hash}->{$key} = CHANGED;
                }
            }
            else {
            	my @keys = keys %{$self->{record_hash}};
            	my $action = $self->{action};
                my $ref = ref $key;
                $self->{log}->log("[DBG] Can't set key: $key value: $value since it is not found in field list for table ".$self->{table_name});
            }
        }
        return 1;
    }
    return $ret_val;
}

sub is_changed {
    my ($self) = @_;

    my $changed_count = scalar (keys %{$self->{record_change_hash}});
    my $changed = ($changed_count eq "0") ? UNCHANGED: CHANGED;
    return $changed;
}

sub save {
    my ($self, $update_keys) = @_;
    
    my $changed = $self->is_changed();
    
    my $id = $self->{record_hash}->{id};
    my $save_field_hash = {};
    # Populate key/value by key/value since
    # my $save_field_hash = $self->{record_hash}; produces a pointer to a pointer
    while (my ($key, $value) = each (%{$self->{record_hash}})){
        $save_field_hash->{$key} = $value;
    }
    delete($save_field_hash->{id});

    my $update_hash = {};
    
    # TODO - Check to ensure it's an array refernce and action is UPDATE
    if (defined $update_keys) {
        foreach (@$update_keys) {
            if (exists $self->{record_change_hash}->{$_}) {
                $update_hash->{$_} = $self->{record_change_hash}->{$_};
            }
            else {
                return;
            }
        } 
    }
    else {
        $update_hash->{id} = $id;
    }

    my $save_str = "";
    
    while (my ($key, $value) = each(%$save_field_hash)){
     $save_str = "$save_str $key=>$value id: $id";
    }
    

    my $save_all = 0;
    my $quiet = 1;

    # For testing purposes - Do save everything that has a value all the time
    if ($save_all eq '1') {
        if ($self->{action} eq UPDATE) {
            # Do an update
            $self->{entity}->update($save_field_hash, $update_hash);
            # We could after this do more changes and then save again
#            $self->set_as_unchanged();
            $self->{log}->log("[DBG] redundant (testing) UPDATE done for table $self->{table_name}") if ($quiet eq '0');
        }
        else {
            # Do an insert
            $self->{entity}->create($save_field_hash);
            $id = $self->{entity}->retrieve_max_id();
            # We could after this do more changes and then save again
#            $self->set_as_unchanged();
            $self->{log}->log("[DBG] redundant (testing) INSERT done for table $self->{table_name}") if ($quiet eq '0');
        }
        return $id;
    }

    # Do an update or insert if anything was changed
    unless ($changed eq UNCHANGED) {
        # Remove any items not changed
        # Remove all fields not changed
        while (my ($key, $value) = each (%{$self->{record_hash}})){
            # If ANY fields have force null set for them then we will call update_allow_nulls
            if (exists $self->{record_update_force_null_hash}->{$key}) {
                $self->{overwrite_nulls} = TRUE;
            }
            # Remove the field from the saved list unless it has been protected for null-ing
            unless (exists $self->{record_change_hash}->{$key}) {
                unless (exists $self->{record_update_force_null_hash}->{$key}) {
                    delete $save_field_hash->{$key};
                }
            }
        }
        if ($self->{action} eq UPDATE) {
            my $count = keys %$save_field_hash;
            # This should not happen!
            if ($count eq '0') {
                $self->{log}->log("[DBG] NO update done for table $self->{table_name} - no non-lookup fields changed") if ($quiet eq '0');
                return $id;
            }
            # Do a check of all fields to be updated and ensure that all NOT NULL fields have a not null value for the update
            # This logic does not check to see if the value makes sense, only that it is not null
            while (my ($key, $value) = each (%{$save_field_hash})){
                # If the value to update is NOT NULLable but is actually set as null (undef) leave without updating
                if (($tables->{$self->{table_name}}->{$key}->{IS_NULLABLE} eq FALSE) && (! defined $value)) {
                    $self->{log}->log("[DBG] NO update done for table $self->{table_name} - NULL value in non-nullable field $key");
                    return;
                }
            }
            # Do an update
            if ($self->{overwrite_nulls} eq TRUE) {
                $self->{entity}->update_allow_null_values($save_field_hash, $update_hash);
            }
            else {
                $self->{entity}->update($save_field_hash, $update_hash);
            }
            # We could after this do more changes and then save again
#            $self->set_as_unchanged();
            my $keys_values_str;
            while ( my ($key, $value) = each(%$save_field_hash) ) {
                $keys_values_str = "$keys_values_str $key => $value";
            }
            $self->{log}->log("[DBG] UPDATE done for table $self->{table_name} $keys_values_str id: $self->{id}") if ($quiet eq '0');
        }
        else {
            # Do a check of all fields and ensure that all NOT NULL fields have a not null value for the insert
            # This logic does not check to see if the value makes sense, only that it is not null
            while (my ($key, $not_needed) = each (%{$self->{record_hash}})) {
                # We don't update the id field so don't deal with that
                next if ($key eq 'id');
                if ($tables->{$self->{table_name}}->{$key}->{IS_NULLABLE} eq FALSE) {
                    my $value = $save_field_hash->{$key};
                    # If the hash containing keys and values to save has a non null value we need to leave without insert
                    if (! defined $value) {
                        $self->{log}->log("[DBG] NO insert done for table $self->{table_name} - NULL value in non-nullable field $key");
                        return;
                    }
                }
            }

            # Do an insert
            # TODO Check to make sure it worked
            $id = $self->{entity}->create($save_field_hash);
#            $id = $self->{entity}->retrieve_max_id();
            # We could after this do more changes and then save again
#            $self->set_as_unchanged();
            # Set the id field now that there is a record
            $self->{record_change_hash}->{id} = $id;
            $self->{log}->log("[DBG] INSERT done for table $self->{table_name}") if ($quiet eq '0');
        }
    }
    else {
        $self->{log}->log("[DBG] UPDATE avoided for table $self->{table_name}") if ($quiet eq '0');
    }
    return $id;
}

1;