# $Id: SpecificStats.pm 3076 2016-04-07 19:54:02Z ian $

package FeedfetcherDeluxe::SpecificStats;

use FileHandle;
use DirHandle;
use XML::LibXML;
use Carp;
use FeedfetcherDeluxe::XtossUtil;
use FeedfetcherDeluxe::SpecificStatsStore;
use Core::Util;
use Core::XmlUtil;
use Core::ActiveFixtures;
use FeedfetcherDeluxe::Entities::Season;
use FeedfetcherDeluxe::Entities::Sub_season;
use FeedfetcherDeluxe::Entities::Person;
use FeedfetcherDeluxe::Entities::Team;
use FeedfetcherDeluxe::Entities::Affiliation;
use FeedfetcherDeluxe::Entities::Event;
use FeedfetcherDeluxe::Entities::Position;

use strict;

#my $loadflexsport = 0;
#my $loadxtoss = 1;

my $xtoss_table_prefix = '';

sub new {
	my ( $class, $dbh, $cfg, $log ) = @_;

	my $self = {};
	bless $self, $class;
	$self->_initialize( $dbh, $cfg, $log );

	return $self;
}

sub _initialize {
	my ( $self, $dbh, $cfg, $log ) = @_;

	$self->{'cfg'} = $cfg;
	$self->{'log'} = $log;

	$self->{xtutil} = new FeedfetcherDeluxe::XtossUtil( $dbh, $self->{log} );
	$self->{'active_fixtures'} =
	  Core::ActiveFixtures::load( $cfg, 'active-fixtures-stats' );
	$self->{'table'}  = {};
	$self->{'fields'} = {};
	$self->{'sports'} = {};
	$self->{dbh}      = $dbh;

	$self->_parse_defs($cfg);
	$self->{'dbstore'} =
	  new FeedfetcherDeluxe::SpecificStatsStore( $dbh, $self->{'fields'}, $cfg,
		$log );

	# Extra SQL Prepare statements

	# (copied from DbStore)

	# xtossql: events / id select
	$self->{'xtoss_event_exists'} = $dbh->prepare(
		q{
		SELECT id
		FROM events
		WHERE event_key=?
		AND publisher_id=?
		}
	);

	# xtossql: teams / select
	$self->{'xtoss_teams_select'} = $dbh->prepare(
		q{
		SELECT id
		FROM teams
		WHERE team_key=? AND publisher_id=?
		}
	);

	# fixit later, also add league_id, home_site_id
	# xtossql: teams / insert
	$self->{'xtoss_teams_insert'} = $dbh->prepare(
		q{
		INSERT INTO teams (team_key, publisher_id)
		VALUES (?, ?)
		}
	);

	# fixit later, also add first_name, last_name, etc.
	# xtossql: display_names / insert
	$self->{'xtoss_display_names_insert'} = $dbh->prepare(
		q{
		INSERT INTO display_names (language, entity_id, entity_type, full_name)
		VALUES (?, ?, ?, ?)
		}
	);

	# xtossql: teams / select -- get ID of most recently created row
	$self->{'xtoss_latest_teams'} = $dbh->prepare(
		q{
		SELECT max(id) from teams
		}
	);

	# xtossql: persons / select
	$self->{'xtoss_persons_select'} = $dbh->prepare(
		q{
		SELECT id
		FROM persons
		WHERE person_key=? AND publisher_id=?
		}
	);

	# xtossql: persons / insert
	$self->{'xtoss_persons_insert'} = $dbh->prepare(
		q{
		INSERT INTO persons (person_key, publisher_id)
		VALUES (?, ?)
		}
	);

	# xtossql: persons / select
	$self->{'xtoss_latest_persons'} = $dbh->prepare(
		q{
		SELECT max(id) from persons
		}
	);
}

sub types {
	my ($self) = @_;

	return $self->{'table'};
}

sub sports {
	my ($self) = @_;
	return keys %{ $self->{'sports'} };
}

sub all_specific_groups {
	my ($self) = @_;

	my @sports = $self->sports();
	my @groups = ();
	foreach my $sport (@sports) {
		push @groups, $self->specific_groups($sport);
	}

	return @groups;
}

sub groups {
	my ( $self, $sport_key ) = @_;

	return () if !defined $sport_key;
	return ( $self->core_groups, $self->specific_groups($sport_key) );
}

sub fields {
	my ( $self, $group ) = @_;

	return $self->{'fields'}->{$group};
}

sub specific_groups {
	my ( $self, $sport_key ) = @_;

	return () if !defined $sport_key;
	my $groups = $self->{'sports'}->{$sport_key};
	if ( !$groups || !@$groups ) {

		# $self->{log}->log("[DBG] none");

		#		die "No tables defined for sport-key: $sport_key";
		return ();
	}
	return @$groups;
}

sub core_groups {
	my ($self) = @_;
	return $self->specific_groups('15000000');
	return ();
}

sub _parse_defs {
	my ( $self, $cfg ) = @_;

	my $log     = $self->{'log'};
	my $def_dir = $cfg->param('db_sport_maps_dir');

	my $dir = new DirHandle $def_dir or croak "No such directory: $def_dir";
	while ( my $file = $dir->read ) {
		next if $file !~ /\.def$/;
		my $path = "$def_dir/$file";
		$self->_parse_def($path);
	}
}

sub _parse_def {
	my ( $self, $filepath ) = @_;

	# Read in def file
	my $file = new FileHandle $filepath
	  or croak "Cannot open definition file: $filepath";

	my ( $group, $stat, $type );
	my $table  = $self->{'table'};
	my $fields = $self->{'fields'};
	my $sports = $self->{'sports'};
	my $sport_key;

	while ( my $line = <$file> ) {
		chomp($line);
		next if $line =~ /^\s*\#/ || $line =~ /^\s*$/;
		$line = Core::Util::trim($line);

		$line = _dashes2underscores($line);
		if ( $line =~ /^!(\w+)\s*$/ ) {
			$sport_key = $1;
			$sports->{$sport_key} = [];
		} elsif ( $line =~ /^\[(\w+)\]/ ) {
			croak "No sport-key defined in $filepath" if !$sport_key;
			$group            = $1;
			$table->{$group}  = {};
			$fields->{$group} = [];
			push @{ $sports->{$sport_key} }, $group;
		} else {
			( $stat, $type ) = Core::Util::trims( split /[;:\t]/, $line );
			$stat =~ s/\|/\:/;
			$table->{$group}->{$stat} = $type;
			push @{ $fields->{$group} }, $stat;
		}
	}
	undef $file;
}

sub store_stats {
	my ( $self, $doc, $fixture_key, $league_key, $publisher_key, $document_id,
		$publisher_id, $document_fixture_id )
	  = @_;

	my $log = $self->{'log'};

	return
	  if !Core::ActiveFixtures::doc_lookup( $self->{'active_fixtures'},
		$doc, $self->{'cfg'} );

	$self->_store_groups_teams( $doc, $publisher_key, $publisher_id );
	$self->_store_groups_players( $doc, $publisher_key, $publisher_id );
	$self->_store_groups_events( $doc, $publisher_key, $publisher_id );
}

sub _store_groups_events {
	my ( $self, $doc, $publisher_key, $publisher_id ) = @_;

	my @event_stats_nodes = $doc->findnodes("//sports-event/event-stats");

	# Unless findnodes found elements, return
	return unless @event_stats_nodes;

	my $sportsml_group_name;

	my $deletable_stat_repository_id;
	my $new_stat_repository_type;
	my $new_stat_repository_id;
	my $season_entity =
	  FeedfetcherDeluxe::Entities::Season->new( $self->{dbh}, $self->{log} );
	my $affiliation_entity =
	  FeedfetcherDeluxe::Entities::Affiliation->new( $self->{dbh},
		$self->{log} );
	my $event_entity =
	  FeedfetcherDeluxe::Entities::Event->new( $self->{dbh}, $self->{log} );
	my $context;
	my $date_coverage_type;
	my $date_coverage_value;
	my $date_coverage_value_id;

	my $stats_record_mapper =
	  FeedfetcherDeluxe::Entities::MetaRecordMapper->new( $self->{dbh},
		$self->{log}, 'stats' );
	my $specific_stats_record_mapper =
	  FeedfetcherDeluxe::Entities::MetaRecordMapper->new( $self->{dbh},
		$self->{log}, undef );

	foreach my $event_stats_node (@event_stats_nodes) {
		my $event_metadata_node =
		  solonode( $event_stats_node, "preceding-sibling::event-metadata" );
		if ( !defined $event_metadata_node ) {
			$self->{log}->log(
"[ADV] Could not find event metadata node for event stats node. Skipping storage."
			);
			next;
		}

		my $event_key = attribute( $event_metadata_node, "event-key" );
		if ( !defined $event_key ) {
			$self->{log}->log(
"[ADV] Could not find event key for event stats node. Skipping storage."
			);
			next;
		}
		my $event_id = $event_entity->retrieve_id_by_key(
			{ 'event_key' => $event_key, 'publisher_id' => $publisher_id } );
		if ( !defined $event_id ) {
			$self->{log}->log(
"[ADV] Could not find event id for event key $event_key for event stats node. Skipping storage."
			);
			next;
		}
		$context                = 'events';
		$date_coverage_type     = 'events';
		$date_coverage_value_id = $event_id;

  #        (
  #            $context, $date_coverage_type, $date_coverage_value,
  #            $date_coverage_value_id
  #        ) = $season_entity->get_coverage( $event_stats_node, $publisher_id );

		my $stat_holder_type = 'events';
		my $stat_holder_id   = $date_coverage_value_id;

		my $membership_type;
		my $membership_id;
		my $league_id =
		  $affiliation_entity->find_league_id( $event_stats_node,
			$publisher_id );
		if ( defined $league_id ) {
			$membership_type = 'affiliations';
			$membership_id   = $league_id;
		}

# Get out if no league found, since this is required for membership storage in stats table
		else {
			$self->{log}->log(
"[ADV] Could not find league for event stats node. Skipping storage."
			);
			last;
		}
		my $sport_key = $self->_get_sport_key($event_stats_node);
		my @groups    = $self->groups($sport_key);

		foreach my $group (@groups) {
			$sportsml_group_name = $group;   #$self->convert_group_name($group);
			my $xtoss_group_name = $self->{xtutil}->get_xtoss_name($group);

			my @group_nodes =
			  nodes( $event_stats_node,
				'descendant::' . _underscores2dashes($group) );

			# Avoid doing work on non-existent nodes
			next unless @group_nodes;

# Spin through all nodes for this group to find out if there is a non-empty one
# If so, assume that previous entries can be deleted and new ones added
# This is a bit simplistic in that it does not guard against all bad data but it keeps
# more simple empty attribute cases at bay
			my $non_empty_group_node = 0;
			foreach my $group_node (@group_nodes) {
				if ( $group_node->hasAttributes() ) {
					$non_empty_group_node = 1;
					last;
				}
			}

# If there's nothing to add (i.e. no attributes to store, don't delete it all then re-add nothing
			next if ( $non_empty_group_node == 0 );

		 # first find the stats group PROPERLY... by doing a join on xtoss_stats
			$deletable_stat_repository_id = '';

			# Find out SportsDB table name
			my $stat_repository_type = $self->{xtutil}->get_xtoss_name($group);

			# Find existing record
			my $tests = {
				'stat_holder_type'     => $stat_holder_type,
				'stat_holder_id'       => $stat_holder_id,
				'stat_coverage_type'   => $date_coverage_type,
				'stat_coverage_id'     => $date_coverage_value_id,
				'stat_repository_type' => $stat_repository_type,
				'stat_membership_type' => $membership_type,
				'stat_membership_id'   => $membership_id,
				'context'              => $context
			};

			$stats_record_mapper->retrieve($tests);
			my $stats_mapper_action = $stats_record_mapper->get_action();
			my $stats_table_id =
			  $stats_record_mapper->get_field_value('stat_repository_id');

			# See if we can find the specific record
			$specific_stats_record_mapper->reset($stat_repository_type);

#                $specific_stats_record_mapper = FeedfetcherDeluxe::Entities::MetaRecordMapper->new($self->{dbh}, $self->{log}, $stat_repository_type);
			$specific_stats_record_mapper->retrieve(
				{ 'id' => $stats_table_id } );
			my $specific_stats_mapper_action =
			  $stats_record_mapper->get_action();

			foreach my $group_node (@group_nodes) {

				my $fields = $self->{'fields'}->{$group};
				croak "No definition file for: $group" if !defined $fields;

				my $rec = {};

				my $has_data = undef;

				# fields are the attributes
				foreach my $field (@$fields) {
					my $sportsdb_field_name = $field;
					$sportsdb_field_name =~ s/-/_/g;
					$sportsdb_field_name =~ s/^xts\://;

					my $value =
					  $group_node->getAttribute( _underscores2dashes($field) );
					$value = undef if ( $value =~ /^\s*$/ );

		   # TODO Verify this fix to matching sporsml/sportsdb attributes/feilds
					my ( $new_field_name, $new_value ) =
					  $self->{xtutil}
					  ->get_xtoss_column_info( $sportsdb_field_name, $value );

		   #        print "field: $field value: $value new_value: $new_value\n";
					$sportsdb_field_name = $new_field_name;
					$value               = $new_value;

					$rec->{$sportsdb_field_name} = $value;
					$has_data = 1 if defined $value;
				}

				if ($has_data) {
					$specific_stats_record_mapper->set($rec);
					my $specific_stats_id =
					  $specific_stats_record_mapper->save();

					if ( $specific_stats_mapper_action eq "INSERT" ) {
						$stats_record_mapper->set(
							{ 'stat_repository_id' => $specific_stats_id } );
					}
					$stats_record_mapper->save();
				}
			}
		}
	}
}

# Store stats for teams, including linkages back to the stats table
sub _store_groups_teams {
	my ( $self, $doc, $publisher_key, $publisher_id ) = @_;

	#        $self->{log}->log("[DBG] Storing stats");

	my $sportsml_group_name;

	my $deletable_stat_repository_id;
	my $new_stat_repository_type;
	my $new_stat_repository_id;
	my $season_entity =
	  FeedfetcherDeluxe::Entities::Season->new( $self->{dbh}, $self->{log} );
	my $team_id;
	my $context;
	my $date_coverage_type;
	my $date_coverage_value;
	my $date_coverage_value_id;
	my $team_entity =
	  FeedfetcherDeluxe::Entities::Team->new( $self->{dbh}, $self->{log} );

	my $stats_record_mapper =
	  FeedfetcherDeluxe::Entities::MetaRecordMapper->new( $self->{dbh},
		$self->{log}, 'stats' );
	my $specific_stats_record_mapper =
	  FeedfetcherDeluxe::Entities::MetaRecordMapper->new( $self->{dbh},
		$self->{log}, undef );

	#	my $test_sport_key = $self->_get_sport_key($doc);
	#	print "sport_key: $test_sport_key\n";

	my @teams = $doc->findnodes("//team");
	foreach my $team (@teams) {

		my $membership_type;
		my $membership_id;
		my $league_id = $team_entity->find_league_id( $team, $publisher_id );
		if ( defined $league_id ) {
			$membership_type = 'affiliations';
			$membership_id   = $league_id;
		}

		my $team_metadata = solonode( $team, 'team-metadata' );
		my $team_key = attribute( $team_metadata, 'team-key' );
		if ( !defined $team_key || $team_key =~ /^\s*$/ ) {
			$self->{log}->log(
"[ADV] required attribute team-key not found in team-metadata node. Skipping."
			);
			next;
		}
		my $team_stat = solonode_optional( $team, 'team-stats' );

# Debug testing code
#        $self->{log}->log("[DBG] team: $team publisher_id: $publisher_id");
		$team_id = $team_entity->retrieve_id_by_key(
			{ 'team_key' => $team_key, 'publisher_id' => $publisher_id } );
		(
			$context, $date_coverage_type, $date_coverage_value,
			$date_coverage_value_id
		) = $season_entity->get_coverage( $team, $publisher_id );

#       	$self->{log}->log("[DBG] ***** context: $context date_coverage_type: $date_coverage_type date_coverage_value: $date_coverage_value date_coverage_value_id: $date_coverage_value_id");

		if ( !defined $context ) {
			$self->{log}->log(
"[DBG] ***** No coverage value found for team stats node. Skipping. $date_coverage_type $date_coverage_value $date_coverage_value_id"
			);
			next;
		}

		if ($team_stat) {    # process actual stats

			my $sport_key = $self->_get_sport_key($team);
			my @groups    = $self->groups($sport_key);

			foreach my $group (@groups) {
				$sportsml_group_name = $group;

				#$self->convert_group_name($group);
				my $xtoss_group_name = $self->{xtutil}->get_xtoss_name($group);

				my @group_nodes =
				  nodes( $team_stat,
					'descendant-or-self::' . _underscores2dashes($group) );

				# Avoid doing work on non-existent nodes
				next unless @group_nodes;

# Spin through all nodes for this group to find out if there is a non-empty one
# If so, assume that previous entries can be deleted and new ones added
# This is a bit simplistic in that it does not guard against all bad data but it keeps
# more simple empty attribute cases at bay
				my $non_empty_group_node = 0;
				foreach my $group_node (@group_nodes) {
					if ( $group_node->hasAttributes() ) {
						$non_empty_group_node = 1;
						last;
					}
				}

#                print "Group: $group, non_empty_group_node: $non_empty_group_node\n";
# If there's nothing to add (i.e. no attributes to store, don't delete it all then re-add nothing
				next if ( $non_empty_group_node == 0 );

		 # first find the stats group PROPERLY... by doing a join on xtoss_stats
				$deletable_stat_repository_id = '';

				# Find out SportsDB table name
				my $stat_repository_type =
				  $self->{xtutil}->get_xtoss_name($group);

				# Find existing record
				my $tests = {
					'stat_holder_type' => 'teams',     # 'teams'
					'stat_holder_id'   => $team_id,    # team_id
					'stat_coverage_type' =>
					  $date_coverage_type,             # events, seasons
					'stat_coverage_id' =>
					  $date_coverage_value_id,         # event_id, season_id
					'stat_repository_type' =>
					  $stat_repository_type,           # table name
					'stat_membership_type' => $membership_type,  # affiliations
					'stat_membership_id'   => $membership_id,    # league_id
					'context'              => $context           # event, season
				};

				foreach my $group_node (@group_nodes) {

					my $stats_scope = $group_node->getAttribute('scope-code')
					  || $group_node->getAttribute('xts:scope-code');

					my $specific_tests = $tests;
					if ( defined $stats_scope ) {
						$specific_tests->{'scope'} = $stats_scope;
					} else {
						$specific_tests->{'scope'} = undef;
						my @keys = keys %$specific_tests;

						#                        print "team context: @keys\n";
					}

					$stats_record_mapper->reset('stats');
					$stats_record_mapper->retrieve( $specific_tests, 1 );
					my $stats_mapper_action =
					  $stats_record_mapper->get_action();
					my $stats_table_id = $stats_record_mapper->get_field_value(
						'stat_repository_id');

					# See if we can find the specific record
					$specific_stats_record_mapper->reset($stat_repository_type);

					$specific_stats_record_mapper->retrieve(
						{ 'id' => $stats_table_id } );
					my $specific_stats_mapper_action =
					  $stats_record_mapper->get_action();

					my $fields = $self->{'fields'}->{$group};
					croak "No definition file for: $group" if !defined $fields;

					my $rec = {};

					my $has_data = undef;

					# fields are the attributes
					foreach my $field (@$fields) {
						my $sportsdb_field_name = $field;
						$sportsdb_field_name =~ s/-/_/g;
						$sportsdb_field_name =~ s/^xts\://;

						my $value = $group_node->getAttribute(
							_underscores2dashes($field) );
						$value = undef if ( $value =~ /^\s*$/ );

		   # TODO Verify this fix to matching sporsml/sportsdb attributes/feilds
						my ( $new_field_name, $new_value ) =
						  $self->{xtutil}
						  ->get_xtoss_column_info( $sportsdb_field_name,
							$value );

		   #        print "field: $field value: $value new_value: $new_value\n";
						$sportsdb_field_name = $new_field_name;
						$value               = $new_value;

				 #						$rec->{$sportsdb_field_name} = $value;
				 # Add the value if it is defined and if another value has not
				 # been added for the same field (for xts: namespace duplicates)
						if ( !$rec->{$sportsdb_field_name} && defined $value ) {

#                           	print "adding: $sportsdb_field_name value: $value\n";
							$rec->{$sportsdb_field_name} = $value;
						}
						$has_data = 1 if defined $value;
					}

			   #					print "table: $stat_repository_type has_data: $has_data\n";

					if ($has_data) {

# Debug testing code
#                $self->{log}->log("[DBG] group: $sportsml_group_name") if $sportsml_group_name =~ /penalty_stats/;
#                $self->{log}->log("[DBG] group: $sportsml_group_name"); # if $sportsml_group_name =~ /team_stats_american_football/;

						$specific_stats_record_mapper->set($rec);
						my $specific_stats_id =
						  $specific_stats_record_mapper->save();

						if ( $specific_stats_mapper_action eq "INSERT" ) {
							$stats_record_mapper->set(
								{ 'stat_repository_id' => $specific_stats_id }
							);
						}
						$stats_record_mapper->save();
					}

#					else {
#                $self->{log}->log("[DBG] unsaved group: $sportsml_group_name"); # if $sportsml_group_name =~ /team_stats_american_football/;
#					}
				}
			}
		}
	}
}

# fixit: copied over from DbStore... should just call it remotely
sub _get_name {
	my ( $self, $name_node ) = @_;

	my ( $first, $middle, $last, $full, $nickname ) = (
		$name_node->getAttribute('first'),
		$name_node->getAttribute('middle'),
		$name_node->getAttribute('last'),
		$name_node->getAttribute('full'),
		$name_node->getAttribute('nickname')
	);

	if ( !$full ) {
		my @names = ();
		push @names, $first  if $first;
		push @names, $middle if $middle;
		push @names, $last   if $last;
		$full = join( " ", @names );
	}

	return $full ? ( $first, $middle, $last, $full, $nickname ) : undef;
}

# TODO Move the seasons work out to season and sub_season entities
sub _store_groups_players {
	my ( $self, $doc, $publisher_key, $publisher_id ) = @_;

	my $sportsml_group_name;

	my $deletable_stat_repository_id;
	my $new_stat_repository_type;
	my $new_stat_repository_id;
	my $season_entity = FeedfetcherDeluxe::Entities::Season->new( $self->{dbh}, $self->{log} );
	my $player_id;
	my $context;
	my $date_coverage_type;
	my $date_coverage_value;
	my $date_coverage_value_id;
	my $person_entity = FeedfetcherDeluxe::Entities::Person->new( $self->{dbh}, $self->{log} );
	my $team_entity = FeedfetcherDeluxe::Entities::Team->new( $self->{dbh}, $self->{log} );
	my $stats_record_mapper = FeedfetcherDeluxe::Entities::MetaRecordMapper->new( $self->{dbh},
		$self->{log}, 'stats' );
	my $specific_stats_record_mapper = FeedfetcherDeluxe::Entities::MetaRecordMapper->new( $self->{dbh},
		$self->{log}, undef );
    my $affiliation_entity = FeedfetcherDeluxe::Entities::Affiliation->new( $self->{dbh}, $self->{log} );

	# Commenting out - this is done already in DbStore.pm
	#    $person_entity->store_all_from_doc($doc, $publisher_id);

	my @players = $doc->findnodes("//player");
	foreach my $player (@players) {
		my $membership_type;
		my $membership_id;
		my $team_id = $person_entity->find_team_id( $player, $publisher_id );
		my $league_id = $affiliation_entity->find_league_id( $player, $publisher_id );

		if ( defined $team_id ) {
			$membership_type = 'teams';
			$membership_id   = $team_id;
		} else {

	 # For sports with no teams but with leagues, try to find league id
	 # and add it as a membership of type 'affiliations' and id of the league id
			if ( defined $league_id ) {
				$membership_type = 'affiliations';
				$membership_id   = $league_id;
			}
		}
		my $player_metadata = solonode( $player, 'player-metadata' );

# Changed required_attribute($player_metadata, 'player-key'); to $player_metadata->getAttribute('player-key');
		my $player_key = $player_metadata->getAttribute('player-key');

		# Missing player keys have come up - imarsman 20080113
		next unless defined $player_key;

		my @stat_nodes = $player->findnodes('child::player-stats');

		$player_id = $person_entity->retrieve_id_by_key(
			{ 'person_key' => $player_key, 'publisher_id' => $publisher_id } );

		# If findnodes found elements
		if (@stat_nodes) {    # process actual stats

			my $sport_key = $self->_get_sport_key($player);
			my @groups    = $self->groups($sport_key);
			foreach my $stats_node (@stat_nodes) {
				($context, $date_coverage_type, $date_coverage_value, $date_coverage_value_id)
				  = $season_entity->get_coverage( $stats_node, $publisher_id );

				if ( $context =~ /career/ ) {
                    $date_coverage_type = 'career';
					my $statistic_metadata_node = firstnode( $stats_node,	'ancestor::statistic/statistic-metadata' );
					if ( defined $statistic_metadata_node ) {
						my $competition_scope = attribute( $statistic_metadata_node, 'competition-scope' );
						if ( !defined $competition_scope || $competition_scope =~ /^\s*$/ )
						{
							next;
						}

                        
						# Get league based on @competition-scope
						$league_id = $affiliation_entity->retrieve_id_by_key(
							{
								'affiliation_type' => 'league',
								'affiliation_key'  => $competition_scope,
								'publisher_id'     => $publisher_id
							}
						);

						# It is possible that the league does not yet exist
						if ( !defined $league_id ) {
							$league_id = $affiliation_entity->create(
								{
									'affiliation_key'  => $competition_scope,
									'affiliation_type' => 'league',
									'publisher_id'     => $publisher_id
								}
							);
						}
					}
                    $membership_type = 'affiliations';
					$membership_id = $league_id;
				}

				# Skip things like career stats for now
				next unless ( defined $context );

#			($context, $date_coverage_type, $date_coverage_value, $date_coverage_value_id) = $season_entity->get_coverage($stats_node, $publisher_id);
				foreach my $group (@groups) {
					$sportsml_group_name =
					  $group;    #$self->convert_group_name($group);
					my $xtoss_group_name =
					  $self->{xtutil}->get_xtoss_name($group);

					my @group_nodes = nodes( $stats_node,
						'descendant-or-self::'
						  . _underscores2dashes($sportsml_group_name) );

					# Avoid doing work on non-existent nodes
					next unless @group_nodes;

                    # Spin through all nodes for this group to find out if there is a non-empty one
                    # If so, assume that previous entries can be deleted and new ones added
                    # This is a bit simplistic in that it does not guard against all bad data but it keeps
                    # more simple empty attribute cases at bay
					my $non_empty_group_node = 0;
					foreach my $group_node (@group_nodes) {
						if ( $group_node->hasAttributes() ) {
							$non_empty_group_node = 1;
							last;
						}
					}

                    # If there's nothing to add (i.e. no attributes to store, don't delete it all then re-add nothing
					next if ( $non_empty_group_node == 0 );

		            # first find the stats group PROPERLY... by doing a join on xtoss_stats
					$deletable_stat_repository_id = '';

                    # 'persons' -> stat_holder_type: the holder of the stats - e.g. persons
                    # team_id -> stat_holder_id: the id for the holder
                    # context: the context of the stat: e.g. career, event, event-play
                    # date_coverage_value_id: the id for the event or season etc.
                    # xtoss_group_name: the table name
                    # membership_type: teams, players
                    # membership_id: id for membership_type

					my $node_name = $stats_node->nodeName;

					# Find out SportsDB table name
					my $stat_repository_type =
					  $self->{xtutil}->get_xtoss_name($group);

					# Find existing record
					my $tests = {
						'stat_holder_type'     => 'persons',
						'stat_holder_id'       => $player_id,
						'stat_coverage_type'   => $date_coverage_type,
						'stat_coverage_id'     => $date_coverage_value_id,
						'stat_repository_type' => $stat_repository_type,
						'stat_membership_type' => $membership_type,
						'stat_membership_id'   => $membership_id,
						'context'              => $context
					};

					foreach my $group_node (@group_nodes) {

						my $stats_scope =
						     $group_node->getAttribute('scope-code')
						  || $group_node->getAttribute('xts:scope-code');

						my $specific_tests = $tests;
						if ( defined $stats_scope ) {
							$specific_tests->{'scope'} = $stats_scope;
						} else {
							$specific_tests->{'scope'} = undef;
							my @keys = keys %$specific_tests;
						}

						$stats_record_mapper->reset('stats');
						$stats_record_mapper->retrieve( $specific_tests, 1 );
						my $stats_mapper_action =
						  $stats_record_mapper->get_action();
						my $stats_table_id =
						  $stats_record_mapper->get_field_value(
							'stat_repository_id');

						# See if we can find the specific record
						$specific_stats_record_mapper->reset(
							$stat_repository_type);

						$specific_stats_record_mapper->retrieve(
							{ 'id' => $stats_table_id } );
						my $specific_stats_mapper_action =
						  $stats_record_mapper->get_action();

						# foo this is what stores the row in specific-stats

						my @attribute_list = $group_node->attributes();

						my $fields = $self->{'fields'}->{$group};
						croak "No definition file for: $group"
						  if !defined $fields;

						my $rec = {};

						my $has_data = undef;

						# fields are the attributes
						foreach my $field (@$fields) {
							my $sportsdb_field_name = $field;
							$sportsdb_field_name =~ s/-/_/g;
							$sportsdb_field_name =~ s/^xts\://;

							my $value = $group_node->getAttribute(
								_underscores2dashes($field) );
							$value = undef if ( $value =~ /^\s*$/ );

                            # Handle position_id for core_stats
                            # If group is 'player_stats' and field is 'position_id' get position_id via other means,
                            # sinct this is not a simple case of retrieving an attribute value
                            # and substitute that for 'value'
							if (   $group eq 'player_stats'
								&& $field eq 'position_id' )
							{
								my $position_entity =
								  FeedfetcherDeluxe::Entities::Position->new(
									$self->{dbh}, $self->{log} );

                                # Grab the player-metadata node, since it contains the position value we need to look up
								my $player_metadata_node =
								  firstnode( $group_node,
									'preceding-sibling::player-metadata' );

                                # If there is something to extract it will be returned, else undef
								$value =
								  $position_entity->get_position_id_from_node(
									$player_metadata_node, 'position-regular',
									$publisher_id );
								$rec->{$sportsdb_field_name} = $value;
							} else {

                                # TODO Verify this fix to matching sporsml/sportsdb attributes/feilds

								my ( $new_field_name, $new_value ) =
								  $self->{xtutil}
								  ->get_xtoss_column_info( $sportsdb_field_name,
									$value );

								$sportsdb_field_name = $new_field_name;
								$value               = $new_value;

                                # Add the value if it is defined and if another value has not
                                # been added for the same field (for xts: namespace duplicates)
								if ( !$rec->{$sportsdb_field_name} && defined $value )
								{
									$rec->{$sportsdb_field_name} = $value;
								}
							}
							$has_data = 1 if defined $value;
						}

						if ($has_data) {
							$specific_stats_record_mapper->set($rec);
							my $specific_stats_id =
							  $specific_stats_record_mapper->save();

							if ( $specific_stats_mapper_action eq "INSERT" ) {
								$stats_record_mapper->set(
									{
										'stat_repository_id' =>
										  $specific_stats_id
									}
								);
							}
							$stats_record_mapper->save();
						}
					}
				}
			}
		}
	}
}

sub _get_sport_key {
	my ( $self, $node ) = @_;

#	my $query = "ancestor::*/*/sports-content-codes/sports-content-code[\@code-type='sport']";
	my $query =
"/sports-content/sports-metadata/sports-content-codes/sports-content-code[\@code-type='sport']";
	my $sports_content_code = firstnode( $node, $query ) or return;
	return required_attribute( $sports_content_code, 'code-key' );
}

sub _get_team_key {
	my ( $self, $node, $team_idref ) = @_;

	my $team_metadata;
	if ( defined $team_idref ) {
		$team_metadata = solonode_optional( $node,
			"ancestor::team[\@id='$team_idref']/team-metadata" );
	}

	if ( !$team_metadata ) {
		$team_metadata = firstnode( $node, "ancestor::team/team-metadata" );
	}
	if ( !$team_metadata ) {
		$self->{log}->log("[DBG] team-metadata node not found");
		return;
	}
	return required_attribute( $team_metadata, 'team-key' );
}

# Store stats to a table and return the name of the table and the row id
# for the new record so that it can be referenced from the sportsdb stats
# table.
sub _store_group {
	my ( $self, $group, $node, $publisher_id ) = @_;

	my $group_name_no_xtoss = $group;

	my $xtoss_group_name = $self->{xtutil}->get_xtoss_name($group);

	my $fields = $self->{'fields'}->{$group};
	croak "No definition file for: $group"
	  if !defined $fields;

	my $rec = [];

	#	print "fields @$fields\n";
	my $has_data = undef;

	# fields are the attributes
	foreach my $field (@$fields) {

# This is not proper - i.e. code here to deal with an exception
# Need a more general solution such as a lookup a-la XtossUtil.pm's key tranlation
# The problem here is that we can look the attribute up fine here but in SpecificStatsStore.pm
# when the stats are to be stored the call is made to a precompiled sql statement, the
# incorrect field names having been compiled in at that point already
# See 	my $rv =	$self->{"xtoss_${xtoss_table}_insert"}->execute(@rec);
# in SpecificStatsStore->store_stats_group
# NOTE leaving out for now
#		if ( $group eq 'penalty_stats' && $field eq 'penalty_type') {
#			$field = 'type';
#		}
		my $value = $node->getAttribute( _underscores2dashes($field) );

# 2009-07-27 - Setting to undef (null) if emtpy - in response to errors in SQL Server
# when trying to store values from empty ice hockey file attributes
# Not sure if the SQL Server errors were from this, but doing this is consistent
# with code practices elsewhere
		$value = undef if ( $value =~ /^\s*$/ );

# Handle position_id for core_stats
# If group is 'player_stats' and field is 'position_id' get position_id via other means,
# sinct this is not a simple case of retrieving an attribute value
# and substitute that for 'value'
		if ( $group eq 'player_stats' && $field eq 'position_id' ) {
			my $position_entity =
			  FeedfetcherDeluxe::Entities::Position->new( $self->{dbh},
				$self->{log} );

# Grab the player-metadata node, since it contains the position value we need to look up
			my $player_metadata_node =
			  firstnode( $node, 'preceding-sibling::player-metadata' );

			# If there is something to extract it will be returned, else undef
			$value =
			  $position_entity->get_position_id_from_node(
				$player_metadata_node, 'position-regular', $publisher_id );

			# Push position_id value was retrieved (or undef) on rec array
			push @$rec, $value;
		} else {
			push @$rec, $value;
		}
		$has_data = 1 if defined $value;
	}

	# foo this is what stores the row in the stats table
	my ( $stat_repository_type, $stat_repository_id ) =
	  $self->{'dbstore'}->store_stats_group( $xtoss_group_name, $rec )
	  if $has_data;

	return ( $stat_repository_type, $stat_repository_id );
}

# TODO Consider moving this into season entity
sub _get_coverage {
	my ( $self, $node ) = @_;
	my $coverage_node;

	my $date_coverage_type;
	my $date_coverage_value;

# Look down child hierarchy first (e.g. player career stats)
#<player id="l.nfl.com-p.1539">
#  <player-metadata player-key="l.nfl.com-p.1539" position-regular="5" xts:position-category-code="1" xts:position-category-name="Defense">
#   <name first="Bryant" last="Young" full="Bryant Young" />
#  </player-metadata>
# <player-stats date-coverage-type="season" date-coverage-value="2006" team-coverage="l.nfl.com-t.31" events-played="16" events-started="16">
	$coverage_node = firstnode( $node, 'child::*/*[@date-coverage-type]' );

	if ( !$coverage_node ) {

# Try for a sibling node containing season information (e.g. teams in player career stats files)
#<statistic-metadata date-coverage-type="season-regular" date-coverage-value="2006" team-coverage="single-team" />
# <team id="l.nfl.com-t.31">
		$coverage_node =
		  firstnode( $node, 'preceding-sibling::*[@date-coverage-type][1]' );
	}
	if ( !$coverage_node ) {
		$coverage_node =
		  firstnode( $node, 'ancestor::*/*[@date-coverage-type]' );
	}

	# Essentially, return undefined values if no node found
	if ( !$coverage_node ) {
		return ( $date_coverage_type, $date_coverage_value );
	}

	$date_coverage_type =
	  required_attribute( $coverage_node, 'date-coverage-type' );
	$date_coverage_value = attribute( $coverage_node, 'date-coverage-value' );
	$date_coverage_value =~ s/\s*//g;

	if ( $date_coverage_value =~ /\d+/ ) {
		$date_coverage_value ||=
		  required_attribute_xpath( $node,
			'ancestor::sports-event/event-metadata/@event-key' )
		  if $date_coverage_type eq 'event';
	} else {
		$date_coverage_value = undef;
	}
	return ( $date_coverage_type, $date_coverage_value );
}

sub _underscores2dashes {
	my ($str) = @_;

	$str =~ s/_/-/g;
	return $str;
}

sub _dashes2underscores {
	my ($str) = @_;

	$str =~ s/-/_/g;
	return $str;
}

1;
