# $Id$

package Util::DateUtils;

use Carp;
use strict;
use File::Find;
use Util::FileUtils;
use Carp qw(croak);
use Time::HiRes qw(usleep gettimeofday);
use Time::Local;
use Date::Pcalc qw( Delta_DHMS Add_Delta_DHMS );


sub new {
    my ($class) = @_;

    my $self = {};
    bless $self, $class;

    $self->{file_utils} = Util::FileUtils->new();
    return $self;
}

sub get_time_from_date_parts {
	my ($self, $sec, $min, $hour, $day, $mon, $year) = @_;
	
	$sec = (defined $sec) ? $sec : 0;
	$min = (defined $min) ? $min : 0;
	$hour = (defined $hour) ? $hour : 0;
	$day = (defined $day) ? $day : 0;
	$mon = (defined $mon) ? $mon : 0;
	$year = (defined $year) ? $year : 0;
	
	# timelocal uses a zero-based month array to do its lookups
	# so decrement month number to compensate
	my $t = timelocal($sec, $min, $hour, $day, $mon-1, $year);
	
	return $t;
}

sub get_SQL_date {
	my ($self, $time) = @_;
	
	$time = time if !defined $time;
	my $time_str = $self->epoch2ISO8601($time);
	
	return $self->ISO8601_date2SQL_date_local($time_str);
}

# NOTE: ASSUMES INCOMING TIME IS EPOCH!!!
sub get_ISO8601_date {
	my ($self, $time) = @_;

	$time = time if !defined $time;
	return $self->epoch2ISO8601($time);
}

# Get date parts from a string in format yyyymmdd with optional part delimiter (- or /)
sub get_UTC_date_parts {
    my ($self, $date) = @_;
    
    my ($year, $mon, $day);
    
    if ($date =~ /^\s*\d{1,2}[\/\-]?\d{1,2}[\/\-]?(19|2[012])\d\d\s*$/) {
#    	print "Going with TSN style date\n";
        $date =~ /^\s*(\d{1,2})[\-\/]?(\d{1,2})[\-\/]?(\d\d\d\d)\s*$/
            or croak "date formatted incorrectly: $date";
        ($mon, $day, $year) = ($1, $2, $3);
    }
    else {
#        print "Going with ISO style date\n";
        $date =~ /^\s*(\d\d\d\d)[\-\/]?(\d\d)[\-\/]?(\d\d)\s*$/
            or croak "date formatted incorrectly: $date";
        ($year, $mon, $day) = ($1, $2, $3);
    }

       
    return ($year, $mon, $day);
}

sub get_UTC_time_parts {
    my ($self, $date_time, $ignore_tz) = @_;
    
    
    $date_time =~ /^(\d\d\d\d)\-?(\d\d)\-?(\d\d)T(\d\d):?(\d\d):?(\d\d)((Z)|([+-])(\d\d):?(\d\d))?$/
        or croak "date-time formatted incorrectly: $date_time";
    
    # Month number was being incorrectly decremented by one
    my ($year, $mon, $day, $hour, $min, $sec, undef, $z, $plusminus, $zone_hr, $zone_min) =
        ($1, $2, $3, $4, $5, $6, $7, $8, $9, $10, $11);
#    print "Date and time extraction common routine\n";
#    print "Time parts: date_time: $date_time year: $year mon: $mon day: $day hour: $hour min: $min sec: $sec z: $z, plusminus: $plusminus zone_hr: $zone_hr zone_min: $zone_min\n";
#    print "Extracted datetime (prior to TZ adjustments): $year $mon $day $hour $min $sec\n";
    

    unless (defined $ignore_tz) {
        if ($z || $plusminus) {
            if ($z eq 'Z') {
                $plusminus = '+';
                $zone_hr = 0;
                $zone_min = 0;
            }

            if ($plusminus eq '+') { $zone_hr = -$zone_hr; $zone_min = -$zone_min; }
            # Factor in timezone
    #    print "$year $mon $day $hour $min $sec\n";
            ($year, $mon, $day, $hour, $min, $sec) =
                Add_Delta_DHMS($year, $mon, $day, $hour, $min, $sec,
                0 ,$zone_hr ,$zone_min ,0);
        }
    }
#    print "Adjusted datetime: $year $mon $day $hour $min $sec\n";
    return ($year, $mon, $day, $hour, $min, $sec);
}

sub epoch2ISO8601 {
	my ($self, $time) = @_;

	croak "No epoch given" if !defined $time;
	my ($utc_sec, $utc_min, $utc_hour, $utc_day, $utc_mon, 
		$utc_year, $utc_wday, $utc_yday) = gmtime($time);
	my ($sec, $min, $hour, $day, $mon, $year, $wday, $yday) = localtime($time);
	$utc_year += 1900;
	$utc_mon++;
	$year += 1900;
	$mon++;

	my ($Dd,$Dh,$Dm,$Ds) = Delta_DHMS(
		$utc_year, $utc_mon, $utc_day, $utc_hour, $utc_min, $utc_sec,
		$year, $mon, $day, $hour, $min, $sec);

	my $sign = '+';
	if ($Dh < 0) { $Dh = -$Dh; $Dm = -$Dm; $sign = '-'; }
	return sprintf("%04d%02d%02dT%02d%02d%02d%s%02d%02d", 
		$year, $mon, $day, $hour, $min, $sec, $sign, $Dh, $Dm);
}

# Transform an ISO8601 formatted datetime into a unix (epoch) formatted one
# NOTE: Returns a timestamp with the UTC timezone, which is necessary to avoid
#       shifting zones for other methods that take a timestamp and assume UTC 
#       in that timestamp.
# e.g. 20081109T132001-0500 -> 1226283601
sub ISO8601_date2time {
	my ($self, $date_time) = @_;
	
    return undef if !defined $date_time;
#    print "ISO8601_date2time\n";
    my ($year, $mon, $day, $hour, $min, $sec) = $self->get_UTC_time_parts($date_time);

	my $time;
#	print "sec, min, hour, day, mon, year: $sec, $min, $hour, $day, $mon, $year\n";
	
	# We NEED timegm since there is no timezon indicator in an epoch timestamp
	# In other words, any functions that take a timestamp assume UTC time zone
	$time = timegm($sec, $min, $hour, $day, $mon-1, $year);
#    print "$sec, $min, $hour, $day, $mon, $year\n";

	return $time;
}

sub ISO8601_date2SQL_date {
    my ($self, $date_time, $ignore_tz) = @_;

    return undef if !defined $date_time;
#    print "ISO8601_date2SQL_date date_time: $date_time\n";
    my ($year, $mon, $day, $hour, $min, $sec) = $self->get_UTC_time_parts($date_time, $ignore_tz);
    
#    $date_time =~ /^(\d\d\d\d)(\d\d)(\d\d)T(\d\d)(\d\d)(\d\d)((Z)|([+-])(\d\d)(\d\d))$/
#        or croak "date-time formatted incorrectly: $date_time";
#
#    my ($year, $mon, $day, $hour, $min, $sec, undef, $z, $plusminus, $zone_hr, $zone_min) =
#        ($1, $2, $3, $4, $5, $6, $7, $8, $9, $10, $11);
#
#    if ($z eq 'Z') {
#        $plusminus = '+';
#        $zone_hr = 0;
#        $zone_min = 0;
#    }
#	# Allow flag to allow ignoring of tz offset adjustment to GMT
#	unless (defined $ignore_tz) {
#	    if ($plusminus eq '+') { $zone_hr = -$zone_hr; $zone_min = -$zone_min; }
#    	# Factor in timezone
#    	($year, $mon, $day, $hour, $min, $sec) =
#        	Add_Delta_DHMS($year, $mon, $day, $hour, $min, $sec,
#        	0 ,$zone_hr ,$zone_min ,0);
#	}
    return sprintf("%04d-%02d-%02d %02d:%02d:%02d",
        $year, $mon, $day, $hour, $min, $sec);
}

sub ISO8601_date2SQL_date_local {
    my ($self, $date_time) = @_;

#    $date_time =~ /^(\d\d\d\d)(\d\d)(\d\d)T(\d\d)(\d\d)(\d\d)/
    $date_time =~ /^(\d\d\d\d)\-?(\d\d)\-?(\d\d)T(\d\d):?(\d\d):?(\d\d)((Z)|([+-])(\d\d)(\d\d))?$/
        or die "date-time formatted incorrectly: $date_time";
    my ($year, $mon, $day, $hour, $min, $sec) =
        ($1, $2, $3, $4, $5, $6);
    return sprintf("%04d-%02d-%02d %02d:%02d:%02d",
        $year, $mon, $day, $hour, $min, $sec);
}

sub get_date_path {
    my ($self) = @_;
    
    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime();
    my ($s, $usec) = gettimeofday();
    $mon += 1;
    $mon = sprintf("%02.2s", $mon);
    $mday = sprintf("%02.2s", $mday);
    $year += 1900;
    my $date_path = $self->{file_utils}->clean_path("${year}/${mon}/${mday}");
    
    return $date_path;
}

sub get_timestamp {
    my ($self, $use_usec) = @_;
    
    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime();
    my ($s, $usec) = gettimeofday();
    $year += 1900;
    $mon += 1;
    $mon = sprintf("%02d", $mon);
    $mday = sprintf("%02d", $mday);
    $hour = sprintf("%02d", $hour);
    $min = sprintf("%02d", $min);
    $sec = sprintf("%02d", $sec);
    $usec = sprintf("%06.6s", $usec);
    my $time_str = "${year}-${mon}-${mday}T${hour}${min}${sec}";
    $time_str = "${time_str}.${usec}" if ($use_usec eq '1');

    $time_str =~ s/^\s*(.*?)\s*$/$1/;
    
    return $time_str;
}

sub sportsml_date2sql_date {
	my ($self, $date_time) = @_;

	return undef if (! defined $date_time);
#	print "sportsml_date2sql_date\n";
    my ($year, $mon, $day, $hour, $min, $sec) = $self->get_UTC_time_parts($date_time);
    
#	$date_time =~ /^(\d\d\d\d)(\d\d)(\d\d)T(\d\d)(\d\d)(\d\d)((Z)|([+-])(\d\d)(\d\d))$/
#		or croak "date-time formatted incorrectly: $date_time";
#
#	my ($year, $mon, $day, $hour, $min, $sec, undef, $z, $plusminus, $zone_hr, $zone_min) =
#		($1, $2, $3, $4, $5, $6, $7, $8, $9, $10, $11);
#
#	if ($z eq 'Z') {
#		$plusminus = '+';
#		$zone_hr = 0;
#		$zone_min = 0;
#	}
#
#	if ($plusminus eq '+') { $zone_hr = -$zone_hr; $zone_min = -$zone_min; }
#	# Factor in timezone
#	($year, $mon, $day, $hour, $min, $sec) =
#		Add_Delta_DHMS($year, $mon, $day, $hour, $min, $sec,
#		0 ,$zone_hr ,$zone_min ,0);
	return sprintf("%04d-%02d-%02d %02d:%02d:%02d",
		$year, $mon, $day, $hour, $min, $sec);
}

sub sportsml_date2sql_date_local {
	my ($self, $date_time) = @_;

#   $date_time =~ /^(\d\d\d\d)(\d\d)(\d\d)T(\d\d)(\d\d)(\d\d)/

    # Make more forgiving
    $date_time =~ /^(\d\d\d\d)\-?(\d\d)\-?(\d\d)T(\d\d):?(\d\d):?(\d\d)/
		or die "date-time formatted incorrectly: $date_time";
	my ($year, $mon, $day, $hour, $min, $sec) =
		($1, $2, $3, $4, $5, $6);
	return sprintf("%04d-%02d-%02d %02d:%02d:%02d",
		$year, $mon, $day, $hour, $min, $sec);
}

1;